// Asynch_IO.cpp,v 4.61 2002/10/29 01:39:09 shuston Exp

#include "TProactor/Asynch_IO.h"

ACE_RCSID(ace, Asynch_IO, "Asynch_IO.cpp,v 4.61 2002/10/29 01:39:09 shuston Exp")

#if (defined (ACE_WIN32) && !defined (ACE_HAS_WINCE)) || (defined (ACE_HAS_AIO_CALLS) || defined(ACE_HAS_AIO_EMULATION))
// This only works on platforms with Asynchronous IO

#include "TProactor/Proactor.h"
#include "ace/Message_Block.h"
#include "ace/INET_Addr.h"
#include "TProactor/Asynch_IO_Impl.h"

size_t
TRB_Asynch_Result::bytes_transferred (void) const
{
  return this->implementation ()->bytes_transferred ();
}

const void *
TRB_Asynch_Result::act (void) const
{
  return this->implementation ()->act ();
}

int
TRB_Asynch_Result::success (void) const
{
  return this->implementation ()->success ();
}

const void *
TRB_Asynch_Result::completion_key (void) const
{
  return this->implementation ()->completion_key ();
}

u_long
TRB_Asynch_Result::error (void) const
{
  return this->implementation ()->error ();
}

ACE_HANDLE
TRB_Asynch_Result::event (void) const
{
  return this->implementation ()->event ();
}

u_long
TRB_Asynch_Result::offset (void) const
{
  return this->implementation ()->offset ();
}

u_long
TRB_Asynch_Result::offset_high (void) const
{
  return this->implementation ()->offset_high ();
}

int
TRB_Asynch_Result::priority (void) const
{
  return this->implementation ()->priority ();
}

int
TRB_Asynch_Result::signal_number (void) const
{
  return this->implementation ()->signal_number ();
}

TRB_Asynch_Result::TRB_Asynch_Result (TRB_Asynch_Result_Impl *implementation)
  : implementation_ (implementation)
{
}

TRB_Asynch_Result::~TRB_Asynch_Result (void)
{
  // Proactor deletes the implementation when the <complete> finishes.  
}

TRB_Asynch_Result_Impl *
TRB_Asynch_Result::implementation (void) const
{
  return this->implementation_;
}

// *********************************************************************

int
TRB_Asynch_Operation::open (TRB_Handler &handler,
                            ACE_HANDLE handle,
                            const void *completion_key,
                            TRB_Proactor *proactor)
{
  return this->implementation ()->open (handler,
                                        handle,
                                        completion_key,
                                        proactor);
}

int
TRB_Asynch_Operation::cancel (void)
{
  if (this->implementation () == 0)
    return 0;
  return this->implementation ()->cancel ();
}

int
TRB_Asynch_Operation::close (void)
{
  if (this->implementation () == 0)
    return 0;
  return this->implementation ()->close();
}

TRB_Proactor *
TRB_Asynch_Operation::proactor (void) const
{
  return this->implementation ()->proactor ();
}

TRB_Asynch_Operation_Impl *
TRB_Asynch_Operation::implementation (void) const
{
  return this->implementation_;
}

TRB_Asynch_Operation::TRB_Asynch_Operation (void)
  : implementation_ (0)
{
}

TRB_Asynch_Operation::~TRB_Asynch_Operation (void)
{
  delete this->implementation_;
  this->implementation_ = 0;
}

void
TRB_Asynch_Operation::implementation (TRB_Asynch_Operation_Impl *implementation)
{
  this->implementation_ = implementation;
}

TRB_Proactor *
TRB_Asynch_Operation::get_proactor (TRB_Proactor *user_proactor,
                                    TRB_Handler &handler) const
{
  if (user_proactor == 0)
    {
      // Grab the singleton proactor if <handler->proactor> is zero
      user_proactor = handler.proactor ();
      if (user_proactor == 0)
        user_proactor = TRB_Proactor::instance ();
    }

  return user_proactor;
}

// ************************************************************

TRB_Asynch_Read_Stream::TRB_Asynch_Read_Stream (void)
  : implementation_ (0)
{
}

TRB_Asynch_Read_Stream::~TRB_Asynch_Read_Stream (void)
{
}

int
TRB_Asynch_Read_Stream::open (TRB_Handler &handler,
                              ACE_HANDLE handle,
                              const void *completion_key,
                              TRB_Proactor *proactor)
{
  // Get a proactor for/from the user.
  proactor = this->get_proactor (proactor, handler);

  // Delete the old implementation.
  delete this->implementation_;
  this->implementation_ = 0;

  // Now let us get the implementation initialized.
  TRB_Asynch_Read_Stream_Impl *implementation = proactor->create_asynch_read_stream ();
  if (implementation == 0)
    return -1;
  
  // Set the implementation class
  this->implementation (implementation);
  
  // Call the <open> method of the base class. 
  return TRB_Asynch_Operation::open (handler,
                                     handle,
                                     completion_key,
                                     proactor);
}
  
int
TRB_Asynch_Read_Stream::read (ACE_Message_Block &message_block,
                              size_t bytes_to_read,
                              const void *act,
                              int priority,
                              int signal_number)
{
  return this->implementation ()->read (message_block,
                                        bytes_to_read,
                                        act,
                                        priority,
                                        signal_number);
}

#if (defined (ACE_WIN32) && !defined (ACE_HAS_WINCE) && (ACE_HAS_WINNT4) && (ACE_HAS_WINNT4 != 0))
int
TRB_Asynch_Read_Stream::readv (ACE_Message_Block &message_block,
                               size_t bytes_to_read,
                               const void *act,
                               int priority,
                               int signal_number)
{
  return this->implementation ()->readv (message_block,
                                         bytes_to_read,
                                         act,
                                         priority,
                                         signal_number);
}
#endif /* (defined (ACE_WIN32) && !defined (ACE_HAS_WINCE) && (ACE_HAS_WINNT4) && (ACE_HAS_WINNT4 != 0)) */

TRB_Asynch_Read_Stream_Impl *
TRB_Asynch_Read_Stream::implementation (void) const
{
  return implementation_;
}

void
TRB_Asynch_Read_Stream::implementation (TRB_Asynch_Read_Stream_Impl *implementation)
{
  this->implementation_ = implementation;

  // Init the base class' implementation also.
  TRB_Asynch_Operation::implementation (this->implementation_);
}

// ************************************************************

size_t
TRB_Asynch_Read_Stream::Result::bytes_to_read (void) const
{
  return this->implementation ()->bytes_to_read ();
}

ACE_Message_Block &
TRB_Asynch_Read_Stream::Result::message_block (void) const
{
  return this->implementation ()->message_block ();
}

ACE_HANDLE
TRB_Asynch_Read_Stream::Result::handle (void) const
{
  return this->implementation ()->handle ();
}

TRB_Asynch_Read_Stream::Result::Result (TRB_Asynch_Read_Stream_Result_Impl *implementation)
  : TRB_Asynch_Result (implementation), 
    implementation_ (implementation)
{
}

TRB_Asynch_Read_Stream::Result::~Result (void)
{
  // Proactor will delete the implementation after <complete> is
  // finished.
}

TRB_Asynch_Read_Stream_Result_Impl *
TRB_Asynch_Read_Stream::Result::implementation (void) const
{
  return this->implementation_;
}

// ***************************************************

TRB_Asynch_Write_Stream::TRB_Asynch_Write_Stream (void)
  : implementation_ (0)
{
}

TRB_Asynch_Write_Stream::~TRB_Asynch_Write_Stream (void)
{
}

int
TRB_Asynch_Write_Stream::open (TRB_Handler &handler,
                               ACE_HANDLE handle,
                               const void *completion_key,
                               TRB_Proactor *proactor)
{
  // Get a proactor for/from the user.
  proactor = this->get_proactor (proactor, handler);

  // Delete the old implementation.
  delete this->implementation_;
  this->implementation_ = 0;

  // Now let us get the implementation initialized.
  TRB_Asynch_Write_Stream_Impl *implementation = proactor->create_asynch_write_stream ();
  if (implementation == 0)
    return -1;
  
  // Set the implementation class
  this->implementation (implementation);
  
  // Call the <open> method of the base class.
  return TRB_Asynch_Operation::open (handler,
                                     handle,
                                     completion_key,
                                     proactor);
}

int
TRB_Asynch_Write_Stream::write (ACE_Message_Block &message_block,
                                size_t bytes_to_write,
                                const void *act,
                                int priority,
                                int signal_number)
{
  return this->implementation ()->write (message_block,
                                         bytes_to_write,
                                         act,
                                         priority,
                                        signal_number);
}

#if (defined (ACE_WIN32) && !defined (ACE_HAS_WINCE) && (ACE_HAS_WINNT4) && (ACE_HAS_WINNT4 != 0))
int
TRB_Asynch_Write_Stream::writev (ACE_Message_Block &message_block,
                                 size_t bytes_to_write,
                                 const void *act,
                                 int priority,
                                 int signal_number)
{
  return this->implementation ()->writev (message_block,
                                          bytes_to_write,
                                          act,
                                          priority,
                                          signal_number);
}
#endif /* (defined (ACE_WIN32) && !defined (ACE_HAS_WINCE) && (ACE_HAS_WINNT4) && (ACE_HAS_WINNT4 != 0)) */

TRB_Asynch_Write_Stream_Impl *
TRB_Asynch_Write_Stream::implementation (void) const
{
  return this->implementation_;
}

void
TRB_Asynch_Write_Stream::implementation (TRB_Asynch_Write_Stream_Impl *implementation)
{
  this->implementation_ = implementation;

  // Init the base class' implementation also.
  TRB_Asynch_Operation::implementation (this->implementation_);
}

// ************************************************************

size_t
TRB_Asynch_Write_Stream::Result::bytes_to_write (void) const
{
  return this->implementation ()->bytes_to_write ();
}

ACE_Message_Block &
TRB_Asynch_Write_Stream::Result::message_block (void) const
{
  return this->implementation ()->message_block ();
}

ACE_HANDLE
TRB_Asynch_Write_Stream::Result::handle (void) const
{
  return this->implementation ()->handle ();
}

TRB_Asynch_Write_Stream::Result::Result (TRB_Asynch_Write_Stream_Result_Impl *implementation)
  : TRB_Asynch_Result (implementation),
    implementation_ (implementation)
{
}

TRB_Asynch_Write_Stream::Result::~Result (void)
{
  // Proactor will delte the implementation when the <complete> call
  // finishes.
}

TRB_Asynch_Write_Stream_Result_Impl *
TRB_Asynch_Write_Stream::Result::implementation (void) const
{
  return this->implementation_;
}

// ************************************************************

TRB_Asynch_Read_File::TRB_Asynch_Read_File (void)
  : implementation_ (0)
{
} 

TRB_Asynch_Read_File::~TRB_Asynch_Read_File (void)
{
} 

int
TRB_Asynch_Read_File::open (TRB_Handler &handler,
                            ACE_HANDLE handle,
                            const void *completion_key,
                            TRB_Proactor *proactor)
{
  // Get a proactor for/from the user.
  proactor = this->get_proactor (proactor, handler);

  // Delete the old implementation.
  delete this->implementation_;
  this->implementation_ = 0;

  // Now let us get the implementation initialized.
  TRB_Asynch_Read_File_Impl *implementation = proactor->create_asynch_read_file ();
  if (implementation == 0)
    return -1;
  
  // Set the implementation class
  this->implementation (implementation);
  
  // Call the <open> method of the base class.
  return TRB_Asynch_Operation::open (handler,
                                     handle,
                                     completion_key,
                                     proactor);
}

int
TRB_Asynch_Read_File::read (ACE_Message_Block &message_block,
                            size_t bytes_to_read,
                            u_long offset,
                            u_long offset_high,
                            const void *act,
                            int priority,
                            int signal_number)
{
  return this->implementation ()->read (message_block,
                                        bytes_to_read,
                                        offset,
                                        offset_high,
                                        act,
                                        priority,
                                        signal_number);
}

#if (defined (ACE_WIN32) && !defined (ACE_HAS_WINCE))
int
TRB_Asynch_Read_File::readv (ACE_Message_Block &message_block,
                             size_t bytes_to_read,
                             u_long offset,
                             u_long offset_high,
                             const void *act,
                             int priority,
                             int signal_number)
{
  return this->implementation ()->readv (message_block,
                                         bytes_to_read,
                                         offset,
                                         offset_high,
                                         act,
                                         priority,
                                         signal_number);
}
#endif /* (defined (ACE_WIN32) && !defined (ACE_HAS_WINCE)) */

TRB_Asynch_Read_File_Impl *
TRB_Asynch_Read_File::implementation (void) const
{
  return this->implementation_;
}

void
TRB_Asynch_Read_File::implementation (TRB_Asynch_Read_File_Impl *implementation)
{
  this->implementation_ = implementation;

  // Set also the base class' implementation.
  TRB_Asynch_Read_Stream::implementation (implementation);
}

// ************************************************************

TRB_Asynch_Read_File::Result::Result (TRB_Asynch_Read_File_Result_Impl *implementation)
  : TRB_Asynch_Read_Stream::Result (implementation),
    implementation_ (implementation)
{
}
  
TRB_Asynch_Read_File::Result::~Result (void)
{
  // Proactor will delete the implementation when <complete> call
  // completes.
}

TRB_Asynch_Read_File_Result_Impl *
TRB_Asynch_Read_File::Result::implementation (void) const
{
  return this->implementation_;
}

// ************************************************************

TRB_Asynch_Write_File::TRB_Asynch_Write_File (void)
  : implementation_ (0)
{
}

TRB_Asynch_Write_File::~TRB_Asynch_Write_File (void)
{
}

int
TRB_Asynch_Write_File::open (TRB_Handler &handler,
                             ACE_HANDLE handle,
                             const void *completion_key,
                             TRB_Proactor *proactor)
{
  // Get a proactor for/from the user.
  proactor = this->get_proactor (proactor, handler);

  // Delete the old implementation.
  delete this->implementation_;
  this->implementation_ = 0;

  // Now let us get the implementation initialized.
  TRB_Asynch_Write_File_Impl *implementation = proactor->create_asynch_write_file ();
  if (implementation == 0)
    return -1;
  
  // Set the implementation class
  this->implementation (implementation);
  
  // Call the <open> method of the base class.
  return TRB_Asynch_Operation::open (handler,
                                     handle,
                                     completion_key,
                                     proactor);
}

int
TRB_Asynch_Write_File::write (ACE_Message_Block &message_block,
                              size_t bytes_to_write,
                              u_long offset,
                              u_long offset_high,
                              const void *act,
                              int priority,
                              int signal_number)
{
  return this->implementation ()->write (message_block,
                                         bytes_to_write,
                                         offset,
                                         offset_high,
                                         act,
                                         priority,
                                         signal_number);
}

#if (defined (ACE_WIN32) && !defined (ACE_HAS_WINCE))
int
TRB_Asynch_Write_File::writev (ACE_Message_Block &message_block,
                               size_t bytes_to_write,
                               u_long offset,
                               u_long offset_high,
                               const void *act,
                               int priority,
                               int signal_number)
{
  return this->implementation ()->writev (message_block,
                                          bytes_to_write,
                                          offset,
                                          offset_high,
                                          act,
                                          priority,
                                          signal_number);
}
#endif /* (defined (ACE_WIN32) && !defined (ACE_HAS_WINCE)) */

TRB_Asynch_Write_File_Impl *
TRB_Asynch_Write_File::implementation (void) const
{
  return this->implementation_;
}

void
TRB_Asynch_Write_File::implementation (TRB_Asynch_Write_File_Impl *implementation)
{
  this->implementation_ = implementation;
  
  // Init the base class' implementation also.
  TRB_Asynch_Write_Stream::implementation (implementation);
}

// ************************************************************

TRB_Asynch_Write_File::Result::Result (TRB_Asynch_Write_File_Result_Impl *implementation)
  : TRB_Asynch_Write_Stream::Result (implementation),
    implementation_ (implementation)
{
}

TRB_Asynch_Write_File::Result::~Result (void)
{
  // Proactor will delete the implementation when the <complete> call
  // completes. 
}

TRB_Asynch_Write_File_Result_Impl *
TRB_Asynch_Write_File::Result::implementation (void) const
{
  return this->implementation_;
}

// *********************************************************************

TRB_Asynch_Accept::TRB_Asynch_Accept (void)
  : implementation_ (0)
{
}

TRB_Asynch_Accept::~TRB_Asynch_Accept (void)
{
}

int
TRB_Asynch_Accept::open (TRB_Handler &handler,
                         ACE_HANDLE handle,
                         const void *completion_key,
                         TRB_Proactor *proactor)
{
  // Get a proactor for/from the user.
  proactor = this->get_proactor (proactor, handler);

  // Delete the old implementation.
  delete this->implementation_;
  this->implementation_ = 0;

  // Now let us get the implementation initialized.
  TRB_Asynch_Accept_Impl *implementation = proactor->create_asynch_accept ();
  if (implementation == 0)
    return -1;
  
  // Set the implementation class
  this->implementation (implementation);
  
  // Call the <open> method of the base class.
  return TRB_Asynch_Operation::open (handler,
                                     handle,
                                     completion_key,
                                     proactor);
}

int
TRB_Asynch_Accept::accept (ACE_Message_Block &message_block,
                           size_t bytes_to_read,
                           ACE_HANDLE accept_handle,
                           const void *act,
                           int priority,
                           int signal_number)
{
  return this->implementation ()->accept (message_block,
                                          bytes_to_read,
                                          accept_handle,
                                          act,
                                          priority,
                                          signal_number);
}

TRB_Asynch_Accept_Impl *
TRB_Asynch_Accept::implementation (void)  const
{
  return this->implementation_;
}

void
TRB_Asynch_Accept::implementation (TRB_Asynch_Accept_Impl *implementation)
{
  this->implementation_ = implementation;
  // Set the implementation in the base class.
  TRB_Asynch_Operation::implementation (implementation);
}

// ************************************************************

size_t
TRB_Asynch_Accept::Result::bytes_to_read (void) const
{
  return this->implementation ()->bytes_to_read ();
}

ACE_Message_Block &
TRB_Asynch_Accept::Result::message_block (void) const
{
  return this->implementation ()->message_block ();
}

ACE_HANDLE
TRB_Asynch_Accept::Result::listen_handle (void) const
{
  return this->implementation ()->listen_handle ();
}

ACE_HANDLE
TRB_Asynch_Accept::Result::accept_handle (void) const
{
  return this->implementation ()->accept_handle ();
}

TRB_Asynch_Accept::Result::Result (TRB_Asynch_Accept_Result_Impl *implementation)
  : TRB_Asynch_Result (implementation),
    implementation_ (implementation)
{
}

TRB_Asynch_Accept::Result::~Result (void)
{
  // Proactor will delete the implementation when the <complete> call
  // completes.
}

TRB_Asynch_Accept_Result_Impl *
TRB_Asynch_Accept::Result::implementation (void) const
{
  return this->implementation_;
}



// *********************************************************************

TRB_Asynch_Connect::TRB_Asynch_Connect (void)
  : implementation_ (0)
{
}

TRB_Asynch_Connect::~TRB_Asynch_Connect (void)
{
}

int
TRB_Asynch_Connect::open (TRB_Handler &handler,
                          ACE_HANDLE handle,
                          const void *completion_key,
                          TRB_Proactor *proactor)
{
  // Get a proactor for/from the user.
  proactor = this->get_proactor (proactor, handler);

  // Delete the old implementation.
  delete this->implementation_;
  this->implementation_ = 0;

  // Now let us get the implementation initialized.
  TRB_Asynch_Connect_Impl *implementation = proactor->create_asynch_connect ();
  if (implementation == 0)
    return -1;
  
  // Set the implementation class
  this->implementation (implementation);
  
  // Call the <open> method of the base class.
  return TRB_Asynch_Operation::open (handler,
                                     handle,
                                     completion_key,
                                     proactor);
}

int
TRB_Asynch_Connect::connect (ACE_HANDLE connect_handle,
                             const ACE_Addr & remote_sap,
                             const ACE_Addr & local_sap,
                             int  reuse_addr,
                             const void *act,
                             int priority,
                             int signal_number)
{
  return this->implementation ()->connect (connect_handle,
                                           remote_sap,
                                           local_sap,
                                           reuse_addr,
                                           act,
                                           priority,
                                           signal_number);
}

TRB_Asynch_Connect_Impl *
TRB_Asynch_Connect::implementation (void)  const
{
  return this->implementation_;
}

void
TRB_Asynch_Connect::implementation (TRB_Asynch_Connect_Impl *implementation)
{
  this->implementation_ = implementation;
  // Set the implementation in the base class.
  TRB_Asynch_Operation::implementation (implementation);
}

// ************************************************************

TRB_Asynch_Connect::Result::Result (TRB_Asynch_Connect_Result_Impl *implementation)
  : TRB_Asynch_Result (implementation),
    implementation_ (implementation)
{
}

TRB_Asynch_Connect::Result::~Result (void)
{
  // Proactor will delete the implementation when the <complete> call
  // completes.
}

const ACE_Addr& 
TRB_Asynch_Connect::Result::remote_address () const
{
  return this->implementation ()->remote_address ();
}

ACE_HANDLE
TRB_Asynch_Connect::Result::connect_handle (void) const
{
  return this->implementation ()->connect_handle ();
}


TRB_Asynch_Connect_Result_Impl *
TRB_Asynch_Connect::Result::implementation (void) const
{
  return this->implementation_;
}

// ************************************************************

TRB_Asynch_Transmit_File::TRB_Asynch_Transmit_File (void)
  : implementation_ (0)
{
}

TRB_Asynch_Transmit_File::~TRB_Asynch_Transmit_File (void)
{
}

int
TRB_Asynch_Transmit_File::open (TRB_Handler &handler,
                                ACE_HANDLE handle,
                                const void *completion_key,
                                TRB_Proactor *proactor)
{
  // Get a proactor for/from the user.
  proactor = this->get_proactor (proactor, handler);

  // Delete the old implementation.
  delete this->implementation_;
  this->implementation_ = 0;

  // Now let us get the implementation initialized.
  TRB_Asynch_Transmit_File_Impl *implementation = proactor->create_asynch_transmit_file ();
  if (implementation == 0)
    return -1;
  
  // Set the implementation class
  this->implementation (implementation);
  
  // Call the <open> method of the base class.
  return TRB_Asynch_Operation::open (handler,
                                     handle,
                                     completion_key,
                                     proactor);
}

int
TRB_Asynch_Transmit_File::transmit_file (ACE_HANDLE file,
                                         Header_And_Trailer *header_and_trailer,
                                         size_t bytes_to_write,
                                         u_long offset,
                                         u_long offset_high,
                                         size_t bytes_per_send,
                                         u_long flags,
                                         const void *act,
                                         int priority,
                                         int signal_number)
{
  return this->implementation ()->transmit_file (file,
                                                 header_and_trailer,
                                                 bytes_to_write,
                                                 offset,
                                                 offset_high,
                                                 bytes_per_send,
                                                 flags,
                                                 act,
                                                 priority,
                                                 signal_number);
}

TRB_Asynch_Transmit_File_Impl *
TRB_Asynch_Transmit_File::implementation (void) const
{
  return this->implementation_;
}

void
TRB_Asynch_Transmit_File::implementation (TRB_Asynch_Transmit_File_Impl *implementation)
{
  this->implementation_ = implementation;
  // Set the base class' implementation also.
  TRB_Asynch_Operation::implementation (this->implementation_);
}

// ***********************************************************************************

ACE_HANDLE
TRB_Asynch_Transmit_File::Result::socket (void) const
{
  return this->implementation ()->socket ();
}

ACE_HANDLE
TRB_Asynch_Transmit_File::Result::file (void) const
{
  return this->implementation ()->file ();
}

TRB_Asynch_Transmit_File::Header_And_Trailer *
TRB_Asynch_Transmit_File::Result::header_and_trailer (void) const
{
  return this->implementation ()->header_and_trailer ();
}

size_t
TRB_Asynch_Transmit_File::Result::bytes_to_write (void) const
{
  return this->implementation ()->bytes_to_write ();
}

size_t
TRB_Asynch_Transmit_File::Result::bytes_per_send (void) const
{
  return this->implementation ()->bytes_per_send ();
}

u_long
TRB_Asynch_Transmit_File::Result::flags (void) const
{
  return this->implementation ()->flags ();
}

TRB_Asynch_Transmit_File::Result::Result (TRB_Asynch_Transmit_File_Result_Impl *implementation)
  : TRB_Asynch_Result (implementation),
    implementation_ (implementation)
{
}

TRB_Asynch_Transmit_File::Result::~Result (void)
{
}

TRB_Asynch_Transmit_File_Result_Impl *
TRB_Asynch_Transmit_File::Result::implementation (void) const
{
  return this->implementation_;
}

// ************************************************************

TRB_Asynch_Transmit_File::Header_And_Trailer::Header_And_Trailer (ACE_Message_Block *header,
                                                                  size_t header_bytes,
                                                                  ACE_Message_Block *trailer,
                                                                  size_t trailer_bytes)
  : header_ (header),
    header_bytes_ (header_bytes),
    trailer_ (trailer),
    trailer_bytes_ (trailer_bytes)
{
}

TRB_Asynch_Transmit_File::Header_And_Trailer::~Header_And_Trailer (void)
{
}

void
TRB_Asynch_Transmit_File::Header_And_Trailer::header_and_trailer (ACE_Message_Block *header,
                                                                  size_t header_bytes,
                                                                  ACE_Message_Block *trailer,
                                                                  size_t trailer_bytes)
{
  this->header (header);
  this->header_bytes (header_bytes);
  this->trailer (trailer);
  this->trailer_bytes (trailer_bytes);
}

ACE_Message_Block *
TRB_Asynch_Transmit_File::Header_And_Trailer::header (void) const
{
  return this->header_;
}

void
TRB_Asynch_Transmit_File::Header_And_Trailer::header (ACE_Message_Block *message_block)
{
  this->header_ = message_block;
}

size_t
TRB_Asynch_Transmit_File::Header_And_Trailer::header_bytes (void) const
{
  return this->header_bytes_;
}

void
TRB_Asynch_Transmit_File::Header_And_Trailer::header_bytes (size_t bytes)
{
  this->header_bytes_ = bytes;
}

ACE_Message_Block *
TRB_Asynch_Transmit_File::Header_And_Trailer::trailer (void) const
{
  return this->trailer_;
}

void
TRB_Asynch_Transmit_File::Header_And_Trailer::trailer (ACE_Message_Block *message_block)
{
  this->trailer_ = message_block;
}

size_t
TRB_Asynch_Transmit_File::Header_And_Trailer::trailer_bytes (void) const
{
  return this->trailer_bytes_;
}

void
TRB_Asynch_Transmit_File::Header_And_Trailer::trailer_bytes (size_t bytes)
{
  this->trailer_bytes_ = bytes;
}

ACE_LPTRANSMIT_FILE_BUFFERS
TRB_Asynch_Transmit_File::Header_And_Trailer::transmit_buffers (void)
{
  // If both are zero, return zero
  if (this->header_ == 0 && this->trailer_ == 0)
    return 0;
  else
    {
      // Something is valid

      // If header is valid, set the fields
      if (this->header_ != 0)
        {
          this->transmit_buffers_.Head = this->header_->rd_ptr ();
#if defined(ACE_WIN64)
          this->transmit_buffers_.HeadLength =
            ACE_static_cast (DWORD, this->header_bytes_);
#else
          this->transmit_buffers_.HeadLength = this->header_bytes_;
#endif /* ACE_WIN64 */
        }
      else
        {
          this->transmit_buffers_.Head = 0;
          this->transmit_buffers_.HeadLength = 0;
        }

      // If trailer is valid, set the fields
      if (this->trailer_ != 0)
        {
          this->transmit_buffers_.Tail = this->trailer_->rd_ptr ();
#if defined(ACE_WIN64)
          this->transmit_buffers_.TailLength =
            ACE_static_cast (DWORD, this->trailer_bytes_);
#else
          this->transmit_buffers_.TailLength = this->trailer_bytes_;
#endif /* ACE_WIN64 */
        }
      else
        {
          this->transmit_buffers_.Tail = 0;
          this->transmit_buffers_.TailLength = 0;
        }

      // Return the transmit buffers
      return &this->transmit_buffers_;
    }
}

// *********************************************************************

TRB_Handler::TRB_Handler (void)
  : proactor_ (0), handle_ (ACE_INVALID_HANDLE)
{
}

TRB_Handler::TRB_Handler (TRB_Proactor *d)
  : proactor_ (d), handle_ (ACE_INVALID_HANDLE)
{
}

TRB_Handler::~TRB_Handler (void)
{
}

void
TRB_Handler::handle_read_stream (const TRB_Asynch_Read_Stream::Result & /* result */)
{
}

void
TRB_Handler::handle_write_stream (const TRB_Asynch_Write_Stream::Result & /* result */)
{
}

void
TRB_Handler::handle_write_dgram (const TRB_Asynch_Write_Dgram::Result & /* result */)
{
}

void
TRB_Handler::handle_read_dgram (const TRB_Asynch_Read_Dgram::Result & /* result */)
{
}

void
TRB_Handler::handle_accept (const TRB_Asynch_Accept::Result & /* result */)
{
}

void
TRB_Handler::handle_connect (const TRB_Asynch_Connect::Result & /* result */)
{
}

void
TRB_Handler::handle_transmit_file (const TRB_Asynch_Transmit_File::Result & /* result */)
{
}

void
TRB_Handler::handle_read_file (const TRB_Asynch_Read_File::Result & /* result */)
{
}

void
TRB_Handler::handle_write_file (const TRB_Asynch_Write_File::Result & /* result */)
{
}

void
TRB_Handler::handle_time_out (const ACE_Time_Value & /* tv */,
                              const void *           /* act */)
{
}
void
TRB_Handler::handle_message (ACE_Message_Block & /*message_block*/,
                              const void *             /* act */)
{
}

void
TRB_Handler::handle_wakeup (void)
{
}

TRB_Proactor *
TRB_Handler::proactor (void)
{
  return this->proactor_;
}

void
TRB_Handler::proactor (TRB_Proactor *p)
{
  this->proactor_ = p;
}

ACE_HANDLE
TRB_Handler::handle (void) const
{
  return this->handle_;
}

void
TRB_Handler::handle (ACE_HANDLE h)
{
  this->handle_ = h;
}

// ************************************************************

TRB_Service_Handler::TRB_Service_Handler (void)
{
}

TRB_Service_Handler::~TRB_Service_Handler (void)
{
}

void
TRB_Service_Handler::addresses (const ACE_INET_Addr & /* remote_address */,
                                const ACE_INET_Addr & /* local_address */ )
{
}

void
TRB_Service_Handler::act (const void *)
{
}

void
TRB_Service_Handler::open (ACE_HANDLE,
                           ACE_Message_Block &)
{
}


// ************************************************************

TRB_Asynch_Read_Dgram::TRB_Asynch_Read_Dgram (void)
  : implementation_ (0)
{
}

TRB_Asynch_Read_Dgram::~TRB_Asynch_Read_Dgram (void)
{
}

int
TRB_Asynch_Read_Dgram::open (TRB_Handler &handler,
                             ACE_HANDLE handle,
                             const void *completion_key,
                             TRB_Proactor *proactor)
{
  // Get a proactor for/from the user.
  proactor = this->get_proactor (proactor, handler);

  // Delete the old implementation.
  delete this->implementation_;
  this->implementation_ = 0;

  // Now let us get the implementation initialized.
  TRB_Asynch_Read_Dgram_Impl *implementation = proactor->create_asynch_read_dgram ();
  if (implementation == 0)
    return -1;
  
  // Set the implementation class
  this->implementation (implementation);

  // Call the <open> method of the base class. 
  return TRB_Asynch_Operation::open (handler,
                                     handle,
                                     completion_key,
                                     proactor);
}

ssize_t
TRB_Asynch_Read_Dgram::recv (ACE_Message_Block *message_block,
                             size_t &number_of_bytes_recvd,
                             int flags,
                             int protocol_family,
                             const void *act,
                             int priority,
                             int signal_number)
{
  return this->implementation ()->recv (message_block,
                                        number_of_bytes_recvd,
                                        flags,
                                        protocol_family,
                                        act,
                                        priority,
                                        signal_number);
}

TRB_Asynch_Read_Dgram_Impl *
TRB_Asynch_Read_Dgram::implementation (void) const
{
  return implementation_;
}

void
TRB_Asynch_Read_Dgram::implementation (TRB_Asynch_Read_Dgram_Impl *implementation)
{
  this->implementation_ = implementation;

  // Init the bas class' implementation also.
  TRB_Asynch_Operation::implementation(this->implementation_);
}

// ************************************************************

const ACE_Addr& 
TRB_Asynch_Read_Dgram::Result::remote_address () const
{
  return this->implementation ()->remote_address ();
}

int
TRB_Asynch_Read_Dgram::Result::remote_address (ACE_Addr& addr) const
{
  return this->implementation ()->remote_address (addr);
}

ACE_Message_Block*
TRB_Asynch_Read_Dgram::Result::message_block (void) const
{
  return this->implementation ()->message_block ();
}

int
TRB_Asynch_Read_Dgram::Result::flags (void) const
{
  return this->implementation ()->flags ();
}

size_t
TRB_Asynch_Read_Dgram::Result::bytes_to_read (void) const
{
  return this->implementation ()->bytes_to_read ();
}

ACE_HANDLE
TRB_Asynch_Read_Dgram::Result::handle (void) const
{
  return this->implementation ()->handle();
}

TRB_Asynch_Read_Dgram::Result::Result (TRB_Asynch_Read_Dgram_Result_Impl *implementation)
: TRB_Asynch_Result (implementation),
  implementation_ (implementation)
{
}

TRB_Asynch_Read_Dgram::Result::~Result (void)
{
}

TRB_Asynch_Read_Dgram_Result_Impl *
TRB_Asynch_Read_Dgram::Result::implementation (void) const
{
  return this->implementation_;
}

// ************************************************************


TRB_Asynch_Write_Dgram::TRB_Asynch_Write_Dgram (void)
  : implementation_ (0)
{
}

TRB_Asynch_Write_Dgram::~TRB_Asynch_Write_Dgram (void)
{
}

int
TRB_Asynch_Write_Dgram::open (TRB_Handler &handler,
                              ACE_HANDLE handle,
                              const void *completion_key,
                              TRB_Proactor *proactor)
{
  // Get a proactor for/from the user.
  proactor = this->get_proactor (proactor, handler);

  // Delete the old implementation.
  delete this->implementation_;
  this->implementation_ = 0;

  // Now let us get the implementation initialized.
  TRB_Asynch_Write_Dgram_Impl *implementation = proactor->create_asynch_write_dgram ();
  if (implementation == 0)
    return -1;
  
  // Set the implementation class
  this->implementation (implementation);
  
  // Call the <open> method of the base class.
  return TRB_Asynch_Operation::open (handler,
                                     handle,
                                     completion_key,
                                     proactor);
}

ssize_t
TRB_Asynch_Write_Dgram::send (ACE_Message_Block *message_block,
                              size_t &number_of_bytes_sent,
                              int flags,
                              const ACE_Addr& remote_addr,
                              const void *act,
                              int priority,
                              int signal_number)
{
  return this->implementation ()->send (message_block,
                                        number_of_bytes_sent,
                                        flags,
                                        remote_addr,
                                        act,
                                        priority,
                                        signal_number);
}

TRB_Asynch_Write_Dgram_Impl *
TRB_Asynch_Write_Dgram::implementation (void) const
{
  return this->implementation_;
}

void
TRB_Asynch_Write_Dgram::implementation (TRB_Asynch_Write_Dgram_Impl *implementation)
{
  this->implementation_ = implementation;

  // Init the base class' implementation also.
  TRB_Asynch_Operation::implementation (this->implementation_);
}

// ************************************************************

const ACE_Addr& 
TRB_Asynch_Write_Dgram::Result::remote_address () const
{
  return this->implementation ()->remote_address ();
}

size_t
TRB_Asynch_Write_Dgram::Result::bytes_to_write (void) const
{
  return this->implementation ()->bytes_to_write ();
}

ACE_Message_Block*
TRB_Asynch_Write_Dgram::Result::message_block () const
{
  return this->implementation ()->message_block ();
}

int
TRB_Asynch_Write_Dgram::Result::flags (void) const
{
  return this->implementation ()->flags ();
}

ACE_HANDLE
TRB_Asynch_Write_Dgram::Result::handle (void) const
{
  return this->implementation ()->handle ();
}

TRB_Asynch_Write_Dgram_Result_Impl *
TRB_Asynch_Write_Dgram::Result::implementation (void) const
{
  return this->implementation_;
}

TRB_Asynch_Write_Dgram::Result::Result (TRB_Asynch_Write_Dgram_Result_Impl *implementation)
: TRB_Asynch_Result (implementation),
  implementation_ (implementation)
{
}

TRB_Asynch_Write_Dgram::Result::~Result (void)
{
}

#endif /* ACE_WIN32 || ACE_HAS_AIO_CALLS */
