/*********************************************************************
** Copyright (C) 2003 Terabit Pty Ltd.  All rights reserved.
**
** This file is part of the POSIX-Proactor module.
**
** This file may be distributed and used only under the terms of the  
** Terabit Public License as defined by Terabit Pty Ltd of Australia   
** and appearing in the file tlicense.txt included in the packaging of
** this module and available at http://www.terabit.com.au/license.php.
**
** Contact support@terabit.com.au if any conditions of this licensing
** are not clear to you.
**********************************************************************/

/* -*- C++ -*- */

//=============================================================================
/**
 *
 *  @file    POSIX_AIO_Dedicated_Processor.h
 *
 *  $Id: $
 *
 *  @author Alexander Libman <libman@terabit.com.au>
 */
//=============================================================================

#ifndef TRB_POSIX_AIO_DEDICATED_PROCESSOR_H 
#define TRB_POSIX_AIO_DEDICATED_PROCESSOR_H 

#include "TProactor/POSIX_AIO_Processor.h"

#if !defined (ACE_LACKS_PRAGMA_ONCE)
#pragma once
#endif /* ACE_LACKS_PRAGMA_ONCE */

#if defined (ACE_HAS_AIO_CALLS) || defined(ACE_HAS_AIO_EMULATION)

#include "ace/Task.h"


/**
 * @class TRB_POSIX_AIO_Dedicated_Processor
 *
 * @brief POSIX implementation .
 */
class TPROACTOR_Export TRB_POSIX_AIO_Dedicated_Processor : 
  public TRB_POSIX_AIO_Processor,
  public ACE_Task<ACE_SYNCH>
{
public:
  
  
  /// Constructor.
  TRB_POSIX_AIO_Dedicated_Processor 
                 (TRB_POSIX_AIO_Dispatcher & dispatcher,
                  const TRB_POSIX_AIO_Config & config);
  /// Virtual destructor.
  virtual ~TRB_POSIX_AIO_Dedicated_Processor (void);
  

  // virtual from TRB_POSIX_AIO_Proccessor
  virtual int start (void);
  
  virtual int stop (void);

  virtual int is_dedicated_leadership (void) const;

  virtual int is_ready_for_leadership (void) const;

  // virtual from Task
  virtual int svc (void);

protected:

  // virtual from TRB_POSIX_AIO_Proccessor
  virtual int activate_leader (int reason);

  enum 
    {
      SS_INIT          = 0,
      SS_ACTIVATING    = 1,
      SS_WAIT_COND_VAR = 2,
      SS_HANDLE_EVENTS = 3,
      SS_STOPPING      = 4,
      SS_STOPPED       = 5
    };

  int svc_state_ ;

  int flg_stop_;


  ///  Data members
  ACE_SYNCH_CONDITION non_empty_condition_;
  ACE_SYNCH_CONDITION ready_condition_;
};

#endif /* ACE_HAS_AIO_CALLS */

#endif //TRB_POSIX_AIO_DEDICATED_PROCESSOR_H 
