/*********************************************************************
** Copyright (C) 2003 Terabit Pty Ltd.  All rights reserved.
**
** This file is part of the POSIX-Proactor module.
**
** This file may be distributed and used only under the terms of the  
** Terabit Public License as defined by Terabit Pty Ltd of Australia   
** and appearing in the file tlicense.txt included in the packaging of
** this module and available at http://www.terabit.com.au/license.php.
**
** Contact support@terabit.com.au if any conditions of this licensing
** are not clear to you.
**********************************************************************/

/* -*- C++ -*- */

//=============================================================================
/**
 *  @file    TRB_POSIX_AIO_Event_Poll_Strategy.h
 *
 *  $Id:  $
 *
 *  @author Alexander Libman <libman@terabit.com.au>
 */
//=============================================================================

#ifndef TRB_POSIX_AIO_EVENT_POLL_STRATEGY_H 
#define TRB_POSIX_AIO_EVENT_POLL_STRATEGY_H 

#include "TProactor/POSIX_AIO_Select_Provider.h"

#if defined (ACE_HAS_LINUX_EPOLL) 


#if defined (ACE_HAS_AIO_CALLS) || defined(ACE_HAS_AIO_EMULATION)

#include /**/ <sys/epoll.h>


/**
 * @class TRB_POSIX_AIO_Event_Poll_Strategy
 *
 */

class TPROACTOR_Export TRB_POSIX_AIO_Event_Poll_Strategy :
  public TRB_POSIX_AIO_Select_Strategy_Base
{
protected:

  friend class TRB_POSIX_AIO_Select_Provider;
  
  TRB_POSIX_AIO_Event_Poll_Strategy (TRB_POSIX_AIO_Select_Provider & provider);
  virtual ~TRB_POSIX_AIO_Event_Poll_Strategy();


  // Wait Strategy Interface

  // use base class
  // virtual TRB_POSIX_AIO_Interrupt_Strategy * create_interrupt_strategy (void);
  
  // use base class
  // virtual void 
  //  destroy_interrupt_strategy (TRB_POSIX_AIO_Interrupt_Strategy * i_strategy);
  virtual int init (void);
  virtual int fini (void);

  virtual int  on_aio_start (TRB_POSIX_Asynch_Result * result);
  
  virtual int  prepare_to_wait (TRB_POSIX_Asynch_Waiting_List & started_list);

  virtual int  wait (ACE_Time_Value * wait_time);

  virtual TRB_POSIX_Asynch_Result * 
    get_next_completed_result (TRB_POSIX_Asynch_Waiting_List & started_list);


  virtual int declare_interest (ACE_HANDLE handle, u_int mask);
  virtual int clear_interest   (ACE_HANDLE handle, u_int mask);

private:


  /**
   * All interactions with epoll are
   * done through this file descriptor.
   */
  ACE_HANDLE poll_fd_;

  /// The pollfd array that epoll will feed its results to.
  struct epoll_event * events_;

  /// Pointer to the next pollfd array element that contains the next
  /// event to be dispatched.
  struct epoll_event * next_event_;
  int num_events_;

};

#endif /* ACE_HAS_AIO_CALLS */
#endif /* ACE_HAS_LINUX_EPOLL */
#endif //TRB_POSIX_AIO_EVENT_POLL_STRATEGY_H 
