/*********************************************************************
** Copyright (C) 2003 Terabit Pty Ltd.  All rights reserved.
**
** This file is part of the POSIX-Proactor module.
**
** This file may be distributed and used only under the terms of the  
** Terabit Public License as defined by Terabit Pty Ltd of Australia   
** and appearing in the file tlicense.txt included in the packaging of
** this module and available at http://www.terabit.com.au/license.php.
**
** Contact support@terabit.com.au if any conditions of this licensing
** are not clear to you.
**********************************************************************/

/* -*- C++ -*- */
// $Id:  $

#include "TProactor/POSIX_AIO_Linux_RT_Strategy.h"

#if (defined(ACE_HAS_AIO_CALLS) || defined(ACE_HAS_AIO_EMULATION)) && defined(linux)

#include "ace/Log_Msg.h"
#include "ace/Addr.h"
#include "ace/ACE.h"

/**
 *****************************************************************
 *  POSIX Select-based AIO Wait Strategy
 *
 *****************************************************************
 */
TRB_POSIX_AIO_Linux_RT_Strategy::TRB_POSIX_AIO_Linux_RT_Strategy
  (TRB_POSIX_AIO_Select_Provider & provider)
  : TRB_POSIX_AIO_Select_Strategy_Base (provider),
    sig_num_ (0),
    flg_sigio_(0),
    pid_ (0),
    tid_ (ACE_OS::NULL_thread)
{
  this->sig_num_ = this->config().wait_sig_num();
}

TRB_POSIX_AIO_Linux_RT_Strategy::~TRB_POSIX_AIO_Linux_RT_Strategy (void)
{
  this->fini ();
}

int 
TRB_POSIX_AIO_Linux_RT_Strategy::init (void)
{
  this->pid_ = ACE_OS::getpid ();
  this->tid_ = ACE_Thread::self ();

  if (this->alloc_repository (ACE_AIO_NO_LIMIT, 1) != 0)
    return -1;;

  ACE_DEBUG ((LM_DEBUG,
    ACE_LIB_TEXT("(%t)TRB_POSIX_AIO_Linux_RT_Strategy pid=%d size=%d\n"), 
    (int) this->pid_,
    (int) this->size_));


  sigemptyset (&this->signal_mask_);  
  sigaddset (&this->signal_mask_, this->sig_num_);  
  sigaddset (&this->signal_mask_, SIGIO);  

  if (this->setup_signal_handler (this->sig_num_, 0) !=0 )
    return -1;

  if (this->setup_signal_handler (SIGIO, 0) != 0)
    return -1;

  if (this->mask_signals (SIG_BLOCK, &this->signal_mask_) != 0)
    return -1;

  return 0;
}  

int 
TRB_POSIX_AIO_Linux_RT_Strategy::fini (void)
{
  this->setup_signal_handler (this->sig_num_, 1);
  this->setup_signal_handler (SIGIO, 1);

  this->free_repository ();
  
  return 0;
}

void 
TRB_POSIX_AIO_Linux_RT_Strategy::sig_handler (int sig_num,
                                           siginfo_t *,
                                           ucontext_t *)
{
  ACE_UNUSED_ARG (sig_num);
  //ACE_DEBUG ((LM_DEBUG,
  //            "%N:%l:(%P | %t)::sig_handler received signal: %d\n",
  //             sig_num));
  return;
}

int
TRB_POSIX_AIO_Linux_RT_Strategy::setup_signal_handler (int sig_num,
                                                    int flg_ign) const
{
  struct sigaction reaction;

  sigemptyset (&reaction.sa_mask);   // Nothing else to mask.


  if (flg_ign)
    {
      reaction.sa_flags = SA_NODEFER;    // Realtime flag.
      reaction.sa_handler = SIG_IGN;
    }
  else
    {
      reaction.sa_flags = SA_SIGINFO |SA_NODEFER;    // Realtime flag.
      reaction.sa_sigaction = 
        ACE_SIGNAL_C_FUNC (TRB_POSIX_AIO_Linux_RT_Strategy::sig_handler);
    }


  // (SIG_DFL);

  int sigaction_return = ACE_OS::sigaction (sig_num,
                                            &reaction,
                                            0);
  if (sigaction_return == -1)
    ACE_ERROR_RETURN ((LM_ERROR,
      ACE_LIB_TEXT("(%t) TRB_POSIX_AIO_Linux_RT_Strategy::%p\n"),
      ACE_LIB_TEXT("setup_signal_handler failed")),
      -1);
  return 0;
}

int
TRB_POSIX_AIO_Linux_RT_Strategy::mask_signals (int how,
                                            const sigset_t *signals) const
{
  // how = SIG_BLOCK/SIG_UNBLOCK/SIG_SETMASK

  if (ACE_OS::sigprocmask (how, signals, 0) != 0)
    ACE_ERROR_RETURN ((LM_ERROR,
      ACE_LIB_TEXT("(%t) TRB_POSIX_AIO_Linux_RT_Strategy::%p\n"),
      ACE_LIB_TEXT("mask_signals failed")),
       -1);
  return 0;
}

int
TRB_POSIX_AIO_Linux_RT_Strategy::on_aio_start (TRB_POSIX_Asynch_Result * result)
{
  ACE_HANDLE handle = result->aio_fildes;

  int flags = ACE::get_flags (handle);
  int sig = ACE_OS::fcntl (handle, F_GETSIG, 0);
  int pid = ACE_OS::fcntl (handle, F_GETOWN, 0);

  if ((sig == this->sig_num_) &&
      ((flags & (O_NONBLOCK|O_ASYNC)) == (O_NONBLOCK|O_ASYNC)) &&
      (pid == this->pid_))
      return 0;

  //  In Linux each thread is a process.
  //  So we  have to use dedicated model
  //  as RT-signals will be delivered only
  //  to the specified thread-process.
  // 
  //if (this->pid_ != ACE_OS::getpid ())
  //
  if (!ACE_OS::thr_equal (ACE_Thread::self(), this->tid_))
    {
      result->set_deferred();
      return 1;
    }

  flags |= (O_NONBLOCK|O_ASYNC);

  if (ACE_OS::fcntl(handle, F_SETFL, flags) < 0) 
    {
      result->set_start_error (errno);

      ACE_ERROR_RETURN ((LM_ERROR,
       ACE_LIB_TEXT("on_aio_start fcntl failed %d %p\n"),
       errno,
       ACE_LIB_TEXT("F_SETFL")),
       -1);
    }


  if (ACE_OS::fcntl(handle, F_SETOWN, this->pid_) < 0) 
    {
      result->set_start_error (errno);

      ACE_ERROR_RETURN ((LM_ERROR,
        ACE_LIB_TEXT("on_aio_start fcntl failed %d %p\n"),
        errno,
        ACE_LIB_TEXT("F_SETOWN")),
        -1);
    }


  if (ACE_OS::fcntl(handle, F_SETSIG, this->sig_num_) < 0) 
    {
      result->set_start_error (errno);

      ACE_ERROR_RETURN ((LM_ERROR,
        ACE_LIB_TEXT("on_aio_start fcntl failed %d %p\n"),
        errno,
        ACE_LIB_TEXT("F_SETSIG")),
        -1);
    }

  if (result->op_code () == TRB_Asynch_Result_Impl::CONNECT)
    {
      if (this->pending_writes_[handle] == IO_READY)
          this->pending_writes_[handle] = 0;

      if (this->pending_reads_ [handle] == IO_READY)
          this->pending_reads_ [handle] = 0;
    }

  return 0; 
}


int
TRB_POSIX_AIO_Linux_RT_Strategy::prepare_to_wait (TRB_POSIX_Asynch_Waiting_List & started_list)
{
  ACE_TRACE ("TRB_POSIX_AIO_Linux_RT_Strategy::prepare_to_wait");
  ACE_UNUSED_ARG (started_list);

  if (this->fd_info_.handle () != ACE_INVALID_HANDLE)
    return WAIT_COMPLETED; 


  return WAIT_ENABLED;
}

int
TRB_POSIX_AIO_Linux_RT_Strategy::wait (ACE_Time_Value * wait_time)
{
  ACE_TRACE ("TRB_POSIX_AIO_Linux_RT_Strategy::wait");

  timespec timeout ;
  timespec * ptime = 0;
  
  if (wait_time != 0) 
    {
      timeout = wait_time->operator timespec();
      ptime = &timeout;
    }
  
  return this->wait_i (ptime);
}

int
TRB_POSIX_AIO_Linux_RT_Strategy::wait_i (timespec * wait_time)
{
  int rc = 0;

  siginfo_t sig_info;

  this->flg_sigio_ = 0;
  this->fd_info_.set (ACE_INVALID_HANDLE, 0);
  
  // Wait for the signals.
  if (wait_time == 0)
    {
      rc = ::sigwaitinfo (&this->signal_mask_,
                          &sig_info);
    }
  else
    {
      rc= ::sigtimedwait (&this->signal_mask_,
                          &sig_info,
                          wait_time);
    }


  if (rc == this->sig_num_)
    {
      this->fd_info_.set (sig_info.si_fd, sig_info.si_band);

      return WAIT_COMPLETED; 
    }

  if (rc == SIGIO) 
    {
      ACE_DEBUG ((LM_DEBUG, "sigwait SIGIO!!!\n"));

      this->flg_sigio_ = 1;
      this->fd_info_.set (0, (POLLIN|POLLOUT));
    
      return WAIT_COMPLETED; 
    }

  switch (errno)
    {
    case EAGAIN:
      return WAIT_TIMEOUT;

    case EINTR :    // Interrupted call
      return WAIT_INTERRUPTED;

    default:
      break;
    }
  return WAIT_ERROR;
}

TRB_POSIX_Asynch_Result *
TRB_POSIX_AIO_Linux_RT_Strategy::get_next_completed_result 
  (TRB_POSIX_Asynch_Waiting_List & started_list)
{
  ACE_UNUSED_ARG (started_list);

  for (;;)
    {

      ACE_HANDLE handle = this->fd_info_.handle();

      if (handle == ACE_INVALID_HANDLE)
        return 0;

      int rc = this->check (this->fd_info_);

      if (rc)
        {
          TRB_POSIX_Asynch_Result * result = this->fd_info_.get_result();
          ACE_ASSERT (result !=0);
          return result;
        }

      if (this->flg_sigio_ == 0 )
        {
          timespec timeout ;

          timeout.tv_sec = 0;
          timeout.tv_nsec = 0;

          this->wait_i (&timeout);
          continue;
        }

      ++handle;

      if (handle >= ACE_static_cast(int, this->size_))
        break;

      this->fd_info_.set(handle, (POLLIN | POLLOUT));
    }


  this->fd_info_.set(ACE_INVALID_HANDLE, 0);
  
  this->setup_signal_handler (this->sig_num_, 1);
  this->setup_signal_handler (this->sig_num_, 0);

  this->flg_sigio_ = 0;

  return 0;
}


// *****************************************************************
//  Declare interest
//  Return code :
//    1 - we should interrupt leader to update wait state
//    0 - we do not have to interrupt leader
// *****************************************************************

int
TRB_POSIX_AIO_Linux_RT_Strategy::declare_interest (ACE_HANDLE handle, 
                                                   u_int mask)
{  
  ACE_UNUSED_ARG (handle);
  ACE_UNUSED_ARG (mask);

  return 0;
}

int
TRB_POSIX_AIO_Linux_RT_Strategy::clear_interest (ACE_HANDLE handle, 
                                                 u_int mask)
{  
  ACE_UNUSED_ARG (handle);
  ACE_UNUSED_ARG (mask);

  return 0;
}


#endif /* ACE_HAS_AIO_CALLS  && linux */
