/*********************************************************************
** Copyright (C) 2003 Terabit Pty Ltd.  All rights reserved.
**
** This file is part of the POSIX-Proactor module.
**
** This file may be distributed and used only under the terms of the  
** Terabit Public License as defined by Terabit Pty Ltd of Australia   
** and appearing in the file tlicense.txt included in the packaging of
** this module and available at http://www.terabit.com.au/license.php.
**
** Contact support@terabit.com.au if any conditions of this licensing
** are not clear to you.
**********************************************************************/

/* -*- C++ -*- */
// $Id:  $

#include "TProactor/POSIX_AIO_SGI_Strategy.h"

#if (defined (ACE_HAS_AIO_CALLS) || defined(ACE_HAS_AIO_EMULATION)) && !defined(sun) && !defined(__Lynx__)

#include "ace/Log_Msg.h"
#include "TProactor/POSIX_AIO_STD_Provider.h"

/*
 ***************************************************************
 * Semaphore_Strategy
 ***************************************************************
 */

TRB_POSIX_AIO_SGI_Strategy::TRB_POSIX_AIO_SGI_Strategy 
  (TRB_POSIX_AIO_STD_Provider & provider)
  : TRB_POSIX_AIO_Wait_Strategy (provider),
    TRB_POSIX_AIO_Interrupt_Strategy (provider),
    std_provider_ (provider),
    sema_ (0)
{
#if defined(__sgi)

   ACE_DEBUG((LM_DEBUG,
              ACE_LIB_TEXT( "SGI IRIX specific: aio_init!\n")));

//typedef struct aioinit {
//    int aio_threads;  /* The number of aio threads to start (5) */
//    int aio_locks;    /* Initial number of preallocated locks (3) */
//    int aio_num;      /* estimated total simultanious aiobc structs (1000) */
//    int aio_usedba;   /* Try to use DBA for raw I/O in lio_listio (0) */
//    int aio_debug;    /* turn on debugging (0) */
//    int aio_numusers; /* max number of user sprocs making aio_* calls (5) */
//    int aio_reserved[3];
//} aioinit_t;

  aioinit_t  aioinit;

  aioinit.aio_threads = 10; /* The number of aio threads to start (5) */
  aioinit.aio_locks = 20;   /* Initial number of preallocated locks (3) */
                     /* estimated total simultaneous aiobc structs (1000) */
  aioinit.aio_num = aiocb_list_max_size_;
  aioinit.aio_usedba = 0;   /* Try to use DBA for raw IO in lio_listio (0) */
  aioinit.aio_debug = 0;    /* turn on debugging (0) */
  aioinit.aio_numusers = 100; /* max number of user sprocs making aio_* calls (5) */
  aioinit.aio_reserved[0] = 0;
  aioinit.aio_reserved[1] = 0;
  aioinit.aio_reserved[2] = 0;

  aio_sgi_init (&aioinit);

#endif
}

TRB_POSIX_AIO_SGI_Strategy::~TRB_POSIX_AIO_SGI_Strategy(void)
{
  ;
}

// *****************************************************************
//   Wait strategy
// *****************************************************************

TRB_POSIX_AIO_Interrupt_Strategy *
TRB_POSIX_AIO_SGI_Strategy::create_interrupt_strategy (void)
{
  return this;
}

void 
TRB_POSIX_AIO_SGI_Strategy::destroy_interrupt_strategy (TRB_POSIX_AIO_Interrupt_Strategy * i_strategy)
{
  //does nothing
  ACE_ASSERT (this == i_strategy);
}

void 
TRB_POSIX_AIO_SGI_Strategy::aio_completion_func (sigval_t cb_data)
{
  // Currently SGI IRIX supports this way of notification 
  // May be Linux will support in the future.
  // This function is called on AIO completion
  // from it system AIO thread.
  
#if defined (__FreeBSD__)
  TRB_POSIX_AIO_SGI_Strategy * strategy =
    ACE_static_cast (TRB_POSIX_AIO_SGI_Strategy *, cb_data.sigval_ptr);
#else
  TRB_POSIX_AIO_SGI_Strategy * strategy =
    ACE_static_cast (TRB_POSIX_AIO_SGI_Strategy *, cb_data.sival_ptr);
#endif

  if (strategy != 0 )
    strategy->interrupt (ACE_OS::NULL_thread);
}

int
TRB_POSIX_AIO_SGI_Strategy::on_aio_start (TRB_POSIX_Asynch_Result * result)
{
  // Currently SGI IRIX supports this way of notification 
  // May be Linux will support in the future.

#if defined(__sgi)
  result->aio_sigevent.sigev_notify = SIGEV_CALLBACK;
  result->aio_sigevent.sigev_func   = aio_completion_func ;
#else
  result->aio_sigevent.sigev_notify = SIGEV_SIGNAL;
#endif /* __sgi */

#if defined (__FreeBSD__)
  result->aio_sigevent.sigev_value.sigval_ptr = this ;
#else
  result->aio_sigevent.sigev_value.sival_ptr = this ;
#endif /* __FreeBSD__ */

  return 0;
}

int
TRB_POSIX_AIO_SGI_Strategy::prepare_to_wait (TRB_POSIX_Asynch_Waiting_List & started_list)
{
  ACE_UNUSED_ARG (started_list);

  // ready to wait
  return WAIT_TIMEOUT;
}

int
TRB_POSIX_AIO_SGI_Strategy::wait (ACE_Time_Value * wait_time)
{
  ACE_TRACE ("TRB_POSIX_AIO_SGI_Strategy::wait");

  int ret_wait=0;

  if (wait_time == 0 ) //  wait indefinetely
    {
      ret_wait = this->sema_.acquire();
    }
  else
    {
      ACE_Time_Value abs_time = ACE_OS::gettimeofday ()
                              + *wait_time;

      ret_wait = this->sema_.acquire(abs_time);
    }

  // Check for errors
  if (ret_wait == -1)
    {
      switch (errno)
        {
        case ETIME:   // Timeout
          return WAIT_TIMEOUT;

        case EINTR :    // Interrupted call
          return WAIT_INTERRUPTED;
          
        default:
          ACE_ERROR((LM_ERROR,
            ACE_LIB_TEXT ("%N:%l:(%P | %t)::%p\n"),
            ACE_LIB_TEXT ("TRB_POSIX_AIO_SGI_Strategy::wait failed")));
          return WAIT_ERROR;

        } //switch
    }

  return WAIT_COMPLETED;
}

TRB_POSIX_Asynch_Result *
TRB_POSIX_AIO_SGI_Strategy::get_next_completed_result 
  (TRB_POSIX_Asynch_Waiting_List & started_list)
{
  return this->std_provider_.get_next_completed_result (started_list);
}

// *****************************************************************
//   Interrupt strategy
// *****************************************************************
int 
TRB_POSIX_AIO_SGI_Strategy::init (void)
{
  return 0;
}

int 
TRB_POSIX_AIO_SGI_Strategy::fini (void)
{
  return 0;
}

int 
TRB_POSIX_AIO_SGI_Strategy::on_leader_activated (TRB_POSIX_AIO_Processor & aio_processor)
{
  ACE_UNUSED_ARG (aio_processor);
  return 0;
}

int 
TRB_POSIX_AIO_SGI_Strategy::on_leader_deactivated (TRB_POSIX_AIO_Processor & aio_processor)
{
  ACE_UNUSED_ARG (aio_processor);
  return 0;
}

int 
TRB_POSIX_AIO_SGI_Strategy::interrupt (ACE_thread_t leader_id)
{
  ACE_UNUSED_ARG (leader_id);

  if (this->sema_.release() < 0)
    ACE_ERROR_RETURN((LM_ERROR,
      ACE_LIB_TEXT("(%P | %t) TRB_POSIX_AIO_SGI_Strategy::%p\n"),
      ACE_LIB_TEXT("interrupt failed")),
      -1);
  return 0;
}

 

#endif /* ACE_HAS_AIO_CALLS */
