/*********************************************************************
** Copyright (C) 2003 Terabit Pty Ltd.  All rights reserved.
**
** This file is part of the POSIX-Proactor module.
**
** This file may be distributed and used only under the terms of the  
** Terabit Public License as defined by Terabit Pty Ltd of Australia   
** and appearing in the file tlicense.txt included in the packaging of
** this module and available at http://www.terabit.com.au/license.php.
**
** Contact support@terabit.com.au if any conditions of this licensing
** are not clear to you.
**********************************************************************/

/* -*- C++ -*- */
// $Id:  $

#include "TProactor/POSIX_AIO_SUN_Port_Strategy.h"

#if defined (ACE_HAS_AIO_CALLS) || defined(ACE_HAS_AIO_EMULATION)

#if defined (ACE_HAS_SUN_COMPLETION_PORTS) && defined (sun)


#include "ace/Log_Msg.h"
#include "ace/Addr.h"
#include "ace/ACE.h"



/**
 *****************************************************************
 *  POSIX Sun Completion Ports based AIO Wait Strategy
 *
 *****************************************************************
 */
TRB_POSIX_AIO_SUN_Port_Strategy::~TRB_POSIX_AIO_SUN_Port_Strategy()
{
  this->fini ();
}

TRB_POSIX_AIO_SUN_Port_Strategy::TRB_POSIX_AIO_SUN_Port_Strategy 
  (TRB_POSIX_AIO_Select_Provider & provider)
  : TRB_POSIX_AIO_Select_Strategy_Base (provider),
    port_fd_    (ACE_INVALID_HANDLE),
    //events_     (0),
    next_event_ (0),
    num_events_ (0)
{
}

int 
TRB_POSIX_AIO_SUN_Port_Strategy::init (void)
{
  if (this->alloc_repository (ACE_AIO_NO_LIMIT, 0) != 0)
    return -1;;

  ACE_DEBUG ((LM_DEBUG,
    ACE_LIB_TEXT("(%t)TRB_POSIX_AIO_SUN_Port_Strategy size=%d\n"), 
    (int) this->size_));

  // Allocate the array before opening the device to avoid a potential
  // resource leak if allocation fails.
  //ACE_NEW_RETURN (this->events_,
  //                port_event_t [POLL_BUF_SIZE+1],
  //                -1);

  // Initialiaze the queue
  this->port_fd_ = ::port_create ();
  if (this->port_fd_ == ACE_INVALID_HANDLE)
    return -1;
  
  return 0;
}  

int 
TRB_POSIX_AIO_SUN_Port_Strategy::fini (void)
{
  //delete [] this->events_;
  //this->events_ = 0;

  if (this->port_fd_ != ACE_INVALID_HANDLE)
    {
       ACE_OS::close (this->port_fd_);
    }

  this->port_fd_ = ACE_INVALID_HANDLE;
  this->num_events_ = 0;
  
  this->free_repository ();

  return 0;
}

int 
TRB_POSIX_AIO_SUN_Port_Strategy::on_aio_start (TRB_POSIX_Asynch_Result * result)
{
  ACE_HANDLE handle = result->aio_fildes;

  int blk_mode = ACE_BIT_DISABLED (ACE::get_flags (handle), ACE_NONBLOCK);

  // Set the handle into non-blocking mode if it's not already in it.
  if (blk_mode)
    ACE::set_flags (handle, ACE_NONBLOCK);

  if (result->op_code () == TRB_Asynch_Result_Impl::CONNECT)
    {
      if (this->pending_writes_[handle] == IO_READY)
          this->pending_writes_[handle] = 0;

      if (this->pending_reads_ [handle] == IO_READY)
          this->pending_reads_ [handle] = 0;
    }

  return 0; 
}


int
TRB_POSIX_AIO_SUN_Port_Strategy::prepare_to_wait (TRB_POSIX_Asynch_Waiting_List & started_list)
{
  ACE_TRACE ("TRB_POSIX_AIO_SUN_Port_Strategy::prepare_to_wait");

  ACE_UNUSED_ARG (started_list);

  if (this->num_events_ > 0)
    return WAIT_COMPLETED; 

  return WAIT_ENABLED;
}

int
TRB_POSIX_AIO_SUN_Port_Strategy::wait (ACE_Time_Value * wait_time)
{
  ACE_TRACE ("TRB_POSIX_AIO_SUN_Port_Strategy::wait");

  timespec_t   timeout;
  timespec_t * ptimeout = 0;

  
  if (wait_time != 0 )
  {
  	 timeout =  wait_time->operator timespec_t ();
     ptimeout = &timeout;
  }

  this->fd_info_.set(ACE_INVALID_HANDLE, 0);
  
  this->next_event_ = this->events_;
  this->num_events_ = 0;

  // Poll for events
//  int rc = ::port_getn (this->port_fd_, 
//                        this->events_,
//                        PORT_GET_BUF_SIZE,   // (sizeof(events_)/sizeof(events_[0])),
//                        &this->num_events_,
//                        ptimeout);
  int rc = ::port_get (this->port_fd_, 
                        this->events_,
                        ptimeout);

  if (rc == 0)
   num_events_=1;

  //ACE_ERROR ((LM_ERROR,
  //   "Port_GetN: rc=%d port=%d num_events=%d ptimeout=%@\n",
  //    rc, port_fd_, num_events_, ptimeout));

  if (rc < 0)
  {
 
    switch (errno)
    {
    case EINTR :    // Interrupted call
      return WAIT_INTERRUPTED;

    case ETIME:
      return WAIT_TIMEOUT;
    default:
      break;
    }
    return WAIT_ERROR;
  }
                                  
  this->find_next_ready_fd ();
      
  return WAIT_COMPLETED;
}

bool 
TRB_POSIX_AIO_SUN_Port_Strategy::find_next_ready_fd ()
{
  bool rc = false;

  for (; !rc && num_events_ > 0;  --num_events_, ++next_event_ )
    {
      if (next_event_->portev_source != PORT_SOURCE_FD)
          continue;
     
      fd_info_.set (next_event_->portev_object,
                    next_event_->portev_events);
      rc = true;
    }

  return rc;
}

TRB_POSIX_Asynch_Result *
TRB_POSIX_AIO_SUN_Port_Strategy::get_next_completed_result 
  (TRB_POSIX_Asynch_Waiting_List & started_list)
{
  ACE_UNUSED_ARG (started_list);

  for (;;)
    {
      ACE_HANDLE handle = this->fd_info_.handle();

      if (handle == ACE_INVALID_HANDLE)
        return 0;

      int rc = this->check (this->fd_info_);
      

      if (rc)
        {
          TRB_POSIX_Asynch_Result * result = this->fd_info_.get_result();
          ACE_ASSERT (result !=0);
          return result;
        }

      u_int mask = get_interest (handle);

      if (mask != 0)
        declare_interest (handle,mask);

      if (!find_next_ready_fd ())
        break;
    }

  return 0;
}

// *****************************************************************
//  Declare interest
//  Return code :
//    1 - we should interrupt leader to update wait state
//    0 - we do not have to interrupt leader
// *****************************************************************

int
TRB_POSIX_AIO_SUN_Port_Strategy::declare_interest (ACE_HANDLE handle, 
                                                   u_int mask)
{  

  int rc = ::port_associate (port_fd_, PORT_SOURCE_FD, handle, mask, 0);

  //ACE_ERROR ((LM_ERROR,
  //   "Declare_interest: rc=%d port=%d handle=%d mask=%x\n",
  //    rc, port_fd_, (int) handle, mask ));

  if (rc != 0)
    ACE_ERROR_RETURN ((LM_ERROR,
        ACE_LIB_TEXT("TRB_POSIX_AIO_SUN_Port_Strategy::declare_interest: %p\n"),
        ACE_LIB_TEXT("port_associate failed")),
        0);

  return 0;
}

int
TRB_POSIX_AIO_SUN_Port_Strategy::clear_interest (ACE_HANDLE handle, 
                                                 u_int mask)
{  
  ACE_UNUSED_ARG (handle);
  ACE_UNUSED_ARG (mask);
  return 0;
}


#endif /* ACE_HAS_SUN_COMPLETION_PORTS */
#endif /* ACE_HAS_AIO_CALLS */
