/*********************************************************************
** Copyright (C) 2003 Terabit Pty Ltd.  All rights reserved.
**
** This file is part of the POSIX-Proactor module.
**
** This file may be distributed and used only under the terms of the  
** Terabit Public License as defined by Terabit Pty Ltd of Australia   
** and appearing in the file tlicense.txt included in the packaging of
** this module and available at http://www.terabit.com.au/license.php.
**
** Contact support@terabit.com.au if any conditions of this licensing
** are not clear to you.
**********************************************************************/

/* -*- C++ -*- */

//=============================================================================
/**
 *  @file    TRB_POSIX_AIO_SUN_Port_Strategy.h
 *
 *  $Id:  $
 *
 *  @author Alexander Libman <libman@terabit.com.au>
 */
//=============================================================================

#ifndef TRB_POSIX_AIO_SUN_PORT_STRATEGY_H 
#define TRB_POSIX_AIO_SUN_PORT_STRATEGY_H 

#include "TProactor/POSIX_AIO_Select_Provider.h"


#if defined (ACE_HAS_SUN_COMPLETION_PORTS) && defined (sun)

#if defined (ACE_HAS_AIO_CALLS) || defined(ACE_HAS_AIO_EMULATION)



#include /**/ <port.h>


/**
 * @class TRB_POSIX_AIO_SUN_Port_Strategy
 *
 */

class TPROACTOR_Export TRB_POSIX_AIO_SUN_Port_Strategy :
  public TRB_POSIX_AIO_Select_Strategy_Base
{
protected:

  friend class TRB_POSIX_AIO_Select_Provider;
  
  TRB_POSIX_AIO_SUN_Port_Strategy (TRB_POSIX_AIO_Select_Provider & provider);
  virtual ~TRB_POSIX_AIO_SUN_Port_Strategy();


  // Wait Strategy Interface

  // use base class
  // virtual TRB_POSIX_AIO_Interrupt_Strategy * create_interrupt_strategy (void);
  
  // use base class
  // virtual void 
  //  destroy_interrupt_strategy (TRB_POSIX_AIO_Interrupt_Strategy * i_strategy);
  virtual int init (void);
  virtual int fini (void);

  virtual int  on_aio_start (TRB_POSIX_Asynch_Result * result);
  
  virtual int  prepare_to_wait (TRB_POSIX_Asynch_Waiting_List & started_list);

  virtual int  wait (ACE_Time_Value * wait_time);

  virtual TRB_POSIX_Asynch_Result * 
    get_next_completed_result (TRB_POSIX_Asynch_Waiting_List & started_list);


  virtual int declare_interest (ACE_HANDLE handle, u_int mask);
  virtual int clear_interest   (ACE_HANDLE handle, u_int mask);

private:
  bool  find_next_ready_fd ();

  enum  {PORT_GET_BUF_SIZE = 100 };


  /**
   * All interactions with epoll are
   * done through this file descriptor.
   */
  ACE_HANDLE port_fd_;

  /// The poll_event_t array that port_get will feed its results to.
  port_event_t      events_ [PORT_GET_BUF_SIZE];

  /// Pointer to the next poll_event_t array element that contains the next
  /// event to be dispatched.
  port_event_t *    next_event_;
  u_int num_events_;

};

#endif /* ACE_HAS_AIO_CALLS */
#endif /* ACE_HAS_SUN_COMPLETION_PORTS */
#endif /* TRB_POSIX_AIO_SUN_PORT_STRATEGY_H */
