/*********************************************************************
** Copyright (C) 2003 Terabit Pty Ltd.  All rights reserved.
**
** This file is part of the POSIX-Proactor module.
**
** This file may be distributed and used only under the terms of the  
** Terabit Public License as defined by Terabit Pty Ltd of Australia   
** and appearing in the file tlicense.txt included in the packaging of
** this module and available at http://www.terabit.com.au/license.php.
**
** Contact support@terabit.com.au if any conditions of this licensing
** are not clear to you.
**********************************************************************/

/* -*- C++ -*- */
// $Id:  $

#include "TProactor/POSIX_AIO_Select_Provider.h"
#include "TProactor/POSIX_AIO_Select_Strategy.h"
#include "TProactor/POSIX_AIO_Poll_Strategy.h"
#include "TProactor/POSIX_AIO_Dev_Poll_Strategy.h"
#include "TProactor/POSIX_AIO_Event_Poll_Strategy.h"
#include "TProactor/POSIX_AIO_Linux_RT_Strategy.h"
#include "TProactor/POSIX_AIO_SUN_Port_Strategy.h"

#if defined (ACE_HAS_AIO_CALLS) || defined(ACE_HAS_AIO_EMULATION)

//----------------------------------------------------------
#include "ace/Version.h"

#if ((ACE_MAJOR_VERSION == 5) && (ACE_MINOR_VERSION ==3))

#include "ace/OS.h"

#else  // we assume ACE 5.4.X

#include "ace/OS_NS_sys_socket.h"

#endif  //((ACE_MAJOR_VERSION == 5) && (ACE_MINOR_VERSION ==3))
//----------------------------------------------------------


#include "ace/Log_Msg.h"
#include "ace/Addr.h"
#include "ace/ACE.h"



TRB_POSIX_AIO_Select_Strategy_Base::Handle_Info::Handle_Info (void)
  : fd_(ACE_INVALID_HANDLE),
    ready_events_(0),
    events_to_clear_(0),
    events_to_set_(0),
    ready_result_(0)
{
  ;
}  

TRB_POSIX_AIO_Select_Strategy_Base::Handle_Info::~Handle_Info (void)
{
  ;
}  

int 
TRB_POSIX_AIO_Select_Strategy_Base::Handle_Info::handle(void) const
{
  return this->fd_;
}

u_int
TRB_POSIX_AIO_Select_Strategy_Base::Handle_Info::events_to_clear(void) const
{
  return this->events_to_clear_;
}

u_int
TRB_POSIX_AIO_Select_Strategy_Base::Handle_Info::events_to_set(void) const
{
  return this->events_to_set_;
}

u_int
TRB_POSIX_AIO_Select_Strategy_Base::Handle_Info::ready_events(void) const
{
  return this->ready_events_;
}

int
TRB_POSIX_AIO_Select_Strategy_Base::Handle_Info::is_read_ready(void) const
{
  return ACE_BIT_ENABLED (this->ready_events_, 
                          (POLLIN|POLLHUP|POLLERR|POLLNVAL));
}

int
TRB_POSIX_AIO_Select_Strategy_Base::Handle_Info::is_write_ready(void) const
{
  return ACE_BIT_ENABLED (this->ready_events_, 
                          (POLLOUT|POLLHUP|POLLERR|POLLNVAL));
}

void
TRB_POSIX_AIO_Select_Strategy_Base::Handle_Info::set(ACE_HANDLE fd, 
                                                      u_int ready_events)
{
  ACE_ASSERT (this->ready_result_ == 0);

  this->fd_ = fd;
  this->ready_events_ = ready_events;
  this->events_to_clear_ = 0;
  this->events_to_set_ = 0;
}

u_int
TRB_POSIX_AIO_Select_Strategy_Base::Handle_Info::clear_ready(u_int events)
{
  this->ready_events_ &= ~events;
  return this->ready_events_;
}

u_int
TRB_POSIX_AIO_Select_Strategy_Base::Handle_Info::request_to_clear(u_int events)
{
  this->events_to_clear_ |= events;
  this->events_to_set_   &= ~events;

  return this->events_to_clear_;
}

u_int
TRB_POSIX_AIO_Select_Strategy_Base::Handle_Info::request_to_set(u_int events)
{
  this->events_to_clear_ &= ~events;
  this->events_to_set_   |= events;

  return this->events_to_set_;
}

TRB_POSIX_Asynch_Result *
TRB_POSIX_AIO_Select_Strategy_Base::Handle_Info::get_result(void)
{
  TRB_POSIX_Asynch_Result * result = this->ready_result_;
  this->ready_result_ = 0;
  return result;
}

void
TRB_POSIX_AIO_Select_Strategy_Base::Handle_Info::set_result(TRB_POSIX_Asynch_Result * result)
{
  ACE_ASSERT (this->ready_result_ == 0);
  this->ready_result_ = result;
}

u_int
TRB_POSIX_AIO_Select_Strategy_Base::Handle_Info::poll (void)
{
  struct pollfd pfd;
  
  pfd.fd      = this->fd_;
  pfd.events  = (POLLIN|POLLOUT);
  pfd.revents = 0;

  ::poll (&pfd,1, 0);

  this->ready_events_ = pfd.revents;
  return this->ready_events_;
}
      
/**
 *****************************************************************
 *  Select_Strategy_Base  - abstract interface for Select_Provider
 *  AIO Wait Strategies:
 *    Select_Strategy
 *    Poll_Strategy
 *    Dev_Poll_Strategy
 *    Linux_RT_Strategy
 *  
 *
 *****************************************************************
 */
TRB_POSIX_AIO_Select_Strategy_Base::~TRB_POSIX_AIO_Select_Strategy_Base()
{
  this->free_repository();
}

TRB_POSIX_AIO_Select_Strategy_Base::TRB_POSIX_AIO_Select_Strategy_Base 
  (TRB_POSIX_AIO_Select_Provider & provider)
  : TRB_POSIX_AIO_Wait_Strategy (provider),
    sel_provider_ (provider),
    pending_reads_(0),
    pending_writes_(0),
    size_(ACE_AIO_MAX_SIZE),
    fd_info_()
{
  ;
}

int 
TRB_POSIX_AIO_Select_Strategy_Base::alloc_repository (size_t size, 
                                                      int flg_ready)
{
  this->free_repository ();
  
  if (size == 0)
    size = ACE_AIO_NO_LIMIT;
    
  size_t max_num_files = ACE_static_cast(size_t, ACE::max_handles ());
  
  if (max_num_files < size)
    size = max_num_files;

  if (size == ACE_AIO_NO_LIMIT)
    size = ACE_AIO_MAX_SIZE;

  TRB_POSIX_Asynch_Result * result = flg_ready ? IO_READY : 0;
  size_t i = 0;

  ACE_NEW_RETURN (this->pending_reads_,
                  TRB_POSIX_Asynch_Result* [this->size_],
                  -1);

  for (i=0; i < this->size_ ; ++i)
    this->pending_reads_ [i]  = result;

  ACE_NEW_RETURN (this->pending_writes_,
                  TRB_POSIX_Asynch_Result* [this->size_],
                  -1);


  for (i=0; i < this->size_ ; ++i)
    this->pending_writes_ [i] = result;

  return 0;
}

int 
TRB_POSIX_AIO_Select_Strategy_Base::free_repository (void)
{
  delete [] this->pending_reads_;
  delete [] this->pending_writes_;

  this->pending_reads_ = 0;
  this->pending_writes_ =0;

  return 0;
}

u_int
TRB_POSIX_AIO_Select_Strategy_Base::get_interest (ACE_HANDLE handle)
{
  u_int rc = 0;
  if (handle >= 0 || ACE_static_cast(size_t,handle) < this->size_ )
  {
     TRB_POSIX_Asynch_Result * result;

     result = this->pending_writes_[handle];
     if (result != 0 && result != IO_READY) // no outstanding writes
         rc |= POLLOUT;

     result = this->pending_reads_[handle];
     if (result != 0 && result != IO_READY) // no outstanding writes
         rc |= POLLIN;
  }
  return rc;
}
        
// *****************************************************************
//  Start operations
//  On return result state contains:
//    STARTED       - "AIO started", i.e. handle is in select mask 
//    DEFFERED      - is never returned for this provider 
//    START_ERROR   - IO unable to start, i.e. invalid handle
//    RESULT_READY  - IO started and completed immediately
//  Return code :
//    1 - we should interrupt leader to update wait state
//    0 - we do not have to interrupt leader
// *****************************************************************

int
TRB_POSIX_AIO_Select_Strategy_Base::start_read (TRB_POSIX_Asynch_Result *result)
{  
  result->cont_ = 0;

  ACE_HANDLE handle = result->aio_fildes;

  if (handle < 0 || ACE_static_cast(size_t,handle) >= this->size_ )
    {
      result->set_start_error (EBADF);
      return 0;
    }
  
  int rc = 0;

  TRB_POSIX_Asynch_Result * first = this->pending_reads_[handle];
  
  if (first == IO_READY) // handle is ready to read (possible)
    {
      rc = this->sel_provider_.do_read (result);

      if (rc != 0)       // read finished 
        return 0;        // no need to interrupt

      first = 0;
    }

  // handle is not ready
  if (first == 0)        // no pending reads
    {
      this->pending_reads_[handle] = result;

      return this->declare_interest (handle, POLLIN);
    }

  // queue result after pending reads
  while (first->cont_ !=0)
    {
      first = first->cont_;
    }

  first->cont_ = result;
  return 0; 

}

int
TRB_POSIX_AIO_Select_Strategy_Base::start_write (TRB_POSIX_Asynch_Result *result)
{  
  result->cont_ = 0;

  ACE_HANDLE handle = result->aio_fildes;

  if (handle < 0 || ACE_static_cast(size_t,handle) >= this->size_ )
    {
      result->set_start_error (EBADF);
      return 0;
    }
  
  int rc = 0;

  TRB_POSIX_Asynch_Result * first = this->pending_writes_[handle];
  
  if (first == IO_READY) // handle is ready to write (possible)
    {
      rc = this->sel_provider_.do_write (result);

      if (rc != 0)       // write finished
        return 0;        // no need to interrupt

      first = 0;
    }

  // handle is not ready
  if (first == 0)        // no pending writes
    {
      this->pending_writes_[handle] = result;

      return this->declare_interest (handle, POLLOUT);
    }

  // queue result after pending writes
  while (first->cont_ !=0)
    {
      first = first->cont_;
    }

  first->cont_ = result;
  return 0; 
}

// *****************************************************************
//  Cancel read/write
// *****************************************************************

int
TRB_POSIX_AIO_Select_Strategy_Base::cancel_read (TRB_POSIX_Asynch_Result * result)
{
  ACE_HANDLE handle = result->aio_fildes;

  if (handle < 0 || ACE_static_cast(size_t,handle) >= this->size_)
    return -1;
  
  TRB_POSIX_Asynch_Result * first = this->pending_reads_[handle];

  if (first == IO_READY)  // no pending reads
    return 1;

  if (first == result)    // we are first
    {
      this->pending_reads_[handle] = result->cont_;
      result->cont_ = 0;
    }
  else for (; first != 0; first = first->cont_)
    {
      if (first->cont_ == result)
        {
           first->cont_ = result->cont_;
           result->cont_ = 0;
           break;
        }
    }

  if (this->pending_reads_[handle] == 0)
    this->clear_interest (handle, POLLIN);

  return 1;
}

int
TRB_POSIX_AIO_Select_Strategy_Base::cancel_write (TRB_POSIX_Asynch_Result * result)
{
  ACE_HANDLE handle = result->aio_fildes;

  if (handle < 0 || ACE_static_cast(size_t,handle) >= this->size_)
    return -1;
  
  TRB_POSIX_Asynch_Result * first = this->pending_writes_[handle];

  if (first == IO_READY)   // no pending writes
    return 1;

  if (first == result)     // we are first
    {
      this->pending_writes_[handle] = result->cont_;
      result->cont_ = 0;
    }
  else for (; first != 0; first = first->cont_)
    {
      if (first->cont_ == result)
        {
           first->cont_ = result->cont_;
           result->cont_ = 0;
           break;
        }
    }

  if (this->pending_writes_[handle] == 0)
    this->clear_interest (handle, POLLOUT);

  return 1;
}

// *****************************************************************
//  check
// *****************************************************************

int
TRB_POSIX_AIO_Select_Strategy_Base::check (Handle_Info & fd_info)
{
  ACE_ASSERT (fd_info.ready_result_ == 0);

  ACE_HANDLE handle = fd_info.handle ();

  TRB_POSIX_Asynch_Result * result=0;

  if (fd_info.is_write_ready())
    {
        result = this->pending_writes_[handle];

        if (result == 0 || result == IO_READY) // no outstanding writes
          {
             this->pending_writes_[handle] = IO_READY;

             fd_info.clear_ready (POLLOUT); // not to repeat
             fd_info.request_to_clear (POLLOUT);
          }
        else 
          {
            int rc = this->sel_provider_.do_write (result);
            
            if (rc == 0)   // opeartion NOT finished
              {
                fd_info.clear_ready (POLLOUT);
                fd_info.request_to_set (POLLOUT);
              }
            else  // finished
              {
                fd_info.set_result (result);
                this->pending_writes_[handle] = result->cont_;
                result->cont_ = 0;

                return 1;   // has result
              }
          }
    }

  if (fd_info.is_read_ready())
    {
        result = this->pending_reads_[handle];

        if (result == 0 || result == IO_READY) // no outstanding reads
          {
             this->pending_reads_[handle] = IO_READY;

             fd_info.clear_ready (POLLIN); // not to repeat
             fd_info.request_to_clear (POLLIN);
          }
        else 
          {
            int rc = this->sel_provider_.do_read (result);
            
            if (rc == 0)   // opeartion NOT finished
              {
                fd_info.clear_ready (POLLIN);
                fd_info.request_to_set (POLLIN);
              }
            else  // finished
              {
                fd_info.set_result (result);
                this->pending_reads_[handle] = result->cont_;
                result->cont_ = 0;

                return 1;   // has result
              }
          }
    }

  return 0;
}

/**
 *****************************************************************
 *  POSIX Select-based AIO Provider
 *
 *****************************************************************
 */
TRB_POSIX_AIO_Select_Provider::~TRB_POSIX_AIO_Select_Provider ()
{
  ACE_ASSERT (this->sel_strategy_ == 0);
}

TRB_POSIX_AIO_Select_Provider::TRB_POSIX_AIO_Select_Provider (const TRB_POSIX_AIO_Config & config)
  : TRB_POSIX_AIO_Provider (config),
    sel_strategy_ (0)
{
  this->max_aio_num_ = ACE_AIO_NO_LIMIT ;
}  

int
TRB_POSIX_AIO_Select_Provider::get_supported_operations_mask () const
{
  return (TRB_Asynch_Result_Impl::READ_STREAM 
         |TRB_Asynch_Result_Impl::WRITE_STREAM 
         |TRB_Asynch_Result_Impl::READ_FILE 
         |TRB_Asynch_Result_Impl::WRITE_FILE
         |TRB_Asynch_Result_Impl::READ_DGRAM 
         |TRB_Asynch_Result_Impl::WRITE_DGRAM 
         |TRB_Asynch_Result_Impl::ACCEPT 
         |TRB_Asynch_Result_Impl::CONNECT);
         
}  
  
 
int 
TRB_POSIX_AIO_Select_Provider::should_start_interrupt_leader () const
{
  // We do  have to inrterrupt select ()
  // when we start new aio
  // return 1;
  return TRB_POSIX_AIO_Provider::should_start_interrupt_leader ();
}   

int TRB_POSIX_AIO_Select_Provider::can_start_in_any_thread () const
{
  // Yes we can
  // return 1;
  return TRB_POSIX_AIO_Provider::can_start_in_any_thread ();
}   

size_t
TRB_POSIX_AIO_Select_Provider::get_max_aio_num () const
{
  return this->max_aio_num_; // base class
}

TRB_POSIX_AIO_Wait_Strategy * 
TRB_POSIX_AIO_Select_Provider::create_wait_strategy (void)
{
  if (this->sel_strategy_ != 0)
    return this->sel_strategy_;

  int ws_type = this->config_.ws_type();

  switch (ws_type)
    {

#if defined (linux)

    case TRB_POSIX_AIO_Config::WST_LINUXRT:

      if (this->config().processor_type ()==
          TRB_POSIX_AIO_Config::PCT_DEDICATED )
        {
          ACE_NEW_RETURN (this->sel_strategy_,
                          TRB_POSIX_AIO_Linux_RT_Strategy (*this),
                          0);
          break;
        }

#endif /* linux */

#if defined (ACE_HAS_SUN_COMPLETION_PORTS) && defined (sun)

    case TRB_POSIX_AIO_Config::WST_SUN_PORT:
      ACE_NEW_RETURN (this->sel_strategy_,
                      TRB_POSIX_AIO_SUN_Port_Strategy (*this),
                      0);
      break;

#endif

#if defined (ACE_HAS_LINUX_EPOLL)

    case TRB_POSIX_AIO_Config::WST_EPOLL:
      ACE_NEW_RETURN (this->sel_strategy_,
                      TRB_POSIX_AIO_Event_Poll_Strategy (*this),
                      0);
      break;

#elif defined (ACE_HAS_DEV_POLL)

    case TRB_POSIX_AIO_Config::WST_DEVPOLL:
      ACE_NEW_RETURN (this->sel_strategy_,
                      TRB_POSIX_AIO_Dev_Poll_Strategy (*this),
                      0);
      break;

#endif /* (ACE_HAS_LINUX_EPOLL) || defined (ACE_HAS_DEV_POLL) */

    case TRB_POSIX_AIO_Config::WST_POLL:
      ACE_NEW_RETURN (this->sel_strategy_,
                      TRB_POSIX_AIO_Poll_Strategy (*this),
                      0);
      break;

    default:
      ACE_NEW_RETURN (this->sel_strategy_,
                      TRB_POSIX_AIO_Select_Strategy (*this),
                      0);

      break;
    }


  return this->sel_strategy_;
}

void
TRB_POSIX_AIO_Select_Provider::destroy_wait_strategy (TRB_POSIX_AIO_Wait_Strategy * w_strategy)
{
  ACE_ASSERT (this->sel_strategy_ == w_strategy);

  delete  this->sel_strategy_;
  this->sel_strategy_ =0;
}

int 
TRB_POSIX_AIO_Select_Provider::register_handle (ACE_HANDLE handle,
                                         const void *completion_key)
{
  ACE_UNUSED_ARG (completion_key);

  ACE::set_flags (handle, ACE_NONBLOCK);

  return 0;
}

int 
TRB_POSIX_AIO_Select_Provider::unregister_handle (ACE_HANDLE handle)
{
  ACE_UNUSED_ARG (handle);

  return 0;
}


// *************************************************************
//   Cancel AIO
// *************************************************************

int
TRB_POSIX_AIO_Select_Provider::cancel_aio (TRB_POSIX_Asynch_Result * result)
{

  result->set_ready (0, ECANCELED, 0);

  int rc = 0;

  switch (result->op_code ())
    {
    case TRB_Asynch_Result_Impl::NO_OP:  
      break;

    case TRB_Asynch_Result_Impl::READ_STREAM:  
    case TRB_Asynch_Result_Impl::READ_FILE:    
    case TRB_Asynch_Result_Impl::READ_DGRAM:   
    case TRB_Asynch_Result_Impl::ACCEPT:
      rc = this->sel_strategy_->cancel_read (result);
      break;

    case TRB_Asynch_Result_Impl::WRITE_STREAM: 
    case TRB_Asynch_Result_Impl::WRITE_FILE:   
    case TRB_Asynch_Result_Impl::WRITE_DGRAM:  
    case TRB_Asynch_Result_Impl::CONNECT: 
      rc = this->sel_strategy_->cancel_write (result);
      break;

    case TRB_Asynch_Result_Impl::TRANSMIT_FILE:
    case TRB_Asynch_Result_Impl::TIMER:        
    default:
      break;
    }

  if (rc > 0)
    return CR_CANCELED;
  
  if (rc == 0)
    return CR_ALL_DONE;

  return CR_CANCEL_ERROR;  // error
}

// *****************************************************************
//  Start AIO
//  On return result state contains:
//    STARTED       - "AIO started", i.e. handle is in select mask 
//    DEFFERED      - is never returned for this provider 
//    START_ERROR   - IO unable to start, i.e. invalid handle
//    RESULT_READY  - IO started and completed immediately
//  Return code :
//    1 - we should interrupt leader to update wait state
//    0 - we do not have to interrupt leader
// *****************************************************************

int
TRB_POSIX_AIO_Select_Provider::start_read_stream (TRB_POSIX_Asynch_Result *result)
{  
  return this->sel_strategy_->start_read (result);
}

int
TRB_POSIX_AIO_Select_Provider::start_write_stream (TRB_POSIX_Asynch_Result *result)
{  
  return this->sel_strategy_->start_write (result);
}

int
TRB_POSIX_AIO_Select_Provider::start_read_file (TRB_POSIX_Asynch_Result *result)
{  
  return this->sel_strategy_->start_read (result);
}

int
TRB_POSIX_AIO_Select_Provider::start_write_file (TRB_POSIX_Asynch_Result *result)
{  
  return this->sel_strategy_->start_write (result);
}

int
TRB_POSIX_AIO_Select_Provider::start_read_dgram (TRB_POSIX_Asynch_Result *result)
{
  return this->sel_strategy_->start_read (result);
}

int
TRB_POSIX_AIO_Select_Provider::start_write_dgram (TRB_POSIX_Asynch_Result *result)
{  
  return this->sel_strategy_->start_write (result);
}

int
TRB_POSIX_AIO_Select_Provider::start_accept (TRB_POSIX_Asynch_Result *result)
{
  return this->sel_strategy_->start_read (result);
}

int
TRB_POSIX_AIO_Select_Provider::start_connect (TRB_POSIX_Asynch_Result *result)
{  
  TRB_POSIX_Asynch_Connect_Result * connect_result =
    ACE_dynamic_cast (TRB_POSIX_Asynch_Connect_Result*, result);

  if (connect_result == 0)
    {
      result->set_start_error (EINVAL);
      return 0;
    }

  const ACE_Addr & addr = connect_result->remote_address ();

  ACE::set_flags (result->aio_fildes, ACE_NONBLOCK);

  for (;;)
    {
      int rc = ACE_OS::connect 
                 (result->aio_fildes,
                  ACE_reinterpret_cast (sockaddr *, addr.get_addr ()),
                  addr.get_size ());

      if (rc == 0)
        {
          result->set_ready (0, 0, 0);
          return 0;
        }
      else if (errno == EINTR)
        continue;
      else if (errno == EWOULDBLOCK || errno == EINPROGRESS )
        break;
      else
        {
          result->set_ready (0, errno, 0);
          return 0;
        }
    }

  return this->sel_strategy_->start_write (result);
}

// *****************************************************************
//  Execute operations : do_read/write_XXXX
//  Return codes:
//  2 - AIO finished, next operation possible on this fd
//  1 - AIO finished, next operation is not possible on this fd
//  0 - not finished
// *****************************************************************
int
TRB_POSIX_AIO_Select_Provider::do_read (TRB_POSIX_Asynch_Result *result)
{
  int rc = -1;

  switch (result->op_code ())
    {
    case TRB_Asynch_Result_Impl::READ_STREAM:  
      rc = this->do_read_stream (result);
      break;
    case TRB_Asynch_Result_Impl::READ_FILE:
      rc = this->do_read_file (result);
      break;
    case TRB_Asynch_Result_Impl::READ_DGRAM:
      rc = this->do_read_dgram (result);
      break;
    case TRB_Asynch_Result_Impl::ACCEPT:
      rc = this->do_accept (result);
      break;
    default:
      ACE_ASSERT (0);
      result->set_ready (0, EINVAL, 0);
      break;
    }

//  if (rc == 0)
//     ACE_DEBUG((LM_DEBUG,"False READ fd=%d\n", result->aio_fildes));

  return rc;   
}

int
TRB_POSIX_AIO_Select_Provider::do_write (TRB_POSIX_Asynch_Result *result)
{
  int rc = -1;

  switch (result->op_code ())
    {
    case TRB_Asynch_Result_Impl::WRITE_STREAM:
      rc = this->do_write_stream (result);
      break;
    case TRB_Asynch_Result_Impl::WRITE_FILE:
      rc = this->do_write_file (result);
      break;
    case TRB_Asynch_Result_Impl::WRITE_DGRAM:
      rc = this->do_write_dgram (result);
      break;
    case TRB_Asynch_Result_Impl::CONNECT: 
      rc = this->do_connect (result);
      break;
    default:
      ACE_ASSERT (0);
      result->set_ready (0, EINVAL, 0);
      break;
    }

//  if (rc == 0)
//     ACE_DEBUG((LM_DEBUG,"False WRITE fd=%d\n", result->aio_fildes));

  return rc;
}


int
TRB_POSIX_AIO_Select_Provider::do_read_stream (TRB_POSIX_Asynch_Result *result)
{ 
  ACE_HANDLE handle = result->aio_fildes;

  int aio_error = 0;
  ssize_t aio_return = ACE_OS::read (handle,
                                     (char *) result->aio_buf,
                                     result->aio_nbytes);

  int rc = 1;  // next op not possible

  if (aio_return < 0)
    {
      aio_return = 0;
      aio_error = ACE_OS::last_error ();

      if (aio_error == EWOULDBLOCK)
        return 0;
    }
  else if (aio_return == 0)
    {
      rc = 2;  // EOF, next op possible
    }
  else if (aio_return == ACE_static_cast (ssize_t,result->aio_nbytes))
    {
      rc = 2;  // next op possible
    }


  result->set_ready (aio_return, aio_error, 0);
  return rc;
}
  
int
TRB_POSIX_AIO_Select_Provider::do_write_stream (TRB_POSIX_Asynch_Result *result)
{  
  ACE_HANDLE handle = result->aio_fildes;

  int aio_error = 0;
  ssize_t aio_return = ACE_OS::write (handle,
                                      (const char *) result->aio_buf,
                                      result->aio_nbytes);

  int rc = 1;  // next op not possible

  if (aio_return < 0)
    {
      aio_return = 0;
      aio_error = ACE_OS::last_error ();

      if (aio_error == EWOULDBLOCK)
        return 0;
    }
  else if (aio_return == 0)
    {
      rc = 2;  // next op possible
    }
  else if (aio_return == ACE_static_cast (ssize_t,result->aio_nbytes))
    {
      rc = 2;  // next op possible
    }



  result->set_ready (aio_return, aio_error, 0);
  return rc;
}

int
TRB_POSIX_AIO_Select_Provider::do_read_file (TRB_POSIX_Asynch_Result *result)
{ 
  off_t pos = ACE_OS::lseek (result->aio_fildes,
                             result->aio_offset,
                             SEEK_SET);

  int aio_error = 0;
  ssize_t aio_return =0;

  if (pos == ACE_static_cast (off_t, -1))
    {
      aio_error = ACE_OS::last_error ();
    }
  else
    {
      aio_return = ACE_OS::read (result->aio_fildes,
                                 (char *) result->aio_buf,
                                 result->aio_nbytes);

      if (aio_return < 0)
        {
          aio_return = 0;
          aio_error = ACE_OS::last_error ();
        }
    }

  if (aio_error == EWOULDBLOCK)
    return 0;

  result->set_ready (aio_return, aio_error, 0);
  return 2;  // regular files -> always ready
}

  
int
TRB_POSIX_AIO_Select_Provider::do_write_file (TRB_POSIX_Asynch_Result *result)
{  
  off_t pos = ACE_OS::lseek (result->aio_fildes,
                             result->aio_offset,
                             SEEK_SET);

  int aio_error = 0;
  ssize_t aio_return =0;

  if (pos == ACE_static_cast (off_t, -1))
    {
      aio_error = ACE_OS::last_error ();
    }
  else
    {
      aio_return = ACE_OS::write (result->aio_fildes,
                                  (const char *) result->aio_buf,
                                  result->aio_nbytes);

      if (aio_return < 0)
        {
          aio_return = 0;
          aio_error = ACE_OS::last_error ();
        }
    }

  result->set_ready (aio_return, aio_error, 0);
  return 2;  // regular files -> always ready
}

int
TRB_POSIX_AIO_Select_Provider::do_read_dgram (TRB_POSIX_Asynch_Result *result)
{
  TRB_POSIX_Asynch_Read_Dgram_Result * rd_result =
    ACE_dynamic_cast (TRB_POSIX_Asynch_Read_Dgram_Result*, result);

  ACE_ASSERT (rd_result != 0);

  ACE_HANDLE handle = result->aio_fildes;

  struct msghdr * msg_hdr = rd_result->msghdr();

  int     aio_error = 0;
  ssize_t aio_return = 0;

  if (msg_hdr->msg_iovlen > 0 && 
      rd_result->bytes_to_read() > 0)
    {
      aio_return =  ACE_OS::recvmsg(handle,
                                    msg_hdr,
                                    rd_result->flags());
  
      if (aio_return < 0)
        {
          aio_return = 0;
          aio_error = ACE_OS::last_error ();
        }
    }
  else
    {
      aio_error = EINVAL;
    }

  if (aio_error == EWOULDBLOCK)
    return 0;

  result->set_ready (aio_return, aio_error, 0);
  return 1; // we do not know if next op is possible
}

int
TRB_POSIX_AIO_Select_Provider::do_write_dgram (TRB_POSIX_Asynch_Result *result)
{  
  TRB_POSIX_Asynch_Write_Dgram_Result * wd_result =
    ACE_dynamic_cast (TRB_POSIX_Asynch_Write_Dgram_Result*, result);

  ACE_ASSERT (wd_result != 0);

  ACE_HANDLE handle = result->aio_fildes;

  struct msghdr * msg_hdr = wd_result->msghdr();

  int     aio_error = 0;
  ssize_t aio_return = 0;

  if (msg_hdr->msg_iovlen > 0 && 
      wd_result->bytes_to_write() > 0)
    {
      aio_return =  ACE_OS::sendmsg(handle,
                                    msg_hdr,
                                    wd_result->flags());
  
      if (aio_return < 0)
        {
          aio_return = 0;
          aio_error = ACE_OS::last_error ();
        }
    }
  else
    {
      aio_error = EINVAL;
    }

  if (aio_error == EWOULDBLOCK)
    return 0;

  result->set_ready (aio_return, aio_error, 0);
  return 1; // we do not know if next op is possible
}

int
TRB_POSIX_AIO_Select_Provider::do_accept (TRB_POSIX_Asynch_Result *result)
{
  TRB_POSIX_Asynch_Accept_Result * accept_result =
    ACE_dynamic_cast (TRB_POSIX_Asynch_Accept_Result*, result);

  ACE_ASSERT (accept_result != 0);

  ACE_HANDLE listen_handle = result->aio_fildes;

  int aio_error = 0;
  
  do
    {
      ACE_HANDLE new_handle = ACE_OS::accept (listen_handle, 0, 0);

      if (new_handle != ACE_INVALID_HANDLE)
        {
          // Store the new handle.
          ACE::clr_flags (new_handle, ACE_NONBLOCK);
          accept_result->accept_handle (new_handle);
        }
      else
          aio_error = ACE_OS::last_error ();

    }
  while (aio_error == EINTR);
    
  if (aio_error == EWOULDBLOCK)  
    return 0;

  accept_result->set_ready (0, aio_error, 0);

  return 2;  // next accept is possible
}

int
TRB_POSIX_AIO_Select_Provider::do_connect (TRB_POSIX_Asynch_Result *result)
{  
  TRB_POSIX_Asynch_Connect_Result * connect_result =
    ACE_dynamic_cast (TRB_POSIX_Asynch_Connect_Result*, result);

  ACE_ASSERT (connect_result != 0);

  ACE_HANDLE handle = result->aio_fildes;

  struct pollfd pfd;
  
  pfd.fd      = handle;
  pfd.events  = (POLLIN|POLLOUT);
  pfd.revents = 0;

  int rc = ::poll (&pfd, 1, 0);

  if (rc == 0)  // not ready
    return 0;


  if (rc < 0)  // errors
    {
      connect_result->set_ready (0, ACE_OS::last_error (), 0);
      return 1;
    }


  int sockerror  = 0 ;
  int lsockerror = sizeof sockerror;

  ACE_OS::getsockopt (handle,
                      SOL_SOCKET,
                      SO_ERROR,
                      (char*) &sockerror,
                      &lsockerror);

  ACE::clr_flags (handle, ACE_NONBLOCK);

  connect_result->set_ready (0, sockerror, 0);
  return 2;   // next op with new handle is possible
}

#endif /* ACE_HAS_AIO_CALLS */
