/*********************************************************************
** Copyright (C) 2003 Terabit Pty Ltd.  All rights reserved.
**
** This file is part of the POSIX-Proactor module.
**
** This file may be distributed and used only under the terms of the  
** Terabit Public License as defined by Terabit Pty Ltd of Australia   
** and appearing in the file tlicense.txt included in the packaging of
** this module and available at http://www.terabit.com.au/license.php.
**
** Contact support@terabit.com.au if any conditions of this licensing
** are not clear to you.
**********************************************************************/

/* -*- C++ -*- */
// $Id:  $

#include "TProactor/POSIX_AIO_Strategies.h"

#if defined (ACE_HAS_AIO_CALLS) || defined(ACE_HAS_AIO_EMULATION)

#include "ace/ACE.h"
#include "ace/Log_Msg.h"
#include "TProactor/POSIX_AIO_Provider.h"
#include "TProactor/POSIX_AIO_Processor.h"

//----------------------------------------------------------
#include "ace/Version.h"

#if ((ACE_MAJOR_VERSION == 5) && (ACE_MINOR_VERSION ==3))

#include "ace/OS.h"

#else  // we assume ACE 5.4.X

#include "ace/OS_NS_stdio.h"

#endif  //((ACE_MAJOR_VERSION == 5) && (ACE_MINOR_VERSION ==3))
//----------------------------------------------------------

/*
 ***************************************************************
 * Base Wait Strategy
 ***************************************************************
 */
TRB_POSIX_AIO_Wait_Strategy::
TRB_POSIX_AIO_Wait_Strategy (TRB_POSIX_AIO_Provider & provider)
  : provider_ (provider)
{
  ;
}

TRB_POSIX_AIO_Wait_Strategy::~TRB_POSIX_AIO_Wait_Strategy (void)
{
}

TRB_POSIX_AIO_Provider &
TRB_POSIX_AIO_Wait_Strategy::provider (void) const
{
  return this->provider_;
}

const TRB_POSIX_AIO_Config &
TRB_POSIX_AIO_Wait_Strategy::config (void) const
{
  return this->provider_.config ();
}

int
TRB_POSIX_AIO_Wait_Strategy::init(void)
{
  return 0;
}

int
TRB_POSIX_AIO_Wait_Strategy::fini(void)
{
  return 0;
}

TRB_POSIX_AIO_Interrupt_Strategy *
TRB_POSIX_AIO_Wait_Strategy::create_interrupt_strategy (void)
{
  return this->config().create_interrupt_strategy (this->provider_);
}

void
TRB_POSIX_AIO_Wait_Strategy::destroy_interrupt_strategy 
                               (TRB_POSIX_AIO_Interrupt_Strategy * i_strategy)
{
  delete i_strategy;
}

/*
 ***************************************************************
 * Base Interrupt Strategy
 ***************************************************************
 */
TRB_POSIX_AIO_Interrupt_Strategy::
TRB_POSIX_AIO_Interrupt_Strategy (TRB_POSIX_AIO_Provider & provider)
  : provider_ (provider)
{
  ;
}

TRB_POSIX_AIO_Interrupt_Strategy::~TRB_POSIX_AIO_Interrupt_Strategy (void)
{
  ;
}

TRB_POSIX_AIO_Provider &
TRB_POSIX_AIO_Interrupt_Strategy::provider (void) const
{
  return this->provider_;
}

const TRB_POSIX_AIO_Config &
TRB_POSIX_AIO_Interrupt_Strategy::config (void) const
{
  return this->provider_.config ();
}

/*
 ***************************************************************
 * Signal_Interrupt_Strategy
 ***************************************************************
 */

TRB_POSIX_AIO_Signal_Interrupt_Strategy::
TRB_POSIX_AIO_Signal_Interrupt_Strategy (TRB_POSIX_AIO_Provider & provider)
  : TRB_POSIX_AIO_Interrupt_Strategy (provider),
    signal_number_ (0)
{
  this->signal_number_ = this->config().interrupt_sig_num();

  sigemptyset (&this->signal_mask_);  
}

TRB_POSIX_AIO_Signal_Interrupt_Strategy::~TRB_POSIX_AIO_Signal_Interrupt_Strategy(void)
{
  ;
}

int 
TRB_POSIX_AIO_Signal_Interrupt_Strategy::init ()
{
  sigemptyset (&this->signal_mask_);  

  if (sigaddset(&this->signal_mask_, this->signal_number_) == -1)
    ACE_ERROR_RETURN
      ((LM_ERROR,
        ACE_LIB_TEXT("(%P | %t) TRB_POSIX_AIO_Signal_Interrupt_Strategy::%p\n"),
        ACE_LIB_TEXT("sigaddset failed")),
        -1);

  return this->setup_signal_handler (this->signal_number_);
}

int 
TRB_POSIX_AIO_Signal_Interrupt_Strategy::fini (void)
{
  return 0;
}

int 
TRB_POSIX_AIO_Signal_Interrupt_Strategy::on_leader_activated (TRB_POSIX_AIO_Processor & aio_processor)
{
  ACE_UNUSED_ARG (aio_processor);

  return this->mask_signals (SIG_UNBLOCK, &this->signal_mask_);
}

int 
TRB_POSIX_AIO_Signal_Interrupt_Strategy::on_leader_deactivated (TRB_POSIX_AIO_Processor & aio_processor)
{
  ACE_UNUSED_ARG (aio_processor);

  return this->mask_signals (SIG_BLOCK, &this->signal_mask_);
}

int 
TRB_POSIX_AIO_Signal_Interrupt_Strategy::interrupt(ACE_thread_t leader_id)
{
  if (ACE_OS::thr_kill (leader_id, this->signal_number_) != 0)
    ACE_ERROR_RETURN((LM_ERROR,
      ACE_LIB_TEXT("(%P|%t) TRB_POSIX_AIO_Signal_Interrupt_Strategy::(%u) %p\n"),
      ACE_static_cast(unsigned int, leader_id),
      ACE_LIB_TEXT("interrupt failed")),
      -1);

  return 0;
}

void 
TRB_POSIX_AIO_Signal_Interrupt_Strategy::sig_handler (int sig_num,
                                                      siginfo_t *,
                                                      ucontext_t *)
{
  ACE_UNUSED_ARG (sig_num);
  //ACE_DEBUG ((LM_DEBUG,
  //            "%N:%l:(%P | %t)::sig_handler received signal: %d\n",
  //             sig_num));
  return;
}

int
TRB_POSIX_AIO_Signal_Interrupt_Strategy::setup_signal_handler (int sig_num) const
{
  struct sigaction reaction;

  sigemptyset (&reaction.sa_mask);   // Nothing else to mask.

  reaction.sa_flags = SA_SIGINFO |SA_NODEFER;    // Realtime flag.

  reaction.sa_sigaction =  // (SIG_DFL);
    ACE_SIGNAL_C_FUNC (TRB_POSIX_AIO_Signal_Interrupt_Strategy::sig_handler); 

  int sigaction_return = ACE_OS::sigaction (sig_num,
                                            &reaction,
                                            0);
  if (sigaction_return == -1)
    ACE_ERROR_RETURN ((LM_ERROR,
      ACE_LIB_TEXT("(%P | %t) TRB_POSIX_AIO_Signal_Interrupt_Strategy::%p\n"),
      ACE_LIB_TEXT("setup_signal_handler failed")),
      -1);
  return 0;
}

int
TRB_POSIX_AIO_Signal_Interrupt_Strategy::mask_signals (int how,
                                                       const sigset_t *signals) const
{
  // how = SIG_BLOCK/SIG_UNBLOCK/SIG_SETMASK

  if (ACE_OS::pthread_sigmask (how, signals, 0) != 0)
    ACE_ERROR_RETURN ((LM_ERROR,
      ACE_LIB_TEXT("(%P | %t) TRB_POSIX_AIO_Signal_Interrupt_Strategy::%p\n"),
      ACE_LIB_TEXT("mask_signals failed")),
       -1);
  return 0;
}

/*
 ***************************************************************
 * Pipe_Interrupt_Strategy
 ***************************************************************
 */
TRB_POSIX_AIO_Pipe_Interrupt_Strategy::TRB_POSIX_AIO_Interrupt_Result::
TRB_POSIX_AIO_Interrupt_Result 
  (TRB_POSIX_AIO_Pipe_Interrupt_Strategy & owner)
  : TRB_POSIX_Asynch_Read_Stream_Result 
    (owner,                          // ACE_Hanlder
     owner.pipe_.read_handle(),      // ACE_HANDLE
     owner.message_block_,           // message block
     sizeof(owner.read_buffer_),     // bytes_to_read
     0,                              // act,
     ACE_INVALID_HANDLE,             // Event
     0,                              // priority,
     0)                              // default signal_number
{
  // Prevent Result from deletion and passing 
  // to the Completion Dispatcher.
  // It will be dispatched internally in AIO_Processor.
  this->flags_ = INTERNAL_RESULT;
}
    
TRB_POSIX_AIO_Pipe_Interrupt_Strategy::TRB_POSIX_AIO_Interrupt_Result::
~TRB_POSIX_AIO_Interrupt_Result (void)
{
}

void
TRB_POSIX_AIO_Pipe_Interrupt_Strategy::TRB_POSIX_AIO_Interrupt_Result::
complete (size_t bytes_transferred,
          int success,
          const void *completion_key,
          u_long error)
{
  // bytes_transferred,completion key, error 
  // are already stored in this result.
  // They are passed as parameters for compatibility with
  // old version.

  ACE_UNUSED_ARG (success);
  ACE_UNUSED_ARG (completion_key);
  ACE_UNUSED_ARG (error);
  ACE_UNUSED_ARG (bytes_transferred);

  TRB_POSIX_AIO_Pipe_Interrupt_Strategy & owner =
     ACE_static_cast(TRB_POSIX_AIO_Pipe_Interrupt_Strategy &, this->handler_);


  owner.bytes_rd_ += bytes_transferred;

  ++(owner.num_rd_);

  // ACE_DEBUG((LM_DEBUG,"PIPE RD=%d\n", (int)owner.num_rd_));

  // this->reuse ();
}


TRB_POSIX_AIO_Pipe_Interrupt_Strategy::
TRB_POSIX_AIO_Pipe_Interrupt_Strategy (TRB_POSIX_AIO_Provider & provider)
  : TRB_POSIX_AIO_Interrupt_Strategy (provider),
    pipe_ (),
    message_block_ (read_buffer_, sizeof(read_buffer_)),
    notify_result_ (0),
    num_wr_  (0),
    bytes_wr_(0),
    num_rd_  (0),
    bytes_rd_(0)
{
  ;
}

TRB_POSIX_AIO_Pipe_Interrupt_Strategy::~TRB_POSIX_AIO_Pipe_Interrupt_Strategy(void)
{
  // closing twice shoud be OK
  this->fini ();

  ACE_TCHAR buf [512];

  ACE_OS::snprintf
            ( buf,
              sizeof(buf)-1,
              ACE_LIB_TEXT("TRB_POSIX_AIO_Pipe_Interrupt_Strategy::fini")
              ACE_LIB_TEXT(" read=%ld(%ld) write=%ld(%ld)\n"),
              this->bytes_rd_,
              this->num_rd_,
              this->bytes_wr_,
              this->num_wr_);

  buf[sizeof(buf)-1] = '\0';

  ACE_DEBUG ((LM_DEBUG,buf));

 
//  %Ld sometimes does not work

//  ACE_DEBUG ((LM_DEBUG,
//              ACE_LIB_TEXT("TRB_POSIX_AIO_Pipe_Interrupt_Strategy::fini")
//              ACE_LIB_TEXT(" read=%Ld(%Ld) write=%Ld(%Ld)\n"),
//              this->bytes_rd_,
//              this->num_rd_,
//              this->bytes_wr_,
//              this->num_wr_));
//  

}

int 
TRB_POSIX_AIO_Pipe_Interrupt_Strategy::fini (void)
{
  this->pipe_.close ();

  TRB_POSIX_Asynch_Result::free (this->notify_result_);
  this->notify_result_ = 0;

  return 0;
}

int 
TRB_POSIX_AIO_Pipe_Interrupt_Strategy::init (void)
{
  // Open the pipe.
  int rc = this->pipe_.open ();

  if (rc < 0)
      return -1;

  // Set write side in NONBLOCK mode
  rc = ACE::set_flags (this->pipe_.write_handle (), ACE_NONBLOCK);
  if (rc < 0)
      return -1;

  // Register read handle in AIO_Provider

  rc = this->provider_.register_handle (this->pipe_.read_handle(), // handle
                                        0);                        // completion key
  if (rc < 0)
      return -1;

  // Create the Asynch_Result.
  ACE_NEW_RETURN (this->notify_result_,
                  TRB_POSIX_AIO_Interrupt_Result (*this),
                  -1);

  return 0;
}

int 
TRB_POSIX_AIO_Pipe_Interrupt_Strategy::on_leader_activated (TRB_POSIX_AIO_Processor & aio_processor)
{
  ACE_UNUSED_ARG (aio_processor);

  if ( this->notify_result_->reuse() > 0 )
    {
      // This means that result is INITIALIZED state,
      // so start notification read
      aio_processor.start_special_aio (this->notify_result_);
    }
  return 0;
}

int 
TRB_POSIX_AIO_Pipe_Interrupt_Strategy::on_leader_deactivated (TRB_POSIX_AIO_Processor & aio_processor)
{
  ACE_UNUSED_ARG (aio_processor);
  return 0;
}

int 
TRB_POSIX_AIO_Pipe_Interrupt_Strategy::interrupt(ACE_thread_t leader_id)
{
  ACE_UNUSED_ARG (leader_id);

  char char_send = 0;
  int ret_val = ACE::send (this->pipe_.write_handle(),
                           &char_send,
                           sizeof (char_send));

  //?? fdatasync (this->pipe_.write_handle());

  if (ret_val < 0)
    {
      if (errno != EWOULDBLOCK)
        ACE_ERROR_RETURN ((LM_ERROR,
          ACE_LIB_TEXT("(%P | %t) TRB_POSIX_AIO_Pipe_Interrupt_Strategy::%p\n"),
          ACE_LIB_TEXT("interrupt failed")),
          -1);
    }
  else
    {
      this->bytes_wr_ += ret_val;
      this->num_wr_ ++;

      // ACE_DEBUG((LM_DEBUG,"PIPE WR=%d\n", (int)this->num_wr_));
    }

  return 0;
}
  

#endif /* ACE_HAS_AIO_CALLS */
