/* -*- C++ -*- */
// POSIX_Asynch_IO.cpp,v 4.46 2002/11/09 23:51:29 shuston Exp

#include "TProactor/POSIX_Asynch_IO.h"

#if defined (ACE_HAS_AIO_CALLS) || defined(ACE_HAS_AIO_EMULATION)

#include "TProactor/Proactor.h"
#include "ace/Message_Block.h"
#include "ace/INET_Addr.h"
#include "ace/UNIX_Addr.h"
#include "ace/Malloc_Base.h"


//----------------------------------------------------------
#include "ace/Version.h"

#if ((ACE_MAJOR_VERSION == 5) && (ACE_MINOR_VERSION ==3))

#include "ace/OS.h"

#else  // we assume ACE 5.4.X

#include "ace/OS_NS_sys_socket.h"
#include "ace/OS_NS_sys_stat.h"

#endif //((ACE_MAJOR_VERSION == 5) && (ACE_MINOR_VERSION ==3))
//----------------------------------------------------------


#include "TProactor/POSIX_Proactor.h"

#if !defined (__ACE_INLINE__)
#include "TProactor/POSIX_Asynch_IO.inl"
#endif /* __ACE_INLINE__ */

/*
 *******************************************************************
 * TRB_POSIX_Asynch_Waiting_List
 *******************************************************************
 */
TRB_POSIX_Asynch_Waiting_List::TRB_POSIX_Asynch_Waiting_List (void)
  : read_marker_(0),
    cursor_(0)
{
}

TRB_POSIX_Asynch_Waiting_List::~TRB_POSIX_Asynch_Waiting_List (void)
{
}

void
TRB_POSIX_Asynch_Waiting_List::add_result(TRB_POSIX_Asynch_Result *result)
{
  if (result->is_read_family())
    {
      this->insert_tail (result);

      if (this->read_marker_ == 0)
        this->read_marker_ = result;
    }
  else
    {
      // If read_marker_  is 0,
      // then  it is also works OK 
      // as it means insert it before <head_>, i.e., at tail.

      this->insert_element (result, 1, this->read_marker_);
    }
}

void
TRB_POSIX_Asynch_Waiting_List::remove_result (TRB_POSIX_Asynch_Result *result)
{
  if ((result->is_read_family()) && (this->read_marker_ == result))
    this->move_read_marker();  // move read marker forward
  
  if (this->cursor_ == result)
    get_next();  // move cursor forward 
   
  this->remove (result);
}

TRB_POSIX_Asynch_Result *  
TRB_POSIX_Asynch_Waiting_List::get_current ()
{
  return this->cursor_;
}

TRB_POSIX_Asynch_Result *  
TRB_POSIX_Asynch_Waiting_List::get_current_and_advance ()
{
  TRB_POSIX_Asynch_Result * result = this->cursor_;
  
  this->get_next();

  return result;
}

TRB_POSIX_Asynch_Result *
TRB_POSIX_Asynch_Waiting_List::get_first(void)
{
  this->cursor_ = this->head_;
  
  return this->get_next();
}

TRB_POSIX_Asynch_Result *
TRB_POSIX_Asynch_Waiting_List::get_last(void)
{
  this->cursor_ = this->head_;
  
  return this->get_prev();
}


TRB_POSIX_Asynch_Result *  
TRB_POSIX_Asynch_Waiting_List::get_next ()
{
  if (this->cursor_ != 0 )
    {
        this->cursor_ = this->cursor_->next_;

        if (this->cursor_ == this->head_)
          this->cursor_ = 0;
    }

  return this->cursor_;
}

TRB_POSIX_Asynch_Result *  
TRB_POSIX_Asynch_Waiting_List::get_prev ()
{
  if (this->cursor_ != 0 )
    {
        this->cursor_ = this->cursor_->prev_;

        if (this->cursor_ == this->head_)
          this->cursor_ = 0;
    }

  return this->cursor_;
}

void  
TRB_POSIX_Asynch_Waiting_List::move_read_marker ()
{
  if (this->read_marker_ != 0 )
    {
        this->read_marker_ = this->read_marker_->next_;

        if (this->read_marker_ == this->head_)
          this->read_marker_ = 0;
    }
}

// **************************************************************
//  TRB_POSIX_Asynch_Result_List and Iterator
// **************************************************************
  
TRB_POSIX_Asynch_Result_List::TRB_POSIX_Asynch_Result_List (void)
{
}

TRB_POSIX_Asynch_Result_List::~TRB_POSIX_Asynch_Result_List (void)
{
  // should we allow to delete all items in base destructor?
  TRB_POSIX_Asynch_Result_List_Iterator itr( *this);

  for (;;)
   {
      TRB_POSIX_Asynch_Result *  result = itr.get_current_and_advance ();
      
      if (result == 0)
        break;
      
      TRB_POSIX_Asynch_Result::free(result);
  }
}


TRB_POSIX_Asynch_Result_List_Iterator::
TRB_POSIX_Asynch_Result_List_Iterator (const TRB_POSIX_Asynch_Result_List & list)
  : ACE_Double_Linked_List_Iterator<TRB_POSIX_Asynch_Result> (list)
{
}

TRB_POSIX_Asynch_Result_List_Iterator::
~TRB_POSIX_Asynch_Result_List_Iterator (void)
{
}

TRB_POSIX_Asynch_Result * 
TRB_POSIX_Asynch_Result_List_Iterator::get_current_and_advance ()
{
  TRB_POSIX_Asynch_Result * result = this->next ();

  this->advance();

  return result;
}

// **************************************************************
//  TRB_POSIX_Asynch_Result
// **************************************************************

ACE_HANDLE
TRB_POSIX_Asynch_Result::event (void) const
{
  return ACE_INVALID_HANDLE;
}

u_long
TRB_POSIX_Asynch_Result::offset (void) const
{
  return this->aio_offset;
}

u_long
TRB_POSIX_Asynch_Result::offset_high (void) const
{
  //
  // @@ Support aiocb64??
  //
  ACE_NOTSUP_RETURN (0);
}

int
TRB_POSIX_Asynch_Result::priority (void) const
{
  return this->aio_reqprio;
}

int
TRB_POSIX_Asynch_Result::signal_number (void) const
{
  return this->aio_sigevent.sigev_signo;
}

TRB_Handler&
TRB_POSIX_Asynch_Result::handler(void)
{
  return this->handler_;
 }
 
int
TRB_POSIX_Asynch_Result::post_completion (TRB_Proactor_Impl *proactor_impl)
{
  // Get to the platform specific implementation.
  TRB_POSIX_Proactor *posix_proactor = ACE_dynamic_cast (TRB_POSIX_Proactor *,
                                                         proactor_impl);

  if (posix_proactor == 0)
    ACE_ERROR_RETURN ((LM_ERROR, "Dynamic cast to POSIX Proactor failed\n"), -1);

  // Post myself.
  return posix_proactor->post_completion (this);
}

TRB_POSIX_Asynch_Result::~TRB_POSIX_Asynch_Result (void)
{
}

TRB_POSIX_Asynch_Result::TRB_POSIX_Asynch_Result (TRB_Handler &handler,
                                                  const void* act,
                                                  ACE_HANDLE event,
                                                  u_long offset,
                                                  u_long offset_high,
                                                  int priority,
                                                  int signal_number)
  : TRB_Asynch_Result_Impl (),
    aiocb (),
    handler_ (handler),
    next_(0),
    prev_(0),
    allocator_(0),
    chunk_(0),
    cont_ (0)
{
  this->act_ = act;

  aio_offset = offset;
  aio_reqprio = priority;
  aio_sigevent.sigev_signo = signal_number;

  //default notification method
  aio_sigevent.sigev_notify = SIGEV_NONE;

  // Event is not used on POSIX.
  ACE_UNUSED_ARG (event);

  //
  // @@ Support offset_high with aiocb64.
  //
  ACE_UNUSED_ARG (offset_high);

  // Other fields in the <aiocb> will be initialized by the
  // subclasses.
}

TRB_Handler TRB_POSIX_Asynch_Result::dummy_handler;

TRB_POSIX_Asynch_Result::TRB_POSIX_Asynch_Result ()
  : TRB_Asynch_Result_Impl (),
    aiocb (),
    handler_ (dummy_handler),
    next_(0),
    prev_(0),
    allocator_(0),
    chunk_(0),
    cont_ (0)
{
  // This constructor is used only for
  // ACE_Double_Linked_List<>
  // So it marked as DUMMY_RESULT 

  this->flags_ = DUMMY_RESULT;
  this->act_ = 0;

  aio_offset = 0;
  aio_reqprio = 0;
  aio_sigevent.sigev_signo = 0;

  //default notification method
  aio_sigevent.sigev_notify = SIGEV_NONE;
}

void
TRB_POSIX_Asynch_Result::complete (size_t bytes_transferred,
                                   int success,
                                   const void *completion_key,
                                   u_long error)
{
  // bytes_transferred,completion key, error 
  // are already stored in this result.
  // They are passed as parameters for compatibility with
  // old version.

  ACE_UNUSED_ARG (success);
  ACE_UNUSED_ARG (completion_key);
  ACE_UNUSED_ARG (error);
  ACE_UNUSED_ARG (bytes_transferred);
}

void
TRB_POSIX_Asynch_Result::free(TRB_POSIX_Asynch_Result * result)
{
  if (result == 0)
      return;

  ACE_Allocator * allocator = result->allocator_;
  void * chunk = result->chunk_;

  if (allocator == 0)
    {
      delete result;
      return;
    }

  result->~TRB_POSIX_Asynch_Result();

  //allocator->free (ACE_static_cast (void*,result));
  allocator->free (chunk);

  return;
}


// ****************************************************************

int
TRB_POSIX_Asynch_Operation::open (TRB_Handler &handler,
                                  ACE_HANDLE handle,
                                  const void *completion_key,
                                  TRB_Proactor *proactor)
{
  this->proactor_ = proactor;
  this->handler_ = &handler;
  this->handle_ = handle;

  // Grab the handle from the <handler> if <handle> is invalid
  if (this->handle_ == ACE_INVALID_HANDLE)
    this->handle_ = this->handler_->handle ();
  if (this->handle_ == ACE_INVALID_HANDLE)
    return -1;

#if 0
  // @@ If <proactor> is 0, let us not bother about getting this
  // Proactor, we have already got the specific implementation
  // Proactor.

  // If no proactor was passed
  if (this->proactor_ == 0)
    {
      // Grab the proactor from the <Service_Config> if
      // <handler->proactor> is zero
      this->proactor_ = this->handler_->proactor ();
      if (this->proactor_ == 0)
        this->proactor_ = TRB_Proactor::instance();
    }
#endif /* 0 */

  return this->posix_proactor_->register_handle(handle, completion_key);
}  

int
TRB_POSIX_Asynch_Operation::cancel (void)
{
  if (!posix_proactor_)
    return -1;

  return posix_proactor_->cancel_aio (this->handle_);
}

int
TRB_POSIX_Asynch_Operation::close (void)
{
  if (this->handle_ != ACE_INVALID_HANDLE)
    {
      if (this->posix_proactor_ != 0)
        this->posix_proactor_->unregister_handle (this->handle_);

      this->handle_ = ACE_INVALID_HANDLE;
    }

  return 0;
}

TRB_Proactor *
TRB_POSIX_Asynch_Operation::proactor (void) const
{
  return this->proactor_;
}

TRB_POSIX_Proactor *
TRB_POSIX_Asynch_Operation::posix_proactor (void) const
{
  return this->posix_proactor_;
}

TRB_POSIX_Asynch_Operation::~TRB_POSIX_Asynch_Operation (void)
{
}

TRB_POSIX_Asynch_Operation::TRB_POSIX_Asynch_Operation (TRB_POSIX_Proactor *posix_proactor)
  : TRB_Asynch_Operation_Impl (),
    posix_proactor_ (posix_proactor),
    handler_ (0),
    handle_  (ACE_INVALID_HANDLE)
{
}

// *********************************************************************

size_t
TRB_POSIX_Asynch_Read_Stream_Result::bytes_to_read (void) const
{
  return this->aio_nbytes;
}

ACE_Message_Block &
TRB_POSIX_Asynch_Read_Stream_Result::message_block (void) const
{
  return this->message_block_;
}

ACE_HANDLE
TRB_POSIX_Asynch_Read_Stream_Result::handle (void) const
{
  return this->aio_fildes;
}

TRB_POSIX_Asynch_Read_Stream_Result::TRB_POSIX_Asynch_Read_Stream_Result (TRB_Handler &handler,
                                                                          ACE_HANDLE handle,
                                                                          ACE_Message_Block &message_block,
                                                                          size_t bytes_to_read,
                                                                          const void* act,
                                                                          ACE_HANDLE event,
                                                                          int priority,
                                                                          int signal_number)
  : TRB_Asynch_Result_Impl (),
    TRB_Asynch_Read_Stream_Result_Impl (),
    TRB_POSIX_Asynch_Result (handler, act, event, 0, 0, priority, signal_number),
    message_block_ (message_block)
{
  this->aio_fildes = handle;
  this->aio_buf = message_block.wr_ptr ();
  this->aio_nbytes = bytes_to_read;
  ACE_UNUSED_ARG (event);
}

void
TRB_POSIX_Asynch_Read_Stream_Result::complete (size_t bytes_transferred,
                                               int success,
                                               const void *completion_key,
                                               u_long error)
{
  // bytes_transferred,completion key, error 
  // are already stored in this result.
  // They are passed as parameters for compatibility with
  // old version.

  ACE_UNUSED_ARG (success);
  ACE_UNUSED_ARG (completion_key);
  ACE_UNUSED_ARG (error);

  // Appropriately move the pointers in the message block.
  this->message_block_.wr_ptr (bytes_transferred);

  // Create the interface result class.
  TRB_Asynch_Read_Stream::Result result (this);

  // Call the application handler.
  this->handler_.handle_read_stream (result);
}

TRB_POSIX_Asynch_Read_Stream_Result::~TRB_POSIX_Asynch_Read_Stream_Result (void)
{
}

// ************************************************************

TRB_POSIX_Asynch_Read_Stream::TRB_POSIX_Asynch_Read_Stream (TRB_POSIX_Proactor  *posix_proactor)
  : TRB_Asynch_Operation_Impl (),
    TRB_Asynch_Read_Stream_Impl (),
    TRB_POSIX_Asynch_Operation (posix_proactor)
{
}

int
TRB_POSIX_Asynch_Read_Stream::read (ACE_Message_Block &message_block,
                                    size_t bytes_to_read,
                                    const void *act,
                                    int priority,
                                    int signal_number)
{
  size_t space = message_block.space ();
  if (bytes_to_read > space)
     bytes_to_read=space;

  if (bytes_to_read == 0)
    {
      errno = ENOSPC;
      return -1;
    }

  // Create the Asynch_Result.
  TRB_POSIX_Proactor *proactor = this->posix_proactor ();

  // Atomic allocation  result and start AIO
  int ret_val = proactor->start_read_stream (*this->handler_,
                                             this->handle_,
                                             message_block,
                                             bytes_to_read,
                                             act,
                                             proactor->get_handle (),
                                             priority,
                                             signal_number);

  return ret_val;
}

TRB_POSIX_Asynch_Read_Stream::~TRB_POSIX_Asynch_Read_Stream (void)
{
}

// *********************************************************************

size_t
TRB_POSIX_Asynch_Write_Stream_Result::bytes_to_write (void) const
{
  return this->aio_nbytes;
}

ACE_Message_Block &
TRB_POSIX_Asynch_Write_Stream_Result::message_block (void) const
{
  return this->message_block_;
}

ACE_HANDLE
TRB_POSIX_Asynch_Write_Stream_Result::handle (void) const
{
  return this->aio_fildes;
}

TRB_POSIX_Asynch_Write_Stream_Result::TRB_POSIX_Asynch_Write_Stream_Result
  (TRB_Handler &handler,
   ACE_HANDLE handle,
   ACE_Message_Block &message_block,
   size_t bytes_to_write,
   const void* act,
   ACE_HANDLE event,
   int priority,
   int signal_number)
  : TRB_Asynch_Result_Impl (),
    TRB_Asynch_Write_Stream_Result_Impl (),
    TRB_POSIX_Asynch_Result (handler, act, event, 0, 0, priority, signal_number),
    message_block_ (message_block)
{
  this->aio_fildes = handle;
  this->aio_buf = message_block.rd_ptr ();
  this->aio_nbytes = bytes_to_write;
  ACE_UNUSED_ARG (event);
}

void
TRB_POSIX_Asynch_Write_Stream_Result::complete (size_t bytes_transferred,
                                                int success,
                                                const void *completion_key,
                                                u_long error)
{
  // bytes_transferred,completion key, error 
  // are already stored in this result.
  // They are passed as parameters for compatibility with
  // old version.

  ACE_UNUSED_ARG (success);
  ACE_UNUSED_ARG (completion_key);
  ACE_UNUSED_ARG (error);

  // Appropriately move the pointers in the message block.
  this->message_block_.rd_ptr (bytes_transferred);

  // Create the interface result class.
  TRB_Asynch_Write_Stream::Result result (this);

  // Call the application handler.
  this->handler_.handle_write_stream (result);
}

TRB_POSIX_Asynch_Write_Stream_Result::~TRB_POSIX_Asynch_Write_Stream_Result (void)
{
}

// *********************************************************************

TRB_POSIX_Asynch_Write_Stream::TRB_POSIX_Asynch_Write_Stream (TRB_POSIX_Proactor *posix_proactor)
  : TRB_Asynch_Operation_Impl (),
    TRB_Asynch_Write_Stream_Impl (),
    TRB_POSIX_Asynch_Operation (posix_proactor)
{
}

int
TRB_POSIX_Asynch_Write_Stream::write (ACE_Message_Block &message_block,
                                      size_t bytes_to_write,
                                      const void *act,
                                      int priority,
                                      int signal_number)
{
  size_t len = message_block.length ();
  if (bytes_to_write > len)
     bytes_to_write = len;

  if (bytes_to_write == 0)
    ACE_ERROR_RETURN 
      ((LM_ERROR,
        ACE_LIB_TEXT ("TRB_POSIX_Asynch_Write_Stream::write:")
        ACE_LIB_TEXT ("Attempt to write 0 bytes\n")),
      -1);

  TRB_POSIX_Proactor *proactor = this->posix_proactor ();

    // Atomic allocation  result and start AIO
  int ret_val = proactor->start_write_stream (*this->handler_,
                                             this->handle_,
                                             message_block,
                                             bytes_to_write,
                                             act,
                                             proactor->get_handle (),
                                             priority,
                                             signal_number);

  return ret_val;
}

TRB_POSIX_Asynch_Write_Stream::~TRB_POSIX_Asynch_Write_Stream (void)
{
}

// *********************************************************************

TRB_POSIX_Asynch_Read_File_Result::TRB_POSIX_Asynch_Read_File_Result
  (TRB_Handler &handler,
   ACE_HANDLE handle,
   ACE_Message_Block &message_block,
   size_t bytes_to_read,
   const void* act,
   u_long offset,
   u_long offset_high,
   ACE_HANDLE event,
   int priority,
   int signal_number)
  : TRB_Asynch_Result_Impl (),
    TRB_Asynch_Read_Stream_Result_Impl (),
    TRB_Asynch_Read_File_Result_Impl (),
    TRB_POSIX_Asynch_Read_Stream_Result (handler,
                                         handle,
                                         message_block,
                                         bytes_to_read,
                                         act,
                                         event,
                                         priority,
                                         signal_number)
{
  this->aio_offset = offset;
  //
  // @@ Use aiocb64??
  //
  ACE_UNUSED_ARG (offset_high);
}

void
TRB_POSIX_Asynch_Read_File_Result::complete (size_t bytes_transferred,
                                             int success,
                                             const void *completion_key,
                                             u_long error)
{
  // bytes_transferred,completion key, error 
  // are already stored in this result.
  // They are passed as parameters for compatibility with
  // old version.

  ACE_UNUSED_ARG (success);
  ACE_UNUSED_ARG (completion_key);
  ACE_UNUSED_ARG (error);

  // Appropriately move the pointers in the message block.
  this->message_block_.wr_ptr (bytes_transferred);

  // Create the interface result class.
  TRB_Asynch_Read_File::Result result (this);

  // Call the application handler.
  this->handler_.handle_read_file (result);
}

TRB_POSIX_Asynch_Read_File_Result::~TRB_POSIX_Asynch_Read_File_Result (void)
{
}

// *********************************************************************

TRB_POSIX_Asynch_Read_File::TRB_POSIX_Asynch_Read_File (TRB_POSIX_Proactor *posix_proactor)
  : TRB_Asynch_Operation_Impl (),
    TRB_Asynch_Read_Stream_Impl (),
    TRB_Asynch_Read_File_Impl (),
    TRB_POSIX_Asynch_Read_Stream (posix_proactor)
{
}

int
TRB_POSIX_Asynch_Read_File::read (ACE_Message_Block &message_block,
                                  size_t bytes_to_read,
                                  u_long offset,
                                  u_long offset_high,
                                  const void *act,
                                  int priority,
                                  int signal_number)
{
  size_t space = message_block.space ();
  if ( bytes_to_read > space )
     bytes_to_read=space;

  if ( bytes_to_read == 0 )
    ACE_ERROR_RETURN 
      ((LM_ERROR,
        ACE_LIB_TEXT ("TRB_POSIX_Asynch_Read_File::read:")
        ACE_LIB_TEXT ("Attempt to read 0 bytes or no space in the message block\n")),
       -1);

  TRB_POSIX_Proactor *proactor = this->posix_proactor ();

  int ret_val = proactor->start_read_file(*this->handler_,
                                          this->handle_,
                                          message_block,
                                          bytes_to_read,
                                          act,
                                          offset,
                                          offset_high,
                                          posix_proactor ()->get_handle (),
                                          priority,
                                          signal_number);

  return ret_val;
}

TRB_POSIX_Asynch_Read_File::~TRB_POSIX_Asynch_Read_File (void)
{
}

int
TRB_POSIX_Asynch_Read_File::read (ACE_Message_Block &message_block,
                                  size_t bytes_to_read,
                                  const void *act,
                                  int priority,
                                  int signal_number)
{
  return TRB_POSIX_Asynch_Read_Stream::read (message_block,
                                                   bytes_to_read,
                                                   act,
                                                   priority,
                                                   signal_number);
}

// ************************************************************

TRB_POSIX_Asynch_Write_File_Result::TRB_POSIX_Asynch_Write_File_Result
  (TRB_Handler &handler,
   ACE_HANDLE handle,
   ACE_Message_Block &message_block,
   size_t bytes_to_write,
   const void* act,
   u_long offset,
   u_long offset_high,
   ACE_HANDLE event,
   int priority,
   int signal_number)
  : TRB_Asynch_Result_Impl (),
    TRB_Asynch_Write_Stream_Result_Impl (),
    TRB_Asynch_Write_File_Result_Impl (),
    TRB_POSIX_Asynch_Write_Stream_Result (handler,
                                          handle,
                                          message_block,
                                          bytes_to_write,
                                          act,
                                          event,
                                          priority,
                                          signal_number)
{
  this->aio_offset = offset;
  //
  // @@ Support offset_high with aiocb64.
  //
  ACE_UNUSED_ARG (offset_high);
}

void
TRB_POSIX_Asynch_Write_File_Result::complete (size_t bytes_transferred,
                                              int success,
                                              const void *completion_key,
                                              u_long error)
{
  // bytes_transferred,completion key, error 
  // are already stored in this result.
  // They are passed as parameters for compatibility with
  // old version.

  ACE_UNUSED_ARG (success);
  ACE_UNUSED_ARG (completion_key);
  ACE_UNUSED_ARG (error);

  // Appropriately move the pointers in the message block.
  this->message_block_.rd_ptr (bytes_transferred);

  // Create the interface result class.
  TRB_Asynch_Write_File::Result result (this);

  // Call the application handler.
  this->handler_.handle_write_file (result);
}

TRB_POSIX_Asynch_Write_File_Result::~TRB_POSIX_Asynch_Write_File_Result  (void)
{
}

// *********************************************************************

TRB_POSIX_Asynch_Write_File::TRB_POSIX_Asynch_Write_File (TRB_POSIX_Proactor *posix_proactor)
  : TRB_Asynch_Operation_Impl (),
    TRB_Asynch_Write_Stream_Impl (),
    TRB_Asynch_Write_File_Impl (),
    TRB_POSIX_Asynch_Write_Stream (posix_proactor)
{
}

int
TRB_POSIX_Asynch_Write_File::write (ACE_Message_Block &message_block,
                                    size_t bytes_to_write,
                                    u_long offset,
                                    u_long offset_high,
                                    const void *act,
                                    int priority,
                                    int signal_number)
{
  size_t len = message_block.length ();
  if (bytes_to_write > len)
     bytes_to_write = len;

  if (bytes_to_write == 0)
    ACE_ERROR_RETURN 
      ((LM_ERROR,
        ACE_LIB_TEXT ("TRB_POSIX_Asynch_Write_File::write:")
        ACE_LIB_TEXT ("Attempt to write 0 bytes\n")),
      -1);

  TRB_POSIX_Proactor *proactor = this->posix_proactor ();

  int ret_val = proactor->start_write_file(*this->handler_,
                                          this->handle_,
                                          message_block,
                                          bytes_to_write,
                                          act,
                                          offset,
                                          offset_high,
                                          posix_proactor ()->get_handle (),
                                          priority,
                                          signal_number);


  return ret_val;
}

TRB_POSIX_Asynch_Write_File::~TRB_POSIX_Asynch_Write_File (void)
{
}

int
TRB_POSIX_Asynch_Write_File::write (ACE_Message_Block &message_block,
                                    size_t bytes_to_write,
                                    const void *act,
                                    int priority,
                                    int signal_number)
{
  return TRB_POSIX_Asynch_Write_Stream::write (message_block,
                                                     bytes_to_write,
                                                     act,
                                                     priority,
                                                     signal_number);
}

// *********************************************************************


size_t
TRB_POSIX_Asynch_Accept_Result::bytes_to_read (void) const
{
  return this->aio_nbytes;
}

ACE_Message_Block &
TRB_POSIX_Asynch_Accept_Result::message_block (void) const
{
  return this->message_block_;
}

ACE_HANDLE
TRB_POSIX_Asynch_Accept_Result::listen_handle (void) const
{
  // now listen handle in aiocb , not accept
  // this reflects right semantic:
  // operation is performed on listen handle
  // accept_hanlde is the result of operation
  return this->aio_fildes;
}

ACE_HANDLE
TRB_POSIX_Asynch_Accept_Result::accept_handle (void) const
{
  return this->accept_handle_;
}

void
TRB_POSIX_Asynch_Accept_Result::accept_handle (ACE_HANDLE handle) 
{
  ACE_ASSERT ( this->accept_handle_ == ACE_INVALID_HANDLE);
  this->accept_handle_ = handle;
}

TRB_POSIX_Asynch_Accept_Result::TRB_POSIX_Asynch_Accept_Result
  (TRB_Handler &handler,
   ACE_HANDLE listen_handle,
   ACE_HANDLE accept_handle,
   ACE_Message_Block &message_block,
   size_t bytes_to_read,
   const void* act,
   ACE_HANDLE event,
   int priority,
   int signal_number)

  : TRB_Asynch_Result_Impl (),
    TRB_Asynch_Accept_Result_Impl (),
    TRB_POSIX_Asynch_Result (handler, act, event, 0, 0, priority, signal_number),
    message_block_ (message_block),
    accept_handle_ (accept_handle)
{
  this->aio_fildes = listen_handle;
  this->aio_nbytes = bytes_to_read;
}

void
TRB_POSIX_Asynch_Accept_Result::complete (size_t bytes_transferred,
                                          int success,
                                          const void *completion_key,
                                          u_long error)
{
  // bytes_transferred,completion key, error 
  // are already stored in this result.
  // They are passed as parameters for compatibility with
  // old version.

  ACE_UNUSED_ARG (success);
  ACE_UNUSED_ARG (completion_key);
  ACE_UNUSED_ARG (error);

  // Appropriately move the pointers in the message block.
  this->message_block_.wr_ptr (bytes_transferred);

  // Create the interface result class.
  TRB_Asynch_Accept::Result result (this);

  // Call the application handler.
  this->handler_.handle_accept (result);
}

TRB_POSIX_Asynch_Accept_Result::~TRB_POSIX_Asynch_Accept_Result (void)
{
}

// *********************************************************************

TRB_POSIX_Asynch_Accept::TRB_POSIX_Asynch_Accept (TRB_POSIX_Proactor * posix_proactor)
  : TRB_Asynch_Operation_Impl (),
    TRB_Asynch_Accept_Impl (),
    TRB_POSIX_Asynch_Operation (posix_proactor)
{
}

TRB_POSIX_Asynch_Accept::~TRB_POSIX_Asynch_Accept (void)
{
  this->close ();
}


int
TRB_POSIX_Asynch_Accept::accept (ACE_Message_Block &message_block,
                                 size_t bytes_to_read,
                                 ACE_HANDLE accept_handle,
                                 const void *act,
                                 int priority,
                                 int signal_number)
{
  ACE_TRACE (ACE_LIB_TEXT("TRB_POSIX_Asynch_Accept::accept\n"));

  // Sanity check: make sure that enough space has been allocated by
  // the caller.
  size_t address_size = sizeof (sockaddr_in) + sizeof (sockaddr);
  size_t space_in_use = message_block.wr_ptr () - message_block.base ();
  size_t total_size = message_block.size ();
  size_t available_space = total_size - space_in_use;
  size_t space_needed = bytes_to_read + 2 * address_size;

  if (available_space < space_needed)
    ACE_ERROR_RETURN ((LM_ERROR,
                       ACE_LIB_TEXT ("Buffer too small\n")),
                       -1);

  // Create future Asynch_Accept_Result
  TRB_POSIX_Proactor *proactor = this->posix_proactor ();

  int ret_val = proactor->start_accept(*this->handler_,
                                        this->handle_,
                                        accept_handle,
                                        message_block,
                                        bytes_to_read,
                                        act,
                                        proactor->get_handle (),
                                        priority,
                                        signal_number);

  return ret_val;
}



int
TRB_POSIX_Asynch_Accept::close ()
{
  //if (this->handle_ != ACE_INVALID_HANDLE)
  //   ACE_OS::closesocket (this->handle_);
     
  TRB_POSIX_Asynch_Operation::close();

  return 0;
}

// *********************************************************************
const ACE_Addr & 
TRB_POSIX_Asynch_Connect_Result::remote_address () const
{
  return *(this->remote_address_);
}

ACE_HANDLE
TRB_POSIX_Asynch_Connect_Result::connect_handle (void) const
{
  return this->aio_fildes;
}

void TRB_POSIX_Asynch_Connect_Result::connect_handle (ACE_HANDLE handle)
{
  this->aio_fildes = handle;
}


TRB_POSIX_Asynch_Connect_Result::TRB_POSIX_Asynch_Connect_Result
  (TRB_Handler &handler,
   ACE_HANDLE connect_handle,
   const ACE_Addr & remote_sap,
   const void* act,
   ACE_HANDLE event,
   int priority,
   int signal_number)

  : TRB_Asynch_Result_Impl (),
    TRB_Asynch_Connect_Result_Impl (),
    TRB_POSIX_Asynch_Result (handler, act, event, 0, 0, priority, signal_number),
    remote_address_(0)
{
  this->aio_fildes = connect_handle;
  this->aio_nbytes = 0;

  switch (remote_sap.get_type())
    {
    case AF_INET:
      ACE_NEW (this->remote_address_, ACE_INET_Addr);
      break;

    case AF_UNIX:
      ACE_NEW (this->remote_address_, ACE_UNIX_Addr);
      break;

    default:
      ACE_ASSERT(0);
      return;
    }

  remote_address_->set_addr (remote_sap.get_addr (),
                             remote_sap.get_size ());
}

TRB_POSIX_Asynch_Connect_Result::~TRB_POSIX_Asynch_Connect_Result (void)
{
  delete this->remote_address_;
}

void
TRB_POSIX_Asynch_Connect_Result::complete (size_t bytes_transferred,
                                           int success,
                                           const void *completion_key,
                                           u_long error)
{
  // bytes_transferred,completion key, error 
  // are already stored in this result.
  // They are passed as parameters for compatibility with
  // old version.

  ACE_UNUSED_ARG (success);
  ACE_UNUSED_ARG (completion_key);
  ACE_UNUSED_ARG (error);
  ACE_UNUSED_ARG (bytes_transferred);

  // Create the interface result class.
  TRB_Asynch_Connect::Result result (this);

  // Call the application handler.
  this->handler_.handle_connect (result);
}

// *********************************************************************

TRB_POSIX_Asynch_Connect::TRB_POSIX_Asynch_Connect (TRB_POSIX_Proactor * posix_proactor)
  : TRB_Asynch_Operation_Impl (),
    TRB_Asynch_Connect_Impl (),
    TRB_POSIX_Asynch_Operation (posix_proactor)
{
}

TRB_POSIX_Asynch_Connect::~TRB_POSIX_Asynch_Connect (void)
{
  this->close ();
}

int
TRB_POSIX_Asynch_Connect::open (TRB_Handler &handler,
                                ACE_HANDLE handle,
                                const void *completion_key,
                                TRB_Proactor *proactor)
{
  ACE_TRACE (ACE_LIB_TEXT("TRB_POSIX_Asynch_Connect::open\n"));

  //int result =
  TRB_POSIX_Asynch_Operation::open (handler,
                                    handle,
                                    completion_key,
                                    proactor);


  return 0;
}

int
TRB_POSIX_Asynch_Connect::connect (ACE_HANDLE connect_handle,
                                   const ACE_Addr & remote_sap,
                                   const ACE_Addr & local_sap,
                                   int reuse_addr,
                                   const void *act,
                                   int priority,
                                   int signal_number)
{
  ACE_TRACE (ACE_LIB_TEXT("TRB_POSIX_Asynch_Connect::connect\n"));


  TRB_POSIX_Proactor *proactor = this->posix_proactor ();

  int ret_val = this->connect_i (connect_handle,
                                 remote_sap,
                                 local_sap,
                                 reuse_addr);

  if (ret_val == 0)
    ret_val = proactor_->register_handle (connect_handle, 0);
  
  if (ret_val == 0)
    ret_val = proactor->start_connect (*this->handler_,
                                       connect_handle,
                                       remote_sap,
                                       act,
                                       proactor->get_handle (),
                                       priority,
                                       signal_number);



  if (ret_val == 0)
     return 0;

  if (connect_handle != ACE_INVALID_HANDLE)
     ACE_OS::closesocket (connect_handle);

  return -1;
}



int
TRB_POSIX_Asynch_Connect::connect_i (ACE_HANDLE & handle,
                                     const ACE_Addr & remote_sap,
                                     const ACE_Addr & local_sap,
                                     int  reuse_addr)
{
  if (handle == ACE_INVALID_HANDLE)
    {
      int protocol_family = remote_sap.get_type ();

      handle = ACE_OS::socket (protocol_family,
                               SOCK_STREAM,
                               0);
      
      if (handle == ACE_INVALID_HANDLE)
        {
          ACE_ERROR_RETURN ((LM_ERROR,
                       ACE_LIB_TEXT("%N:%l:TRB_POSIX_Asynch_Connect::connect_i: ")
                       ACE_LIB_TEXT(" ACE_OS::socket failed\n")),
                      -1);
        }

      // Reuse the address
      int one = 1;
      if (reuse_addr != 0 &&
           ACE_OS::setsockopt (handle,
                               SOL_SOCKET,
                               SO_REUSEADDR,
                               (const char*) &one,
                               sizeof one) == -1 )
        {
          ACE_ERROR_RETURN ((LM_ERROR,
                       ACE_LIB_TEXT("%N:%l:TRB_POSIX_Asynch_Connect::connect_i: ")
                       ACE_LIB_TEXT(" ACE_OS::setsockopt failed\n")),
                      -1);
        }
    }

  if (local_sap != ACE_Addr::sap_any)
    {
      sockaddr * laddr = ACE_reinterpret_cast (sockaddr *,
                                               local_sap.get_addr ());
      size_t size = local_sap.get_size ();

      if (ACE_OS::bind (handle, laddr, size) == -1)
        {
           ACE_ERROR_RETURN ((LM_ERROR,
                       ACE_LIB_TEXT("%N:%l:TRB_POSIX_Asynch_Connect::connect_i: ")
                       ACE_LIB_TEXT(" ACE_OS::bind failed\n")),
                      -1);
        }
    }

  // set non blocking mode
  if (ACE::set_flags (handle, ACE_NONBLOCK) != 0)
    {
      ACE_ERROR_RETURN ((LM_ERROR,
                         ACE_LIB_TEXT("%N:%l:TRB_POSIX_Asynch_Connect::connect_i: ")
                         ACE_LIB_TEXT(" ACE::set_flags failed\n")),
                        -1);
    }

  return 0;
}


int
TRB_POSIX_Asynch_Connect::cancel (void)
{
  ACE_TRACE (ACE_LIB_TEXT("TRB_POSIX_Asynch_Connect::cancel\n"));
  
  if (this->posix_proactor() == 0 || this->handler_ == 0)
    return -1;

  return posix_proactor_->cancel_aio (this->handler_);
}


int
TRB_POSIX_Asynch_Connect::close (void)
{
  ACE_TRACE (ACE_LIB_TEXT("TRB_POSIX_Asynch_Connect::close\n"));

  //if (this->handle_ != ACE_INVALID_HANDLE)
  //   ACE_OS::closesocket (this->handle_);

  TRB_POSIX_Asynch_Operation::close();

  return 0;
}


// *********************************************************************

ACE_HANDLE
TRB_POSIX_Asynch_Transmit_File_Result::socket (void) const
{
  return this->socket_;
}

ACE_HANDLE
TRB_POSIX_Asynch_Transmit_File_Result::file (void) const
{
  return this->aio_fildes;
}

TRB_Asynch_Transmit_File::Header_And_Trailer *
TRB_POSIX_Asynch_Transmit_File_Result::header_and_trailer (void) const
{
  return this->header_and_trailer_;
}

size_t
TRB_POSIX_Asynch_Transmit_File_Result::bytes_to_write (void) const
{
  return this->aio_nbytes;
}

size_t
TRB_POSIX_Asynch_Transmit_File_Result::bytes_per_send (void) const
{
  return this->bytes_per_send_;
}

u_long
TRB_POSIX_Asynch_Transmit_File_Result::flags (void) const
{
  return this->flags_;
}

TRB_POSIX_Asynch_Transmit_File_Result::TRB_POSIX_Asynch_Transmit_File_Result
  (TRB_Handler &handler,
   ACE_HANDLE socket,
   ACE_HANDLE file,
   TRB_Asynch_Transmit_File::Header_And_Trailer *header_and_trailer,
   size_t bytes_to_write,
   u_long offset,
   u_long offset_high,
   size_t bytes_per_send,
   u_long flags,
   const void *act,
   ACE_HANDLE event,
   int priority,
   int signal_number)

  : TRB_Asynch_Result_Impl (),
    TRB_Asynch_Transmit_File_Result_Impl (),
    TRB_POSIX_Asynch_Result (handler, act, event, offset, offset_high, priority, signal_number),
    socket_ (socket),
    header_and_trailer_ (header_and_trailer),
    bytes_per_send_ (bytes_per_send),
    flags_ (flags)
{
  this->aio_fildes = file;
  this->aio_nbytes = bytes_to_write;
}

void
TRB_POSIX_Asynch_Transmit_File_Result::complete (size_t bytes_transferred,
                                                 int success,
                                                 const void *completion_key,
                                                 u_long error)
{
  // bytes_transferred,completion key, error 
  // are already stored in this result.
  // They are passed as parameters for compatibility with
  // old version.

  ACE_UNUSED_ARG (success);
  ACE_UNUSED_ARG (completion_key);
  ACE_UNUSED_ARG (error);
  ACE_UNUSED_ARG (bytes_transferred);

  // We will not do this because (a) the header and trailer blocks may
  // be the same message_blocks and (b) in cases of failures we have
  // no idea how much of what (header, data, trailer) was sent.
  /*
    if (this->success_ && this->header_and_trailer_ != 0)
    {
    ACE_Message_Block *header = this->header_and_trailer_->header ();
    if (header != 0)
    header->rd_ptr (this->header_and_trailer_->header_bytes ());

    ACE_Message_Block *trailer = this->header_and_trailer_->trailer ();
    if (trailer != 0)
    trailer->rd_ptr (this->header_and_trailer_->trailer_bytes ());
    }
  */

  // Create the interface result class.
  TRB_Asynch_Transmit_File::Result result (this);

  // Call the application handler.
  this->handler_.handle_transmit_file (result);
}

TRB_POSIX_Asynch_Transmit_File_Result::~TRB_POSIX_Asynch_Transmit_File_Result (void)
{
}


// *********************************************************************

/**
 * @class TRB_POSIX_Asynch_Transmit_Handler
 *
 * @brief Auxillary handler for doing <Asynch_Transmit_File> in
 * Unix. <TRB_POSIX_Asynch_Transmit_File> internally uses this.
 *
 * This is a helper class for implementing
 * <TRB_POSIX_Asynch_Transmit_File> in Unix systems.
 */
class TPROACTOR_Export TRB_POSIX_Asynch_Transmit_Handler : public TRB_Handler
{
public:
  /// Constructor. Result pointer will have all the information to do
  /// the file transmission (socket, file, application handler, bytes
  /// to write).
  TRB_POSIX_Asynch_Transmit_Handler 
      (TRB_POSIX_Proactor *posix_proactor,
       TRB_Handler &handler,
       ACE_HANDLE socket,
       ACE_HANDLE file,
       TRB_Asynch_Transmit_File::Header_And_Trailer * header_and_trailer,
       size_t bytes_to_write,
       u_long offset,
       u_long offset_high,
       size_t bytes_per_send,
       u_long flags,
       const void *act,
       ACE_HANDLE event,
       int priority,
       int signal_number);


  /// Destructor.
  virtual ~TRB_POSIX_Asynch_Transmit_Handler (void);

  /// Do the transmission. All the info to do the transmission is in
  /// the <result> member.
  int transmit (void);

protected:
  /// This is called when asynchronous writes from the socket complete.
  virtual void handle_write_stream (const TRB_Asynch_Write_Stream::Result &result);

  /// This is called when asynchronous reads from the file complete.
  virtual void handle_read_file (const TRB_Asynch_Read_File::Result &result);


  /// returns : 
  ///  -1 done with errors
  ///   0 AIO started
  ///   1 all is done OK, nothing more to do 
  int  transmit_header  ();
  int  transmit_trailer ();
  int  initiate_read_file (void);

  void do_callback ();
  void set_completion (u_long error);
  

  /// The asynch result pointer made from the initial transmit file
  /// request.
  TRB_POSIX_Asynch_Transmit_File_Result result_;

  /// Message bloack used to do the transmission.
  ACE_Message_Block mb_;

  enum ACT
  {
    HEADER_ACT  = 1,
    DATA_ACT    = 2,
    TRAILER_ACT = 3
  };


  /// Current offset of the file being transmitted.
  size_t file_offset_;

  /// Total size of the file.
  size_t file_size_;

  /// Number of bytes transferred on the stream.
  size_t bytes_transferred_;

  /// To read from the file to be transmitted.
  TRB_POSIX_Asynch_Read_File rf_;

  /// Write stream to write the header, trailer and the data.
  TRB_POSIX_Asynch_Write_Stream ws_;
};

// ************************************************************

// Constructor.
TRB_POSIX_Asynch_Transmit_Handler::TRB_POSIX_Asynch_Transmit_Handler
      (TRB_POSIX_Proactor * posix_proactor,
       TRB_Handler & handler,
       ACE_HANDLE socket,
       ACE_HANDLE file,
       TRB_Asynch_Transmit_File::Header_And_Trailer * header_and_trailer,
       size_t bytes_to_write,
       u_long offset,
       u_long offset_high,
       size_t bytes_per_send,
       u_long flags,
       const void *act,
       ACE_HANDLE event,
       int priority,
       int signal_number)
  : result_ (handler,
             socket,
             file,
             header_and_trailer,
             bytes_to_write,
             offset,
             offset_high,
             bytes_per_send,
             flags,
             act,
             event,
             priority,
             signal_number),
    mb_  (bytes_per_send),
    file_offset_ (offset),
    file_size_ (ACE_OS::filesize (file)),
    bytes_transferred_ (0),
    rf_ (posix_proactor),
    ws_ (posix_proactor)
{
  // Allocate memory for the message block.
  //ACE_NEW (this->mb_,
  //         ACE_Message_Block (this->result_->bytes_per_send ()
  //                            + 1));

  // Init the file size.
  //file_size_ = ACE_OS::filesize (this->result_->file ());
}

// Destructor.
TRB_POSIX_Asynch_Transmit_Handler::~TRB_POSIX_Asynch_Transmit_Handler (void)
{
  //delete result_;
  //mb_->release ();
}

void
TRB_POSIX_Asynch_Transmit_Handler::set_completion (u_long error)
{
    this->result_.set_ready (this->bytes_transferred_,
                             error,
                             result_.completion_key());
}

void
TRB_POSIX_Asynch_Transmit_Handler::do_callback ()
{
  ACE_SEH_TRY
  {
    if (this->result_.error() != 0)  // Failure.
    {
        ACE_ERROR ((LM_ERROR, "Asynch_Transmit_File failed.\n"));
    }


    this->result_.complete (this->result_.bytes_transferred(),
                            this->result_.success(),         // Failure.
                            this->result_.completion_key() , // @@ Completion key.
                            this->result_.error());          // @@ Error no.
  }
  ACE_SEH_FINALLY
  {
     // the result pointer also.
     delete this;
  }
}
// Do the transmission.
// Initiate transmitting the header. When that completes
// handle_write_stream will be called, there start transmitting the file.
int
TRB_POSIX_Asynch_Transmit_Handler::transmit (void)
{
  // No proactor is given for the <open>'s. Because we are using the
  // concrete implementations of the  Asynch_Operations, and we have
  // already given them the specific proactor, so they wont need the
  // general <proactor> interface pointer.

  // Open Asynch_Read_File.
  if (this->rf_.open (*this,
                      this->result_.file (),
                      0,
                      0) == -1)
    ACE_ERROR_RETURN ((LM_ERROR,
                       "TRB_Asynch_Transmit_Handler:read_file open failed\n"),
                      -1);

  // Open Asynch_Write_Stream.
  if (this->ws_.open (*this,
                      this->result_.socket (),
                      0,
                      0) == -1)
    ACE_ERROR_RETURN ((LM_ERROR,
                       "TRB_Asynch_Transmit_Handler:write_stream open failed\n"),
                      -1);

  result_.set_started();

  return this->transmit_header ();

}

void
TRB_POSIX_Asynch_Transmit_Handler::handle_write_stream (const TRB_Asynch_Write_Stream::Result &result)
{
  // Update bytes transferred so far.
  this->bytes_transferred_ += result.bytes_transferred ();

  // Check the success parameter.
  if (result.success () == 0)
    {
       set_completion (result.error());
       this->do_callback ();
       return;
    }
 
  // Write stream successful.

  // Partial write to socket.
  int unsent_data = result.bytes_to_write () - result.bytes_transferred ();
  if (unsent_data != 0)
    {
      // this is ok!!
      //ACE_DEBUG ((LM_DEBUG,
      //            "%N:%l:Partial write to socket: Asynch_write called again\n"));

      // Duplicate the message block and retry remaining data
      if (this->ws_.write (result.message_block (),
                           unsent_data,
                           result.act (),
                           this->result_.priority (),
                           this->result_.signal_number ()) == -1)
        {

          // @@ Handle this error.
          u_long err = errno ? errno : EFAULT;
          set_completion (err);
          do_callback ();
          return;
        }

      // @@ Handling *partial write* to a socket.  Let us not continue
      // further before this write finishes. Because proceeding with
      // another read and then write might change the order of the
      // file transmission, because partial write to the stream is
      // always possible.
      return;
    }

  // Not a partial write. A full write.

  // Check ACT to see what was sent.
  ACT act = (ACT) ACE_reinterpret_cast (int, 
                                  ACE_const_cast (void*, result.act ()));

  switch (act)
    {
    case TRAILER_ACT:
      set_completion (0);
      break ;  // all done

    case HEADER_ACT:
    case DATA_ACT:
      // If header/data was sent, initiate the file data transmission.
      if (initiate_read_file () == 0)
      {
          return ;
      }
      break;

    default:
      // @@ Handle this error.
      ACE_ERROR ((LM_ERROR,
          "Error:TRB_Asynch_Transmit_Handler::handle_write_stream::Unexpected act\n"));

      set_completion (EFAULT);
      break;
    }
    do_callback ();
}

void
TRB_POSIX_Asynch_Transmit_Handler::handle_read_file (const TRB_Asynch_Read_File::Result &result)
{
  // Failure.
  if (result.success () == 0)
    {
       set_completion (result.error());
       this->do_callback ();
       return;
    }

  // Read successful.
  if (result.bytes_transferred () == 0)
  {
      if (transmit_trailer () != 0)
      {
          do_callback ();
      }
      return;
  }

  // Increment offset.
  this->file_offset_ += result.bytes_transferred ();

  // Write data to network.
  if (this->ws_.write (result.message_block (),
                       result.bytes_transferred (),
                       (void *) DATA_ACT,
                       this->result_.priority (),
                       this->result_.signal_number ()) == -1)
    {
      // @@ Handle this error.
      ACE_ERROR ((LM_ERROR,
          "Error:TRB_Asynch_Transmit_File : write to the stream failed\n"));

      u_long err = errno ? errno : EFAULT;
      set_completion (err);
      do_callback ();
      return;
    }
}

int
TRB_POSIX_Asynch_Transmit_Handler::transmit_header (void)
{
    TRB_Asynch_Transmit_File::Header_And_Trailer * header_and_trailer =
        this->result_.header_and_trailer();

    if (header_and_trailer == 0)
    {
        return initiate_read_file ();
    }

    ACE_Message_Block * mb = header_and_trailer->header();
    size_t nbytes = header_and_trailer->header_bytes ();

    if (mb == 0 || nbytes == 0)
    {
        return initiate_read_file ();
    }

    if (this->ws_.write (*mb,
                          nbytes,
                          (void *) HEADER_ACT,
                          this->result_.priority (),
                          this->result_.signal_number ()) == -1)
    {
        ACE_ERROR((LM_ERROR,
           ACE_TEXT("Error:Asynch_Transmit_Handler:write_stream writing header failed\n")));

        u_long err = errno ? errno : EFAULT;
        set_completion (err);
        return -1; // done with ERRORS
    }
    return 0;   // started AIO
}

int
TRB_POSIX_Asynch_Transmit_Handler::transmit_trailer (void)
{
    TRB_Asynch_Transmit_File::Header_And_Trailer * header_and_trailer =
        this->result_.header_and_trailer();

    if (header_and_trailer == 0)
    {
        set_completion (0);
        return 1;   // done OK
    }

    ACE_Message_Block * mb = header_and_trailer->trailer ();
    size_t nbytes = header_and_trailer->trailer_bytes ();

    if (mb == 0 || nbytes == 0)
    {
        set_completion (0);
        return 1;  // done OK
    }

    if (this->ws_.write (*mb,
                         nbytes,
                         (void *) TRAILER_ACT,
                         this->result_.priority (),
                         this->result_.signal_number ()) == -1)
    {
        ACE_ERROR((LM_ERROR,
            ACE_TEXT("Error:Asynch_Transmit_Handler:write_stream writing trailer failed\n")));

        u_long err = errno ? errno : EFAULT;
        set_completion (err);
        return -1; // done with ERRORS
    }
    return 0;   // started AIO
}

int
TRB_POSIX_Asynch_Transmit_Handler::initiate_read_file (void)
{
    // Is there something to read.
    if (this->file_offset_ >= this->file_size_)
    {
        return transmit_trailer (); // File is sent. Send the trailer.
    }

    this->mb_.reset ();
    if (this->rf_.read (this->mb_,
                        this->mb_.size (),
                        this->file_offset_,
                        0, // @@ offset_high !!! if aiocb64 is used.
                        0, // Act
                        this->result_.priority (),
                        this->result_.signal_number ()) == -1)
    {

        ACE_ERROR ((LM_ERROR,
            ACE_TEXT("Error:Asynch_Transmit_Handler::read from file failed\n")));
        u_long err = errno ? errno : EFAULT;
        set_completion (err);
        return -1;
    }
    return 0;
}

// *********************************************************************

TRB_POSIX_Asynch_Transmit_File::TRB_POSIX_Asynch_Transmit_File (TRB_POSIX_Proactor *posix_proactor)
  : TRB_Asynch_Operation_Impl (),
    TRB_Asynch_Transmit_File_Impl (),
    TRB_POSIX_Asynch_Operation (posix_proactor),
    file_ (ACE_INVALID_HANDLE)
{
}

int
TRB_POSIX_Asynch_Transmit_File::transmit_file (ACE_HANDLE file,
                                               TRB_Asynch_Transmit_File::Header_And_Trailer *header_and_trailer,
                                               size_t bytes_to_write,
                                               u_long offset,
                                               u_long offset_high,
                                               size_t bytes_per_send,
                                               u_long flags,
                                               const void *act,
                                               int priority,
                                               int signal_number)
{
  // Adjust these parameters if there are default values specified.
  ssize_t file_size = ACE_OS::filesize (file);

  if (file_size == -1)
    ACE_ERROR_RETURN ((LM_ERROR,
                       "Error:%N:%l:%p\n",
                       "POSIX_Asynch_Transmit_File:filesize failed"),
                      -1);

 
  if (offset > (size_t) file_size)
    ACE_ERROR_RETURN ((LM_ERROR,
                       "Error:%p\n",
                       "Asynch_Transmit_File:File size is less than offset"),
                      -1);

  size_t  limit = ((size_t)  file_size) - offset;

  if (bytes_to_write == 0)
    bytes_to_write = file_size;

 if (bytes_to_write > limit)
    bytes_to_write = limit;

  //if (offset != 0)
    // Unclear why was + 1 ???
    //bytes_to_write = file_size - offset + 1;

  if (bytes_per_send == 0)
    bytes_per_send = bytes_to_write;
  
  if (bytes_per_send > 16384)
    bytes_per_send = 16384;
  

  // Configure the result parameter.

  // Make the auxillary handler and initiate transmit.
  TRB_POSIX_Asynch_Transmit_Handler *transmit_handler = 0;

  ACE_NEW_RETURN (transmit_handler,
                  ::TRB_POSIX_Asynch_Transmit_Handler (this->posix_proactor (),
                                                       *this->handler_,
                                                       this->handle_,
                                                       file,
                                                       header_and_trailer,
                                                       bytes_to_write,
                                                       offset,
                                                       offset_high,
                                                       bytes_per_send,
                                                       flags,
                                                       act,
                                                       this->posix_proactor ()->get_handle (),
                                                       priority,
                                                       signal_number),
                  -1);


  // set file handle for cancel ()
  file_ = file;

  ssize_t return_val = transmit_handler->transmit ();

  if (return_val == -1)
    {
      // This deletes the <result> in it too.
      file_ = ACE_INVALID_HANDLE;
      delete transmit_handler;
      return -1;
    }

  return 0;
}

TRB_POSIX_Asynch_Transmit_File::~TRB_POSIX_Asynch_Transmit_File (void)
{
}

int
TRB_POSIX_Asynch_Transmit_File::cancel (void)
{
  if (!posix_proactor_)
     return -1;
  
  if (this->file_ != ACE_INVALID_HANDLE)
    posix_proactor_->cancel_aio (this->file_);
    
  return posix_proactor_->cancel_aio (this->handle_);
}

// *********************************************************************

size_t
TRB_POSIX_Asynch_Read_Dgram_Result::bytes_to_read (void) const
{
  return this->aio_nbytes;
}

ACE_HANDLE
TRB_POSIX_Asynch_Read_Dgram_Result::handle (void) const
{
  return this->aio_fildes;
}

ACE_Message_Block*
TRB_POSIX_Asynch_Read_Dgram_Result::message_block () const
{
  return this->message_block_;
}

int
TRB_POSIX_Asynch_Read_Dgram_Result::flags (void) const
{
  return this->flags_;
}

struct msghdr *
TRB_POSIX_Asynch_Read_Dgram_Result::msghdr(void)
{
  return & this->msg_hdr_;
}

int
TRB_POSIX_Asynch_Read_Dgram_Result::remote_address (ACE_Addr& addr) const
{
  int retVal = -1;  // failure

  // make sure the addresses are of the same type
  if (this->remote_address_ != 0 &&
      addr.get_type () == this->remote_address_->get_type ())
    { 
      // copy the remote_address_ into addr
      addr.set_addr (this->remote_address_->get_addr (),
                   this->remote_address_->get_size ());

      retVal = 0; // success
  }

  return retVal;
}

const ACE_Addr & 
TRB_POSIX_Asynch_Read_Dgram_Result::remote_address () const
{
  if (this->remote_address_ == 0)
    return ACE_Addr::sap_any;

  return *(this->remote_address_);
}
sockaddr *
TRB_POSIX_Asynch_Read_Dgram_Result::saddr () const
{
  if (this->remote_address_ == 0)
    return 0;

  return (sockaddr *) this->remote_address_->get_addr ();
}


TRB_POSIX_Asynch_Read_Dgram_Result::TRB_POSIX_Asynch_Read_Dgram_Result
  (TRB_Handler &handler,
   ACE_HANDLE handle,
   ACE_Message_Block *message_block,
   size_t bytes_to_read,
   int flags,
   int protocol_family,
   const void* act,
   ACE_HANDLE event,
   int priority,
   int signal_number)

  : TRB_Asynch_Result_Impl (),
    TRB_Asynch_Read_Dgram_Result_Impl(),
    TRB_POSIX_Asynch_Result (handler, act, event, 0, 0, priority, signal_number),
    message_block_ (message_block),
    remote_address_ (0),
    flags_ (flags)
{

  ACE_UNUSED_ARG(bytes_to_read);

  this->aio_fildes = handle;
  this->aio_nbytes = 0;
  
  this->msg_hdr_.msg_name = 0;
  this->msg_hdr_.msg_namelen = 0;
  this->msg_hdr_.msg_iov = 0;
  this->msg_hdr_.msg_iovlen = 0;

#if (defined(sun) && defined(_XOPEN_SOURCE) && (_XOPEN_VERSION - 0 == 4)) /* XPG4 or XPG4v2 */
  this->msg_hdr_.msg_control = 0;
  this->msg_hdr_.msg_controllen = 0;
  this->msg_hdr_.msg_flags = 0;
#endif

  switch (protocol_family)
    {
    case PF_INET:
      ACE_NEW_NORETURN (this->remote_address_, ACE_INET_Addr);
      break;
    case PF_UNIX:
      ACE_NEW_NORETURN (this->remote_address_, ACE_UNIX_Addr);
      break;
    default:
      ACE_ASSERT(0);
      break;
    }

  if (this->remote_address_ != 0)
    {
      
#if defined (ACE_HAS_SOCKADDR_MSG_NAME)
      this->msg_hdr_.msg_name = 
        (struct sockaddr *) this->remote_address_->get_addr ();
#else
      this->msg_hdr_.msg_name = 
        (char *)  this->remote_address_->get_addr ();
#endif /* ACE_HAS_SOCKADDR_MSG_NAME */

      this->msg_hdr_.msg_namelen = this->remote_address_->get_size ();
    }

  iovec  iov[ACE_IOV_MAX];
  int    iovcnt = 0;

  for (const ACE_Message_Block* msg = message_block;
       msg != 0 && iovcnt < ACE_IOV_MAX;
       msg = msg->cont () )
    {
      size_t msg_space = msg->space ();


      // Make as many iovec as needed to fit all of msg_len.
      size_t wr_ptr_offset = 0;
      while (msg_space > 0 && iovcnt < ACE_IOV_MAX)
        {
          u_long this_chunk_length = ULONG_MAX;

          if (msg_space < ULONG_MAX)
            this_chunk_length = ACE_static_cast (u_long, msg_space);

          // Collect the data in the iovec.
          iov[iovcnt].iov_base = msg->wr_ptr () + wr_ptr_offset;
          iov[iovcnt].iov_len  = this_chunk_length;
          iovcnt++;

          this->aio_nbytes += this_chunk_length;

          msg_space -= this_chunk_length;
          wr_ptr_offset += this_chunk_length;

        }

      if (msg_space > 0)    // Ran out of iovecs before msg_space exhausted
        {
          errno = ERANGE;
          return;
        }
     }

  
  if (iovcnt > 0)
      ACE_NEW (this->msg_hdr_.msg_iov, iovec[iovcnt]);

  this->msg_hdr_.msg_iovlen = iovcnt;

  for (int i=0; i < iovcnt; ++i)
    {
      this->msg_hdr_.msg_iov[i] = iov[i];
    }

  errno =0;
  return;
}

void
TRB_POSIX_Asynch_Read_Dgram_Result::complete (size_t bytes_transferred,
                                              int success,
                                              const void *completion_key,
                                              u_long error)
{
  // bytes_transferred,completion key, error 
  // are already stored in this result.
  // They are passed as parameters for compatibility with
  // old version.

  ACE_UNUSED_ARG (success);
  ACE_UNUSED_ARG (completion_key);
  ACE_UNUSED_ARG (error);

  // Appropriately move the pointers in the message block.
  for (ACE_Message_Block* mb = this->message_block_;
       (mb != 0) && (bytes_transferred > 0);
       mb = mb->cont ())
    {
      size_t len_part = mb->space ();
       
      if (len_part == 0)
        continue; 

      if ( len_part > bytes_transferred)
        len_part = bytes_transferred;

      mb->wr_ptr (len_part);

      bytes_transferred -= len_part;
    }

  // Adjust the address length
  if (this->remote_address_ != 0)
    this->remote_address_->set_size (this->msg_hdr_.msg_namelen);

#if (defined(sun) && defined(_XOPEN_SOURCE) && (_XOPEN_VERSION - 0 == 4)) /* XPG4 or XPG4v2 */
  // replace input flags with output flags
  this->flags_ = this->msg_hdr_.msg_flags;
#endif


  // Create the interface result class.
  TRB_Asynch_Read_Dgram::Result result (this);

  // Call the application handler.
  this->handler_.handle_read_dgram (result);
}

TRB_POSIX_Asynch_Read_Dgram_Result::~TRB_POSIX_Asynch_Read_Dgram_Result (void)
{
  delete this->remote_address_;
  delete [] this->msg_hdr_.msg_iov;
}

/***************************************************************************/
TRB_POSIX_Asynch_Read_Dgram::~TRB_POSIX_Asynch_Read_Dgram (void)
{
}

ssize_t
TRB_POSIX_Asynch_Read_Dgram::recv (ACE_Message_Block *message_block,
                                   size_t &number_of_bytes_recvd,
                                   int flags,
                                   int protocol_family,
                                   const void *act,
                                   int priority,
                                   int signal_number)
{

  TRB_POSIX_Proactor *proactor = this->posix_proactor ();

  int ret_val = proactor->start_read_dgram(*this->handler_,
                                        this->handle_,
                                        message_block,
                                        number_of_bytes_recvd,
                                        flags,
                                        protocol_family,
                                        act,
                                        proactor->get_handle (),
                                        priority,
                                        signal_number);

  //useless parameter
  number_of_bytes_recvd = 0;

  return ret_val;
}

TRB_POSIX_Asynch_Read_Dgram::TRB_POSIX_Asynch_Read_Dgram (TRB_POSIX_Proactor *posix_proactor)
  : TRB_Asynch_Operation_Impl (),
    TRB_Asynch_Read_Dgram_Impl (),
    TRB_POSIX_Asynch_Operation (posix_proactor)
{
}

//***************************************************************************

size_t
TRB_POSIX_Asynch_Write_Dgram_Result::bytes_to_write (void) const
{
  return this->aio_nbytes;
}

ACE_HANDLE
TRB_POSIX_Asynch_Write_Dgram_Result::handle (void) const
{
  return this->aio_fildes;
}

ACE_Message_Block*
TRB_POSIX_Asynch_Write_Dgram_Result::message_block () const
{
  return this->message_block_;
}

int
TRB_POSIX_Asynch_Write_Dgram_Result::flags (void) const
{
  return this->flags_;
}

struct msghdr *
TRB_POSIX_Asynch_Write_Dgram_Result::msghdr(void)
{
  return & this->msg_hdr_;
}

int
TRB_POSIX_Asynch_Write_Dgram_Result::remote_address (ACE_Addr& addr) const
{
  int retVal = -1;  // failure

  // make sure the addresses are of the same type
  if (this->remote_address_ != 0 &&
      addr.get_type () == this->remote_address_->get_type ())
    { 
      // copy the remote_address_ into addr
      addr.set_addr (this->remote_address_->get_addr (),
                   this->remote_address_->get_size ());

      retVal = 0; // success
  }

  return retVal;
}

const ACE_Addr & 
TRB_POSIX_Asynch_Write_Dgram_Result::remote_address () const
{
  if (this->remote_address_ == 0)
    return ACE_Addr::sap_any;

  return *(this->remote_address_);
}

sockaddr *
TRB_POSIX_Asynch_Write_Dgram_Result::saddr () const
{
  if (this->remote_address_ == 0)
    return 0;

  return (sockaddr *) this->remote_address_->get_addr ();
}


TRB_POSIX_Asynch_Write_Dgram_Result::TRB_POSIX_Asynch_Write_Dgram_Result
  (TRB_Handler &handler,
   ACE_HANDLE handle,
   const ACE_Addr & remote_sap,
   ACE_Message_Block *message_block,
   size_t bytes_to_write,
   int flags,
   const void* act,
   ACE_HANDLE event,
   int priority,
   int signal_number)

  : TRB_Asynch_Result_Impl (),
    TRB_Asynch_Write_Dgram_Result_Impl(),
    TRB_POSIX_Asynch_Result (handler, act, event, 0, 0, priority, signal_number),
    message_block_ (message_block),
    remote_address_(0),
    flags_ (flags)
{
  ACE_UNUSED_ARG(bytes_to_write);

  this->aio_fildes = handle;
  this->aio_nbytes = 0;
  
  this->msg_hdr_.msg_name = 0;
  this->msg_hdr_.msg_namelen = 0;
  this->msg_hdr_.msg_iov = 0;
  this->msg_hdr_.msg_iovlen = 0;

#if (defined(sun) && defined(_XOPEN_SOURCE) && (_XOPEN_VERSION - 0 == 4)) /* XPG4 or XPG4v2 */
  this->msg_hdr_.msg_control = 0;
  this->msg_hdr_.msg_controllen = 0;
  this->msg_hdr_.msg_flags = flags_;
#endif

  switch (remote_sap.get_type())
    {
    case PF_INET:
      ACE_NEW_NORETURN (this->remote_address_, ACE_INET_Addr);
      break;
    case PF_UNIX:
      ACE_NEW_NORETURN (this->remote_address_, ACE_UNIX_Addr);
      break;
    default:
      ACE_ASSERT(0);
      break;
    }

  if (this->remote_address_ != 0)
    {
      this->remote_address_->set_addr (remote_sap.get_addr (),
                                       remote_sap.get_size ());

#if defined (ACE_HAS_SOCKADDR_MSG_NAME)
      this->msg_hdr_.msg_name = 
        (struct sockaddr *) this->remote_address_->get_addr ();
#else
      this->msg_hdr_.msg_name = 
        (char *)  this->remote_address_->get_addr ();
#endif /* ACE_HAS_SOCKADDR_MSG_NAME */

      this->msg_hdr_.msg_namelen = this->remote_address_->get_size ();

    }

  iovec  iov[ACE_IOV_MAX];
  int    iovcnt = 0;

  for (const ACE_Message_Block* msg = message_block;
       msg != 0 && iovcnt < ACE_IOV_MAX;
       msg = msg->cont () )
    {
      size_t msg_len = msg->length ();


      // Make as many iovec as needed to fit all of msg_len.
      size_t rd_ptr_offset = 0;
      while (msg_len > 0 && iovcnt < ACE_IOV_MAX)
        {
          u_long this_chunk_length = ULONG_MAX;

          if (msg_len < ULONG_MAX)
            this_chunk_length = ACE_static_cast (u_long, msg_len);

          // Collect the data in the iovec.
          iov[iovcnt].iov_base = msg->rd_ptr () + rd_ptr_offset;
          iov[iovcnt].iov_len  = this_chunk_length;
          ++iovcnt;

          this->aio_nbytes += this_chunk_length;

          msg_len -= this_chunk_length;
          rd_ptr_offset += this_chunk_length;
        }

      if (msg_len > 0)       // Ran out of iovecs before msg_space exhausted
      {
        errno = ERANGE;
        return;
      }
    }

  if (iovcnt > 0)
      ACE_NEW (this->msg_hdr_.msg_iov, iovec[iovcnt]);

  this->msg_hdr_.msg_iovlen = iovcnt;

  for (int i=0; i < iovcnt; ++i)
    {
      this->msg_hdr_.msg_iov[i] = iov[i];
    }

  errno =0;
  return;
}

TRB_POSIX_Asynch_Write_Dgram_Result::~TRB_POSIX_Asynch_Write_Dgram_Result (void)
{
  delete this->remote_address_;
  delete [] this->msg_hdr_.msg_iov;
}

void
TRB_POSIX_Asynch_Write_Dgram_Result::complete (size_t bytes_transferred,
                                               int success,
                                               const void *completion_key,
                                               u_long error)
{
  // bytes_transferred,completion key, error 
  // are already stored in this result.
  // They are passed as parameters for compatibility with
  // old version.

  ACE_UNUSED_ARG (success);
  ACE_UNUSED_ARG (completion_key);
  ACE_UNUSED_ARG (error);


  // Appropriately move the pointers in the message block.
  for (ACE_Message_Block* mb = this->message_block_;
       (mb != 0) && (bytes_transferred > 0);
       mb = mb->cont ())
    {
      size_t len_part = mb->length ();

      if (len_part == 0)
        continue;

      if ( len_part > bytes_transferred)
        len_part = bytes_transferred;

      mb->rd_ptr (len_part);

      bytes_transferred -= len_part;
    }

  // Create the interface result class.
  TRB_Asynch_Write_Dgram::Result result (this);

  // Call the application handler.
  this->handler_.handle_write_dgram (result);
}

//***************************************************************************

TRB_POSIX_Asynch_Write_Dgram::~TRB_POSIX_Asynch_Write_Dgram (void)
{
}

ssize_t
TRB_POSIX_Asynch_Write_Dgram::send (ACE_Message_Block *message_block,
                                    size_t & bytes_to_write,
                                    int flags,
                                    const ACE_Addr &addr,
                                    const void *act,
                                    int priority,
                                    int signal_number)
{
  TRB_POSIX_Proactor *proactor = this->posix_proactor ();

  int ret_val = proactor->start_write_dgram(*this->handler_,
                                        this->handle_,
                                        addr,
                                        message_block,
                                        bytes_to_write,
                                        flags,
                                        act,
                                        proactor->get_handle (),
                                        priority,
                                        signal_number);

  //useless parameter
  bytes_to_write = 0;

  return ret_val;
}

TRB_POSIX_Asynch_Write_Dgram::TRB_POSIX_Asynch_Write_Dgram
  (TRB_POSIX_Proactor *posix_proactor)
  : TRB_Asynch_Operation_Impl (),
    TRB_Asynch_Write_Dgram_Impl (),
    TRB_POSIX_Asynch_Operation (posix_proactor)
{
}

// *********************************************************************

#if defined (ACE_HAS_EXPLICIT_TEMPLATE_INSTANTIATION)

template class ACE_Double_Linked_List<TRB_POSIX_Asynch_Result>;
template class ACE_Double_Linked_List_Iterator_Base<TRB_POSIX_Asynch_Result>;
template class ACE_Double_Linked_List_Iterator<TRB_POSIX_Asynch_Result>;


#elif  defined (ACE_HAS_TEMPLATE_INSTANTIATION_PRAGMA)

#pragma instantiate ACE_Double_Linked_List<TRB_POSIX_Asynch_Result>
#pragma instantiate ACE_Double_Linked_List_Iterator_Base<TRB_POSIX_Asynch_Result>
#pragma instantiate ACE_Double_Linked_List_Iterator<TRB_POSIX_Asynch_Result>


#endif /* ACE_HAS_EXPLICIT_TEMPLATE_INSTANTIATION */


#endif /* ACE_HAS_AIO_CALLS */
