/* -*- C++ -*- */
// POSIX_Proactor.cpp,v 4.55 2002/12/21 23:29:24 shuston Exp

#include "TProactor/POSIX_Proactor.h"

#if defined (ACE_HAS_AIO_CALLS) || defined(ACE_HAS_AIO_EMULATION)

# if defined (ACE_HAS_SYSINFO)
#   include /**/ <sys/systeminfo.h>
# endif /* ACE_HAS_SYS_INFO */


#include "ace/Task_T.h"
#include "ace/Log_Msg.h"
#include "ace/Object_Manager.h"
#include "TProactor/POSIX_AIO_Provider.h"

#include "ace/Malloc_T.h"
#include "TProactor/POSIX_AIO_Allocator.h"

//----------------------------------------------------------
#include "ace/Version.h"

#if ((ACE_MAJOR_VERSION == 5) && (ACE_MINOR_VERSION ==3))

#include "ace/OS.h"

#else  // we assume ACE 5.4.X

# if defined (sun)
#   include "ace/OS_NS_strings.h"
# endif /* sun */

#include "ace/OS_NS_sys_socket.h"
#include "ace/OS_NS_signal.h"

#endif //((ACE_MAJOR_VERSION == 5) && (ACE_MINOR_VERSION ==3))
//----------------------------------------------------------

typedef 
  TRB_Cached_List_Allocator <TRB_POSIX_MAX_Result, ACE_SYNCH_NULL_MUTEX>
  RESULT_ALLOCATOR;

// *********************************************************************

TRB_POSIX_Wakeup_Completion::TRB_POSIX_Wakeup_Completion (TRB_Handler &handler,
                                                          const void *act,
                                                          ACE_HANDLE event,
                                                          int priority,
                                                          int signal_number)
  : TRB_Asynch_Result_Impl (),
    TRB_POSIX_Asynch_Result (handler, act, event, 0, 0, priority, signal_number)
{
}

TRB_POSIX_Wakeup_Completion::~TRB_POSIX_Wakeup_Completion (void)
{
}

void
TRB_POSIX_Wakeup_Completion::complete (size_t       /* bytes_transferred */,
                                       int          /* success */,
                                       const void * /* completion_key */,
                                       u_long       /*  error */)
{

  this->handler_.handle_wakeup ();
}

// *********************************************************************

TRB_POSIX_Asynch_Timer::TRB_POSIX_Asynch_Timer (TRB_Handler &handler,
                                                const void *act,
                                                const ACE_Time_Value &tv,
                                                ACE_HANDLE event,
                                                int priority,
                                                int signal_number)
  : TRB_Asynch_Result_Impl (),
    TRB_POSIX_Asynch_Result (handler, act, event, 0, 0, priority, signal_number),
    time_ (tv)
{
}

void
TRB_POSIX_Asynch_Timer::complete (size_t       /* bytes_transferred */,
                                  int          /* success */,
                                  const void * /* completion_key */,
                                  u_long       /* error */)
{
  this->handler_.handle_time_out (this->time_, this->act ());
}


// *********************************************************************


// *********************************************************************

int 
TRB_POSIX_Proactor::create_result_allocator (void)
{
  ACE_MT (ACE_GUARD_RETURN (ACE_SYNCH_MUTEX, ace_mon, this->mutex_, -1));

  if (this->aio_allocator_ != 0)
    return 0;


  ACE_NEW_RETURN (this->aio_allocator_,
                  RESULT_ALLOCATOR(),
                  -1);
             

  return 0;
}

void 
TRB_POSIX_Proactor::delete_result_allocator (void)
{
  ACE_MT (ACE_GUARD (ACE_SYNCH_MUTEX, ace_mon, this->mutex_));

  delete this->aio_allocator_;
  this->aio_allocator_ = 0;
}

void * 
TRB_POSIX_Proactor::allocate_any_result ()
{
  if (this->aio_allocator_ != 0)
    return this->aio_allocator_->malloc (sizeof(TRB_POSIX_MAX_Result));

  return 0;
}

// *********************************************************************
int 
TRB_POSIX_Proactor::get_os_id(void)
{
  int os_id = OS_UNDEFINED;

#if defined(sun)

  os_id = OS_SUN; // set family

  char Buf [32];

  ::memset(Buf,0,sizeof(Buf));

  ACE_OS::sysinfo (SI_RELEASE , Buf, sizeof(Buf)-1);

  if (ACE_OS_String::strcasecmp (Buf , "5.6") == 0)
    os_id = OS_SUN_56;
  else if (ACE_OS_String::strcasecmp (Buf , "5.7") == 0)
    os_id = OS_SUN_57;
  else if (ACE_OS_String::strcasecmp (Buf , "5.8") == 0)
    os_id = OS_SUN_58;

#elif defined(HPUX)

  os_id = OS_HPUX;   // set family

#elif defined(__sgi)

  os_id = OS_IRIX;   // set family

#elif defined(__OpenBSD)

  os_id = OS_OPENBSD; // set family

  // do the same

//#else defined (LINUX, __FreeBSD__ ...)
//setup here os_id_
#endif

  return os_id;
}

// *********************************************************************
TRB_POSIX_Proactor::TRB_POSIX_Proactor (const TRB_POSIX_AIO_Config & prime_processor_cfg,
                                        const TRB_POSIX_AIO_Config & second_processor_cfg)
  :  wakeup_handler_ (),
     os_id_          (TRB_POSIX_Proactor::get_os_id()),
     flg_open_       (0),
     flg_disable_start_(0),
     mutex_          (),
     dispatcher_     (mutex_),
     prime_config_   (prime_processor_cfg),
     second_config_  (second_processor_cfg),
     prime_aio_processor_  (0),
     second_aio_processor_ (0),
     aio_allocator_        (0)
{
  this->create_result_allocator();
  
  if (this->open() !=0)
    this->close();
}

TRB_POSIX_Proactor::TRB_POSIX_Proactor (size_t max_op,
                                        int sig_num,
                                        int leader_type,
                                        int start_aio_type)
  :  wakeup_handler_ (),
     os_id_          (TRB_POSIX_Proactor::get_os_id()),
     flg_open_       (0),
     flg_disable_start_(0),
     mutex_          (),
     dispatcher_     (mutex_),
     prime_config_   (max_op,
                      (leader_type != 0)
                        ?  TRB_POSIX_AIO_Config::PCT_DEDICATED
                        :  TRB_POSIX_AIO_Config::PCT_SHARED,
                      TRB_POSIX_AIO_Config::PCF_ZERO,
                      TRB_POSIX_AIO_Config::PVT_SELECT, // POSIX,
                      (start_aio_type != 0)
                        ?  TRB_POSIX_AIO_Config::PVF_AIO_START_ANY_THREAD |
                           TRB_POSIX_AIO_Config::PVF_AIO_START_INTERRUPT_WAIT
                        :  0,
                      TRB_POSIX_AIO_Config::WST_DEFAULT,
                      ACE_SIGRTMIN,
                      (sig_num == 0) 
                        ?  TRB_POSIX_AIO_Config::IST_PIPE
                        :  TRB_POSIX_AIO_Config::IST_SIGNAL,
                      sig_num),
     second_config_   (max_op,
                       TRB_POSIX_AIO_Config::PCT_NONE), // NOT EXIST
     prime_aio_processor_  (0),
     second_aio_processor_ (0),
     aio_allocator_        (0)
{
  this->create_result_allocator();

  if (this->open() !=0)
    this->close();
}

TRB_POSIX_Proactor::~TRB_POSIX_Proactor (void)
{
  this->close ();
  this->delete_result_allocator();
}

int
TRB_POSIX_Proactor::open (void)
{
  ACE_MT (ACE_GUARD_RETURN (ACE_SYNCH_MUTEX, ace_mon, this->mutex_, -1));

  ACE_DEBUG((LM_DEBUG,
             ACE_LIB_TEXT("=======Prime AIO Processor Configuration\n")));
  this->prime_config_.dump();

  ACE_DEBUG((LM_DEBUG,
             ACE_LIB_TEXT("=======Second AIO Processor Configuration\n")));
  this->second_config_.dump();

  ACE_DEBUG((LM_DEBUG,
             ACE_LIB_TEXT("===========================================\n")));

  if (this->flg_open_ != 0)
    return 0;

  //
  this->prime_aio_processor_ = 
    this->prime_config_.create_processor (this->dispatcher_);

  if (this->prime_aio_processor_ != 0)
    {
      if (this->prime_aio_processor_->create_provider () != 0)
        return -1;

      if (this->prime_aio_processor_->start () != 0)
        return -1;
      
    }
  ++this->flg_open_;

  //
  this->second_aio_processor_ = 
    this->second_config_.create_processor(this->dispatcher_);

  if (this->second_aio_processor_ != 0)
    {
      if (this->second_aio_processor_->create_provider () != 0)
        return -1;

      if (this->second_aio_processor_->start () != 0)
        return -1;
      
    }
  ++this->flg_open_;

  if (flg_open_ == 0)
    return -1;


  if (this->dispatcher_.init (this->prime_aio_processor_,
                              this->second_aio_processor_) !=0)
    return -1;

  return 0;
}
int
TRB_POSIX_Proactor::close (void)
{
  ACE_MT (ACE_GUARD_RETURN (ACE_SYNCH_MUTEX, ace_mon, this->mutex_, -1));

  if (this->second_aio_processor_ != 0)
    {
      this->second_aio_processor_->stop ();
      this->second_aio_processor_->destroy_provider ();

      delete this->second_aio_processor_;
      this->second_aio_processor_ = 0;
    }

  if (this->prime_aio_processor_ != 0)
    {
      this->prime_aio_processor_->stop ();
      this->prime_aio_processor_->destroy_provider ();

      delete this->prime_aio_processor_;
      this->prime_aio_processor_ = 0;
    }

  this->dispatcher_.fini ();

  this->flg_open_ =0;
  return 0;
}

int
TRB_POSIX_Proactor::register_handle (ACE_HANDLE handle,
                                     const void *completion_key)
{
  ACE_UNUSED_ARG (handle);
  ACE_UNUSED_ARG (completion_key);
  return 0;
}

int
TRB_POSIX_Proactor::unregister_handle (ACE_HANDLE handle)
{
  ACE_UNUSED_ARG (handle);
  return 0;
}

int
TRB_POSIX_Proactor::wake_up_dispatch_threads (void)
{
  return 0;
}

int
TRB_POSIX_Proactor::close_dispatch_threads (int)
{
  return 0;
}

size_t
TRB_POSIX_Proactor::number_of_threads (void) const
{
  // @@ Implement it.
  ACE_NOTSUP_RETURN (0);
}

void
TRB_POSIX_Proactor::number_of_threads (size_t threads)
{
  // @@ Implement it.
  ACE_UNUSED_ARG (threads);
}

ACE_HANDLE
TRB_POSIX_Proactor::get_handle (void) const
{
  return ACE_INVALID_HANDLE;
}
int
TRB_POSIX_Proactor::post_wakeup_completions (int how_many)
{
  TRB_POSIX_Wakeup_Completion *wakeup_completion = 0;

  for (int ci = 0; ci < how_many; ci++)
    {
      ACE_NEW_RETURN (wakeup_completion,
                      TRB_POSIX_Wakeup_Completion (this->wakeup_handler_),
                      -1);
      if (this->post_completion (wakeup_completion) == -1)
        return -1;
    }

  return 0;
}

int
TRB_POSIX_Proactor::post_completion (TRB_POSIX_Asynch_Result *result)
{
  ACE_MT (ACE_GUARD_RETURN (ACE_SYNCH_MUTEX, ace_mon, this->mutex_, -1));

  if (this->flg_open_ == 0)
    return -1;

  return this->dispatcher_.post_result (result,
                                        TRB_POSIX_AIO_Dispatcher::NOTIFY_STRONG);
}


int
TRB_POSIX_Proactor::start_aio (TRB_POSIX_Asynch_Result *result)
{
  ACE_TRACE ("TRB_POSIX_Proactor::start_aio");

  //ACE_MT (ACE_GUARD_RETURN (ACE_SYNCH_MUTEX, ace_mon, this->mutex_, -1));

  int rc = -1;

  if (this->flg_open_ != 0 && this->flg_disable_start_ == 0)
    {
      int op = result->op_code ();

      if ((this->prime_aio_processor_ != 0) &&
          (this->prime_aio_processor_->is_supported (op)))
        {
          rc = this->prime_aio_processor_->start_aio (result);
        }
      else if ((this->second_aio_processor_ != 0) &&
          (this->second_aio_processor_->is_supported (op)))
        {
          rc = this->second_aio_processor_->start_aio (result);
        }
    }

  if (rc != 0)
    TRB_POSIX_Asynch_Result::free(result);

  return rc;
}

int
TRB_POSIX_Proactor::cancel_all_aio (void)
{
  ACE_TRACE ("TRB_POSIX_Proactor::cancel_aio");

  ACE_MT (ACE_GUARD_RETURN (ACE_SYNCH_MUTEX, ace_mon, this->mutex_, -1));

  if (this->flg_open_ == 0)
    return -1;

  this->flg_disable_start_ = 1;

  int rc1 = TRB_POSIX_AIO_Provider::CR_ALL_DONE;
  int rc2 = TRB_POSIX_AIO_Provider::CR_ALL_DONE;

  if (this->prime_aio_processor_ != 0)
      rc1 =  this->prime_aio_processor_->cancel_all_aio ();

  if (this->second_aio_processor_ != 0)
      rc1 =  this->second_aio_processor_->cancel_all_aio ();

  if (rc1 == TRB_POSIX_AIO_Provider::CR_CANCEL_ERROR ||
      rc2 == TRB_POSIX_AIO_Provider::CR_CANCEL_ERROR )
      return -1;

  if (rc1 == TRB_POSIX_AIO_Provider::CR_NOT_CANCELED ||
      rc2 == TRB_POSIX_AIO_Provider::CR_NOT_CANCELED )
      return 2;  // NOT CANCELLED

  if (rc1 == TRB_POSIX_AIO_Provider::CR_CANCELED ||
      rc2 == TRB_POSIX_AIO_Provider::CR_CANCELED )
      return 0;  // CANCELLED

  return 1;  // ALLDONE
}

int
TRB_POSIX_Proactor::cancel_aio (TRB_Handler * handler)
{
  ACE_TRACE ("TRB_POSIX_Proactor::cancel_aio");

  ACE_MT (ACE_GUARD_RETURN (ACE_SYNCH_MUTEX, ace_mon, this->mutex_, -1));

  if (this->flg_open_ == 0)
    return -1;

  int rc1 = TRB_POSIX_AIO_Provider::CR_ALL_DONE;
  int rc2 = TRB_POSIX_AIO_Provider::CR_ALL_DONE;

  if (this->prime_aio_processor_ != 0)
      rc1 =  this->prime_aio_processor_->cancel_aio (handler);

  if (this->second_aio_processor_ != 0)
      rc1 =  this->second_aio_processor_->cancel_aio (handler);

  if (rc1 == TRB_POSIX_AIO_Provider::CR_CANCEL_ERROR ||
      rc2 == TRB_POSIX_AIO_Provider::CR_CANCEL_ERROR )
      return -1;

  if (rc1 == TRB_POSIX_AIO_Provider::CR_NOT_CANCELED ||
      rc2 == TRB_POSIX_AIO_Provider::CR_NOT_CANCELED )
      return 2;  // NOT CANCELLED

  if (rc1 == TRB_POSIX_AIO_Provider::CR_CANCELED ||
      rc2 == TRB_POSIX_AIO_Provider::CR_CANCELED )
      return 0;  // CANCELLED

  return 1;  // ALLDONE
}

int
TRB_POSIX_Proactor::cancel_aio (ACE_HANDLE handle)
{
  ACE_TRACE ("TRB_POSIX_Proactor::cancel_aio");

  ACE_MT (ACE_GUARD_RETURN (ACE_SYNCH_MUTEX, ace_mon, this->mutex_, -1));

  if (this->flg_open_ == 0)
    return -1;

  int rc1 = TRB_POSIX_AIO_Provider::CR_ALL_DONE;
  int rc2 = TRB_POSIX_AIO_Provider::CR_ALL_DONE;

  if (this->prime_aio_processor_ != 0)
      rc1 =  this->prime_aio_processor_->cancel_aio (handle);

  if (this->second_aio_processor_ != 0)
      rc1 =  this->second_aio_processor_->cancel_aio (handle);

  if (rc1 == TRB_POSIX_AIO_Provider::CR_CANCEL_ERROR ||
      rc2 == TRB_POSIX_AIO_Provider::CR_CANCEL_ERROR )
      return -1;

  if (rc1 == TRB_POSIX_AIO_Provider::CR_NOT_CANCELED ||
      rc2 == TRB_POSIX_AIO_Provider::CR_NOT_CANCELED )
      return 2;  // NOT CANCELLED

  if (rc1 == TRB_POSIX_AIO_Provider::CR_CANCELED ||
      rc2 == TRB_POSIX_AIO_Provider::CR_CANCELED )
      return 0;  // CANCELLED

  return 1;  // ALLDONE
}

int
TRB_POSIX_Proactor::handle_events (ACE_Time_Value &wait_time)
{
  return this->handle_events_i (&wait_time);
}

int
TRB_POSIX_Proactor::handle_events (void)
{
  return this->handle_events_i (0);
}

int
TRB_POSIX_Proactor::handle_events_i (ACE_Time_Value *timeout)
{
  ACE_TRACE ("TRB_POSIX_Proactor::handle_events.inl");

  ACE_MT (ACE_GUARD_RETURN (ACE_SYNCH_MUTEX, ace_mon, this->mutex_, -1));

  if (this->flg_open_ == 0)
    return -1;

  return this->dispatcher_.handle_events (timeout);
}

TRB_POSIX_Proactor::Proactor_Type 
TRB_POSIX_Proactor::get_impl_type (void)
{
  return PROACTOR_POSIX;
}

// **********************************************************************

TRB_Asynch_Read_Stream_Impl *
TRB_POSIX_Proactor::create_asynch_read_stream (void)
{
  TRB_Asynch_Read_Stream_Impl *implementation = 0;
  ACE_NEW_RETURN (implementation,
                  TRB_POSIX_Asynch_Read_Stream (this),
                  0);
  return implementation;
}

TRB_Asynch_Read_Stream_Result_Impl *
TRB_POSIX_Proactor::create_asynch_read_stream_result (TRB_Handler &handler,
                                                      ACE_HANDLE handle,
                                                      ACE_Message_Block &message_block,
                                                      size_t bytes_to_read,
                                                      const void* act,
                                                      ACE_HANDLE event,
                                                      int priority,
                                                      int signal_number)
{
  TRB_Asynch_Read_Stream_Result_Impl *implementation;
  ACE_NEW_RETURN (implementation,
                  TRB_POSIX_Asynch_Read_Stream_Result (handler,
                                                       handle,
                                                       message_block,
                                                       bytes_to_read,
                                                       act,
                                                       event,
                                                       priority,
                                                       signal_number),
                  0);
  return implementation;
}


TRB_Asynch_Write_Stream_Impl *
TRB_POSIX_Proactor::create_asynch_write_stream (void)
{
  TRB_Asynch_Write_Stream_Impl *implementation = 0;
  ACE_NEW_RETURN (implementation,
                  TRB_POSIX_Asynch_Write_Stream (this),
                  0);
  return implementation;
}

TRB_Asynch_Write_Stream_Result_Impl *
TRB_POSIX_Proactor::create_asynch_write_stream_result (TRB_Handler &handler,
                                                       ACE_HANDLE handle,
                                                       ACE_Message_Block &message_block,
                                                       size_t bytes_to_write,
                                                       const void* act,
                                                       ACE_HANDLE event,
                                                       int priority,
                                                       int signal_number)
{
  TRB_Asynch_Write_Stream_Result_Impl *implementation;
  ACE_NEW_RETURN (implementation,
                  TRB_POSIX_Asynch_Write_Stream_Result (handler,
                                                        handle,
                                                        message_block,
                                                        bytes_to_write,
                                                        act,
                                                        event,
                                                        priority,
                                                        signal_number),
                  0);
  return implementation;
}


TRB_Asynch_Read_File_Impl *
TRB_POSIX_Proactor::create_asynch_read_file (void)
{
  TRB_Asynch_Read_File_Impl *implementation = 0;
  ACE_NEW_RETURN (implementation,
                  TRB_POSIX_Asynch_Read_File (this),
                  0);
  return  implementation;
}

TRB_Asynch_Read_File_Result_Impl *
TRB_POSIX_Proactor::create_asynch_read_file_result (TRB_Handler &handler,
                                                    ACE_HANDLE handle,
                                                    ACE_Message_Block &message_block,
                                                    size_t bytes_to_read,
                                                    const void* act,
                                                    u_long offset,
                                                    u_long offset_high,
                                                    ACE_HANDLE event,
                                                    int priority,
                                                    int signal_number)
{
  TRB_Asynch_Read_File_Result_Impl *implementation;
  ACE_NEW_RETURN (implementation,
                  TRB_POSIX_Asynch_Read_File_Result (handler,
                                                     handle,
                                                     message_block,
                                                     bytes_to_read,
                                                     act,
                                                     offset,
                                                     offset_high,
                                                     event,
                                                     priority,
                                                     signal_number),
                  0);
  return implementation;
}


TRB_Asynch_Write_File_Impl *
TRB_POSIX_Proactor::create_asynch_write_file (void)
{
  TRB_Asynch_Write_File_Impl *implementation = 0;
  ACE_NEW_RETURN (implementation,
                  TRB_POSIX_Asynch_Write_File (this),
                  0);
  return  implementation;
}

TRB_Asynch_Write_File_Result_Impl *
TRB_POSIX_Proactor::create_asynch_write_file_result (TRB_Handler &handler,
                                                     ACE_HANDLE handle,
                                                     ACE_Message_Block &message_block,
                                                     size_t bytes_to_write,
                                                     const void* act,
                                                     u_long offset,
                                                     u_long offset_high,
                                                     ACE_HANDLE event,
                                                     int priority,
                                                     int signal_number)
{
  TRB_Asynch_Write_File_Result_Impl *implementation;
  ACE_NEW_RETURN (implementation,
                  TRB_POSIX_Asynch_Write_File_Result (handler,
                                                      handle,
                                                      message_block,
                                                      bytes_to_write,
                                                      act,
                                                      offset,
                                                      offset_high,
                                                      event,
                                                      priority,
                                                      signal_number),
                  0);
  return implementation;
}


TRB_Asynch_Read_Dgram_Impl *
TRB_POSIX_Proactor::create_asynch_read_dgram (void)
{
    TRB_Asynch_Read_Dgram_Impl *implementation = 0;
    ACE_NEW_RETURN (implementation,
                    TRB_POSIX_Asynch_Read_Dgram (this),
                    0);
    return implementation;
}

TRB_Asynch_Read_Dgram_Result_Impl *
TRB_POSIX_Proactor::create_asynch_read_dgram_result (TRB_Handler &handler,
                                                     ACE_HANDLE handle,
                                                     ACE_Message_Block *message_block,
                                                     size_t bytes_to_read,
                                                     int flags,
                                                     int protocol_family,
                                                     const void* act,
                                                     ACE_HANDLE event ,
                                                     int priority ,
                                                     int signal_number)
{
  TRB_Asynch_Read_Dgram_Result_Impl *implementation=0;
  ACE_NEW_RETURN (implementation,
                  TRB_POSIX_Asynch_Read_Dgram_Result(handler,
                                                     handle,
                                                     message_block,
                                                     bytes_to_read,
                                                     flags,
                                                     protocol_family,
                                                     act,
                                                     event,
                                                     priority,
                                                     signal_number),
                  0);

  return implementation;
}


TRB_Asynch_Write_Dgram_Impl *
TRB_POSIX_Proactor::create_asynch_write_dgram (void)
{
        TRB_Asynch_Write_Dgram_Impl *implementation = 0;
        ACE_NEW_RETURN (implementation,
                        TRB_POSIX_Asynch_Write_Dgram (this),
                        0);

    return implementation;
}

TRB_Asynch_Write_Dgram_Result_Impl *
TRB_POSIX_Proactor::create_asynch_write_dgram_result (TRB_Handler &handler,
                                                      ACE_HANDLE handle,
                                                      const ACE_Addr & remote_sap,
                                                      ACE_Message_Block *message_block,
                                                      size_t bytes_to_write,
                                                      int flags,
                                                      const void* act,
                                                      ACE_HANDLE event,
                                                      int priority ,
                                                      int signal_number)
{
  TRB_Asynch_Write_Dgram_Result_Impl *implementation=0;
  ACE_NEW_RETURN (implementation,
                  TRB_POSIX_Asynch_Write_Dgram_Result(handler,
                                                      handle,
                                                      remote_sap,
                                                      message_block,
                                                      bytes_to_write,
                                                      flags,
                                                      act,
                                                      event,
                                                      priority,
                                                      signal_number),
                  0);

  return implementation;
}


TRB_Asynch_Accept_Impl *
TRB_POSIX_Proactor::create_asynch_accept (void)
{
  TRB_Asynch_Accept_Impl *implementation = 0;
  ACE_NEW_RETURN (implementation,
                  TRB_POSIX_Asynch_Accept (this),
                  0);

  return implementation;
}

TRB_Asynch_Accept_Result_Impl *
TRB_POSIX_Proactor::create_asynch_accept_result (TRB_Handler &handler,
                                                 ACE_HANDLE listen_handle,
                                                 ACE_HANDLE accept_handle,
                                                 ACE_Message_Block &message_block,
                                                 size_t bytes_to_read,
                                                 const void* act,
                                                 ACE_HANDLE event,
                                                 int priority,
                                                 int signal_number)
{
  TRB_Asynch_Accept_Result_Impl *implementation;
  ACE_NEW_RETURN (implementation,
                  TRB_POSIX_Asynch_Accept_Result (handler,
                                                  listen_handle,
                                                  accept_handle,
                                                  message_block,
                                                  bytes_to_read,
                                                  act,
                                                  event,
                                                  priority,
                                                  signal_number),
                  0);
  return implementation;
}


TRB_Asynch_Connect_Impl *
TRB_POSIX_Proactor::create_asynch_connect (void)
{
  TRB_Asynch_Connect_Impl *implementation = 0;
  ACE_NEW_RETURN (implementation,
                  TRB_POSIX_Asynch_Connect (this),
                  0);

  return implementation;
}

TRB_Asynch_Connect_Result_Impl *
TRB_POSIX_Proactor::create_asynch_connect_result (TRB_Handler &handler,
                                                  ACE_HANDLE  connect_handle,
                                                  const ACE_Addr & remote_sap,
                                                  const void* act,
                                                  ACE_HANDLE event,
                                                  int priority,
                                                  int signal_number)
{
  TRB_Asynch_Connect_Result_Impl *implementation;
  ACE_NEW_RETURN (implementation,
                  TRB_POSIX_Asynch_Connect_Result (handler,
                                                   connect_handle,
                                                   remote_sap,
                                                   act,
                                                   event,
                                                   priority,
                                                   signal_number),
                  0);
  return implementation;
}


TRB_Asynch_Transmit_File_Impl *
TRB_POSIX_Proactor::create_asynch_transmit_file (void)
{
  TRB_Asynch_Transmit_File_Impl *implementation = 0;
  ACE_NEW_RETURN (implementation,
                  TRB_POSIX_Asynch_Transmit_File (this),
                  0);
  return implementation;
}

TRB_Asynch_Transmit_File_Result_Impl *
TRB_POSIX_Proactor::create_asynch_transmit_file_result (TRB_Handler &handler,
                                                        ACE_HANDLE socket,
                                                        ACE_HANDLE file,
                                                        TRB_Asynch_Transmit_File::Header_And_Trailer *header_and_trailer,
                                                        size_t bytes_to_write,
                                                        u_long offset,
                                                        u_long offset_high,
                                                        size_t bytes_per_send,
                                                        u_long flags,
                                                        const void *act,
                                                        ACE_HANDLE event,
                                                        int priority,
                                                        int signal_number)
{
  TRB_Asynch_Transmit_File_Result_Impl *implementation;
  ACE_NEW_RETURN (implementation,
                  TRB_POSIX_Asynch_Transmit_File_Result (handler,
                                                         socket,
                                                         file,
                                                         header_and_trailer,
                                                         bytes_to_write,
                                                         offset,
                                                         offset_high,
                                                         bytes_per_send,
                                                         flags,
                                                         act,
                                                         event,
                                                         priority,
                                                         signal_number),
                  0);
  return implementation;
}

TRB_Asynch_Result_Impl *
TRB_POSIX_Proactor::create_asynch_timer (TRB_Handler &handler,
                                         const void *act,
                                         const ACE_Time_Value &tv,
                                         ACE_HANDLE event,
                                         int priority,
                                         int signal_number)
{
  TRB_Asynch_Result_Impl *implementation;
  ACE_NEW_RETURN (implementation,
                  TRB_POSIX_Asynch_Timer (handler,
                                          act,
                                          tv,
                                          event,
                                          priority,
                                          signal_number),
                  0);
  return implementation;
}
// *********************************************************************

int
TRB_POSIX_Proactor::start_read_stream (TRB_Handler &handler,
                                       ACE_HANDLE handle,
                                       ACE_Message_Block &message_block,
                                       size_t bytes_to_read,
                                       const void *act,
                                       ACE_HANDLE event,
                                       int priority,
                                       int signal_number)
{
  ACE_MT (ACE_GUARD_RETURN (ACE_SYNCH_MUTEX, ace_mon, this->mutex_, -1));

  void * chunk = this->allocate_any_result ();

  TRB_POSIX_Asynch_Read_Stream_Result * result = 0;

  ACE_NEW_MALLOC_RETURN 
    (result,
     ACE_static_cast (TRB_POSIX_Asynch_Read_Stream_Result *, chunk),
     TRB_POSIX_Asynch_Read_Stream_Result (handler,
                                          handle,
                                          message_block,
                                          bytes_to_read,
                                          act,
                                          event,
                                          priority,
                                          signal_number),
      -1);

  result->allocator_ = this->aio_allocator_;
  result->chunk_ = chunk;

  int rc = this->start_aio (result);

  return rc;
}                
                
int
TRB_POSIX_Proactor::start_write_stream (TRB_Handler &handler,
                                        ACE_HANDLE handle,
                                        ACE_Message_Block &message_block,
                                        size_t bytes_to_write,
                                        const void *act,
                                        ACE_HANDLE event,
                                        int priority,
                                        int signal_number)
{
  ACE_MT (ACE_GUARD_RETURN (ACE_SYNCH_MUTEX, ace_mon, this->mutex_, -1));

  void * chunk = this->allocate_any_result ();

  TRB_POSIX_Asynch_Write_Stream_Result * result = 0;
  ACE_NEW_MALLOC_RETURN 
    (result,
     ACE_static_cast (TRB_POSIX_Asynch_Write_Stream_Result *, chunk),
     TRB_POSIX_Asynch_Write_Stream_Result (handler,
                                           handle,
                                           message_block,
                                           bytes_to_write,
                                           act,
                                           event,
                                           priority,
                                           signal_number),
     -1);

  result->allocator_ = this->aio_allocator_;
  result->chunk_ = chunk;

  int rc = this->start_aio (result);

  return rc;
}  
              
int                
TRB_POSIX_Proactor::start_read_file (TRB_Handler &handler,
                                     ACE_HANDLE handle,
                                     ACE_Message_Block &message_block,
                                     size_t bytes_to_read,
                                     const void *act,
                                     u_long offset,
                                     u_long offset_high,
                                     ACE_HANDLE event,
                                     int priority,
                                     int signal_number)
{
  ACE_MT (ACE_GUARD_RETURN (ACE_SYNCH_MUTEX, ace_mon, this->mutex_, -1));

  void * chunk = this->allocate_any_result ();

  TRB_POSIX_Asynch_Read_File_Result *  result = 0;
  ACE_NEW_MALLOC_RETURN 
    (result,
     ACE_static_cast (TRB_POSIX_Asynch_Read_File_Result *, chunk),
     TRB_POSIX_Asynch_Read_File_Result (handler,
                                        handle,
                                        message_block,
                                        bytes_to_read,
                                        act,
                                        offset,
                                        offset_high,
                                        event,
                                        priority,
                                        signal_number),
     -1);

  result->allocator_ = this->aio_allocator_;
  result->chunk_ = chunk;

  int rc = this->start_aio (result);

  return rc;
}  

int
TRB_POSIX_Proactor::start_write_file (TRB_Handler &handler,
                                      ACE_HANDLE handle,
                                      ACE_Message_Block &message_block,
                                      size_t bytes_to_write,
                                      const void *act,
                                      u_long offset,
                                      u_long offset_high,
                                      ACE_HANDLE event,
                                      int priority,
                                      int signal_number)

{
  ACE_MT (ACE_GUARD_RETURN (ACE_SYNCH_MUTEX, ace_mon, this->mutex_, -1));

  void * chunk = this->allocate_any_result ();

  TRB_POSIX_Asynch_Write_File_Result *  result = 0;
  ACE_NEW_MALLOC_RETURN 
    (result,
     ACE_static_cast (TRB_POSIX_Asynch_Write_File_Result*, chunk),
     TRB_POSIX_Asynch_Write_File_Result (handler,
                                        handle,
                                        message_block,
                                        bytes_to_write,
                                        act,
                                        offset,
                                        offset_high,
                                        event,
                                        priority,
                                        signal_number),
     -1);

  result->allocator_ = this->aio_allocator_;
  result->chunk_ = chunk;

  int rc = this->start_aio (result);

  return rc;
}
  
int
TRB_POSIX_Proactor::start_read_dgram (TRB_Handler &handler,
                                      ACE_HANDLE handle,
                                      ACE_Message_Block *message_block,
                                      size_t bytes_to_read,
                                      int flags,
                                      int protocol_family,
                                      const void* act,
                                      ACE_HANDLE event,
                                      int priority,
                                      int signal_number)

{
  ACE_MT (ACE_GUARD_RETURN (ACE_SYNCH_MUTEX, ace_mon, this->mutex_, -1));

  void * chunk = this->allocate_any_result ();

  TRB_POSIX_Asynch_Read_Dgram_Result * result=0;
  ACE_NEW_MALLOC_RETURN 
    (result,
     ACE_static_cast (TRB_POSIX_Asynch_Read_Dgram_Result *, chunk),
     TRB_POSIX_Asynch_Read_Dgram_Result(handler,
                                        handle,
                                        message_block,
                                        bytes_to_read,
                                        flags,
                                        protocol_family,
                                        act,
                                        event,
                                        priority,
                                        signal_number),
     -1);

  result->allocator_ = this->aio_allocator_;
  result->chunk_ = chunk;

  int rc = this->start_aio (result);

  return rc;
} 
 
int
TRB_POSIX_Proactor::start_write_dgram (TRB_Handler &handler,
                                       ACE_HANDLE handle,
                                       const ACE_Addr & remote_sap,
                                       ACE_Message_Block *message_block,
                                       size_t bytes_to_write,
                                       int flags,
                                       const void* act,
                                       ACE_HANDLE event,
                                       int priority,
                                       int signal_number)

{
  ACE_MT (ACE_GUARD_RETURN (ACE_SYNCH_MUTEX, ace_mon, this->mutex_, -1));

  void * chunk = this->allocate_any_result ();

  TRB_POSIX_Asynch_Write_Dgram_Result * result=0;
  ACE_NEW_MALLOC_RETURN 
    (result,
     ACE_static_cast (TRB_POSIX_Asynch_Write_Dgram_Result *, chunk),
     TRB_POSIX_Asynch_Write_Dgram_Result(handler,
                                         handle,
                                         remote_sap,
                                         message_block,
                                         bytes_to_write,
                                         flags,
                                         act,
                                         event,
                                         priority,
                                         signal_number),
     -1);

  result->allocator_ = this->aio_allocator_;
  result->chunk_ = chunk;

  int rc = this->start_aio (result);

  return rc;
}
  
int
TRB_POSIX_Proactor::start_accept (TRB_Handler &handler,
                                  ACE_HANDLE listen_handle,
                                  ACE_HANDLE accept_handle,
                                  ACE_Message_Block &message_block,
                                  size_t bytes_to_read,
                                  const void *act,
                                  ACE_HANDLE event,
                                  int priority,
                                  int signal_number)
                

{
  ACE_MT (ACE_GUARD_RETURN (ACE_SYNCH_MUTEX, ace_mon, this->mutex_, -1));


  void * chunk = this->allocate_any_result ();

  TRB_POSIX_Asynch_Accept_Result * result =0;
  ACE_NEW_MALLOC_RETURN 
    (result,
     ACE_static_cast (TRB_POSIX_Asynch_Accept_Result *,chunk),
     TRB_POSIX_Asynch_Accept_Result (handler,
                                     listen_handle,
                                     accept_handle,
                                     message_block,
                                     bytes_to_read,
                                     act,
                                     event,
                                     priority,
                                     signal_number),
     -1);

  result->allocator_ = this->aio_allocator_;
  result->chunk_ = chunk;

  int rc = this->start_aio (result);

  return rc;
} 
 
int
TRB_POSIX_Proactor::start_connect (TRB_Handler & handler,
                                   ACE_HANDLE connect_handle,
                                   const ACE_Addr & remote_sap,
                                   const void *act,
                                   ACE_HANDLE event,
                                   int priority,
                                   int signal_number)
                
{
  ACE_MT (ACE_GUARD_RETURN (ACE_SYNCH_MUTEX, ace_mon, this->mutex_, -1));

  void * chunk = this->allocate_any_result ();

  TRB_POSIX_Asynch_Connect_Result  * result =0;
  ACE_NEW_MALLOC_RETURN 
    (result,
     ACE_static_cast (TRB_POSIX_Asynch_Connect_Result *, chunk),
     TRB_POSIX_Asynch_Connect_Result (handler,
                                      connect_handle,
                                      remote_sap,
                                      act,
                                      event,
                                      priority,
                                      signal_number),
     -1);

  result->allocator_ = this->aio_allocator_;
  result->chunk_ = chunk;

  int rc = this->start_aio (result);

  return rc;
}  
                
// **********************************************************************

TRB_POSIX_AIOCB_Proactor::TRB_POSIX_AIOCB_Proactor (size_t max_op, 
                                                    int sig_num,
                                                    int leader_type,
                                                    int start_aio_type)
  : TRB_POSIX_Proactor 
      (TRB_POSIX_AIO_Config (max_op,
                             (leader_type != 0)
                               ? TRB_POSIX_AIO_Config::PCT_DEDICATED
                               : TRB_POSIX_AIO_Config::PCT_SHARED,
                             TRB_POSIX_AIO_Config::PCF_ZERO,
                             TRB_POSIX_AIO_Config::PVT_POSIX,
                             (start_aio_type != 0)
                               ?  TRB_POSIX_AIO_Config::PVF_AIO_START_ANY_THREAD |
                                  TRB_POSIX_AIO_Config::PVF_AIO_START_INTERRUPT_WAIT
                               :  0,
                             TRB_POSIX_AIO_Config::WST_DEFAULT,
                             ACE_SIGRTMIN,
                             (sig_num == 0)
                               ? TRB_POSIX_AIO_Config::IST_PIPE
                               : TRB_POSIX_AIO_Config::IST_SIGNAL,
                             sig_num),
       TRB_POSIX_AIO_Config (max_op,
                             TRB_POSIX_AIO_Config::PCT_DEDICATED,
                             TRB_POSIX_AIO_Config::PCF_ZERO,
                             TRB_POSIX_AIO_Config::PVT_SELECT,
                             TRB_POSIX_AIO_Config::PVF_AIO_START_ANY_THREAD |
                             TRB_POSIX_AIO_Config::PVF_AIO_START_INTERRUPT_WAIT,
                             TRB_POSIX_AIO_Config::WST_DEFAULT,
                             ACE_SIGRTMIN,
                             (sig_num == 0) 
                               ? TRB_POSIX_AIO_Config::IST_PIPE
                               : TRB_POSIX_AIO_Config::IST_SIGNAL,
                             sig_num))
{
}

TRB_POSIX_SIG_Proactor::TRB_POSIX_SIG_Proactor (size_t max_op,
                                                int sig_num,
                                                int leader_type,
                                                int start_aio_type)
  : TRB_POSIX_Proactor 
      (TRB_POSIX_AIO_Config (max_op,
                             (leader_type != 0)
                               ? TRB_POSIX_AIO_Config::PCT_DEDICATED
                               : TRB_POSIX_AIO_Config::PCT_SHARED,
                             TRB_POSIX_AIO_Config::PCF_ZERO,
                             TRB_POSIX_AIO_Config::PVT_POSIX,
                             (start_aio_type != 0)
                               ?  TRB_POSIX_AIO_Config::PVF_AIO_START_ANY_THREAD 
                               :  0,
                             TRB_POSIX_AIO_Config::WST_SIGWAIT,
                             ACE_SIGRTMIN,
                             (sig_num == 0) 
                               ? TRB_POSIX_AIO_Config::IST_PIPE
                               : TRB_POSIX_AIO_Config::IST_SIGNAL,
                             sig_num),
       TRB_POSIX_AIO_Config (max_op,
                             TRB_POSIX_AIO_Config::PCT_DEDICATED,
                             TRB_POSIX_AIO_Config::PCF_ZERO,
                             TRB_POSIX_AIO_Config::PVT_SELECT,
                             TRB_POSIX_AIO_Config::PVF_AIO_START_ANY_THREAD |
                             TRB_POSIX_AIO_Config::PVF_AIO_START_INTERRUPT_WAIT,
                             TRB_POSIX_AIO_Config::WST_DEFAULT,
                             ACE_SIGRTMIN,
                             (sig_num == 0)
                               ? TRB_POSIX_AIO_Config::IST_PIPE
                               : TRB_POSIX_AIO_Config::IST_SIGNAL,
                             sig_num))
{
}

// ********************************************************************

TRB_Select_Proactor::TRB_Select_Proactor (size_t max_op, 
                                          int sig_num,
                                          int leader_type,
                                          int start_aio_type)
  : TRB_POSIX_Proactor 
      (TRB_POSIX_AIO_Config (max_op,
                             (leader_type != 0)
                               ? TRB_POSIX_AIO_Config::PCT_DEDICATED
                               : TRB_POSIX_AIO_Config::PCT_SHARED,
                             TRB_POSIX_AIO_Config::PCF_ZERO,
                             TRB_POSIX_AIO_Config::PVT_SELECT,
                             (start_aio_type != 0)
                               ?  TRB_POSIX_AIO_Config::PVF_AIO_START_ANY_THREAD |
                                  TRB_POSIX_AIO_Config::PVF_AIO_START_INTERRUPT_WAIT
                               :  0,
                             TRB_POSIX_AIO_Config::WST_SELECT,
                             ACE_SIGRTMIN,
                             (sig_num == 0)
                               ? TRB_POSIX_AIO_Config::IST_PIPE
                               : TRB_POSIX_AIO_Config::IST_SIGNAL,
                             sig_num),
       TRB_POSIX_AIO_Config (max_op,
                            TRB_POSIX_AIO_Config::PCT_NONE)) // NOT EXIST
{
}

TRB_Poll_Proactor::TRB_Poll_Proactor (size_t max_op, 
                                      int sig_num,
                                      int leader_type,
                                      int start_aio_type)
  : TRB_POSIX_Proactor 
      (TRB_POSIX_AIO_Config (max_op,
                             (leader_type != 0)
                               ? TRB_POSIX_AIO_Config::PCT_DEDICATED
                               : TRB_POSIX_AIO_Config::PCT_SHARED,
                             TRB_POSIX_AIO_Config::PCF_ZERO,
                             TRB_POSIX_AIO_Config::PVT_SELECT,
                             (start_aio_type != 0)
                               ?  TRB_POSIX_AIO_Config::PVF_AIO_START_ANY_THREAD |
                                  TRB_POSIX_AIO_Config::PVF_AIO_START_INTERRUPT_WAIT
                               :  0,
                             TRB_POSIX_AIO_Config::WST_POLL,
                             ACE_SIGRTMIN,
                             (sig_num == 0)
                               ? TRB_POSIX_AIO_Config::IST_PIPE
                               : TRB_POSIX_AIO_Config::IST_SIGNAL,
                             sig_num),
       TRB_POSIX_AIO_Config (max_op,
                            TRB_POSIX_AIO_Config::PCT_NONE)) // NOT EXIST
{
}

TRB_Dev_Poll_Proactor::TRB_Dev_Poll_Proactor (size_t max_op, 
                                              int sig_num,
                                              int leader_type,
                                              int start_aio_type)
  : TRB_POSIX_Proactor 
      (TRB_POSIX_AIO_Config (max_op,
                             (leader_type != 0)
                               ? TRB_POSIX_AIO_Config::PCT_DEDICATED
                               : TRB_POSIX_AIO_Config::PCT_SHARED,
                             TRB_POSIX_AIO_Config::PCF_ZERO,
                             TRB_POSIX_AIO_Config::PVT_SELECT,
//                             (start_aio_type != 0)
//                               ?  TRB_POSIX_AIO_Config::PVF_AIO_START_ANY_THREAD 
//                               :  0,
//
                             TRB_POSIX_AIO_Config::PVF_AIO_START_INTERRUPT_WAIT,
//
                             TRB_POSIX_AIO_Config::WST_DEVPOLL,
                             ACE_SIGRTMIN,
                             (sig_num == 0)
                               ? TRB_POSIX_AIO_Config::IST_PIPE
                               : TRB_POSIX_AIO_Config::IST_SIGNAL,
                             sig_num),
       TRB_POSIX_AIO_Config (max_op,
                            TRB_POSIX_AIO_Config::PCT_NONE)) // NOT EXIST
{
    ACE_UNUSED_ARG (start_aio_type);
}

TRB_Event_Poll_Proactor::TRB_Event_Poll_Proactor (size_t max_op, 
                                              int sig_num,
                                              int leader_type,
                                              int start_aio_type)
  : TRB_POSIX_Proactor 
      (TRB_POSIX_AIO_Config (max_op,
                             (leader_type != 0)
                               ? TRB_POSIX_AIO_Config::PCT_DEDICATED
                               : TRB_POSIX_AIO_Config::PCT_SHARED,
                             TRB_POSIX_AIO_Config::PCF_ZERO,
                             TRB_POSIX_AIO_Config::PVT_SELECT,
                             (start_aio_type != 0)
                               ?  TRB_POSIX_AIO_Config::PVF_AIO_START_ANY_THREAD 
                               :  0,
                             TRB_POSIX_AIO_Config::WST_EPOLL,
                             ACE_SIGRTMIN,
                             (sig_num == 0)
                               ? TRB_POSIX_AIO_Config::IST_PIPE
                               : TRB_POSIX_AIO_Config::IST_SIGNAL,
                             sig_num),
       TRB_POSIX_AIO_Config (max_op,
                            TRB_POSIX_AIO_Config::PCT_NONE)) // NOT EXIST
{
}

TRB_LINUX_Proactor::TRB_LINUX_Proactor (size_t max_op, 
                                        int sig_num,
                                        int leader_type,
                                        int start_aio_type)
  : TRB_POSIX_Proactor 
      (TRB_POSIX_AIO_Config (max_op,
                             (leader_type != 0)
                               ? TRB_POSIX_AIO_Config::PCT_DEDICATED
                               : TRB_POSIX_AIO_Config::PCT_SHARED,
                             TRB_POSIX_AIO_Config::PCF_ZERO,
                             TRB_POSIX_AIO_Config::PVT_LINUX,
                             (start_aio_type != 0)
                               ?  TRB_POSIX_AIO_Config::PVF_AIO_START_ANY_THREAD 
                               :  0,
                             TRB_POSIX_AIO_Config::WST_NATIVE,
                             ACE_SIGRTMIN,
                             (sig_num == 0)
                               ? TRB_POSIX_AIO_Config::IST_PIPE
                               : TRB_POSIX_AIO_Config::IST_SIGNAL,
                             sig_num),
       TRB_POSIX_AIO_Config (max_op,
                             TRB_POSIX_AIO_Config::PCT_DEDICATED,
                             TRB_POSIX_AIO_Config::PCF_ZERO,
                             TRB_POSIX_AIO_Config::PVT_SELECT,
                             TRB_POSIX_AIO_Config::PVF_AIO_START_ANY_THREAD |
                             TRB_POSIX_AIO_Config::PVF_AIO_START_INTERRUPT_WAIT,
                             TRB_POSIX_AIO_Config::WST_DEFAULT,
                             ACE_SIGRTMIN,
                             (sig_num == 0) 
                               ? TRB_POSIX_AIO_Config::IST_PIPE
                               : TRB_POSIX_AIO_Config::IST_SIGNAL,
                             sig_num))
{
  ACE_UNUSED_ARG (leader_type);
}

TRB_Linux_RT_Proactor::TRB_Linux_RT_Proactor (size_t max_op, 
                                              int sig_num,
                                              int leader_type,
                                              int start_aio_type)
  : TRB_POSIX_Proactor 
      (TRB_POSIX_AIO_Config (max_op,
                             TRB_POSIX_AIO_Config::PCT_DEDICATED,
                             TRB_POSIX_AIO_Config::PCF_ZERO,
                             TRB_POSIX_AIO_Config::PVT_SELECT,
                             (start_aio_type != 0)
                               ?  TRB_POSIX_AIO_Config::PVF_AIO_START_ANY_THREAD 
                               :  0,
                             TRB_POSIX_AIO_Config::WST_LINUXRT,
                             ACE_SIGRTMIN,
                             (sig_num == 0)
                               ? TRB_POSIX_AIO_Config::IST_PIPE
                               : TRB_POSIX_AIO_Config::IST_SIGNAL,
                             sig_num),
       TRB_POSIX_AIO_Config (max_op,
                            TRB_POSIX_AIO_Config::PCT_NONE)) // NOT EXIST
{
  ACE_UNUSED_ARG (leader_type);
}

TRB_SUN_Port_Proactor::TRB_SUN_Port_Proactor (size_t max_op, 
                                              int sig_num,
                                              int leader_type,
                                              int start_aio_type)
  : TRB_POSIX_Proactor 
      (TRB_POSIX_AIO_Config (max_op,
                             (leader_type != 0)
                               ? TRB_POSIX_AIO_Config::PCT_DEDICATED
                               : TRB_POSIX_AIO_Config::PCT_SHARED,
                             TRB_POSIX_AIO_Config::PCF_ZERO,
                             TRB_POSIX_AIO_Config::PVT_SELECT,
                             (start_aio_type != 0)
                               ?  TRB_POSIX_AIO_Config::PVF_AIO_START_ANY_THREAD 
                               :  0,
                             TRB_POSIX_AIO_Config::WST_SUN_PORT,
                             ACE_SIGRTMIN,
                             (sig_num == 0)
                               ? TRB_POSIX_AIO_Config::IST_PIPE
                               : TRB_POSIX_AIO_Config::IST_SIGNAL,
                             sig_num),
       TRB_POSIX_AIO_Config (max_op,
                            TRB_POSIX_AIO_Config::PCT_NONE)) // NOT EXIST
{
}

TRB_Dummy_Proactor::TRB_Dummy_Proactor ()
  : TRB_POSIX_Proactor 
      (TRB_POSIX_AIO_Config (0,
                            TRB_POSIX_AIO_Config::PCT_NONE), // NOT EXIST
       TRB_POSIX_AIO_Config (0,
                            TRB_POSIX_AIO_Config::PCT_NONE)) // NOT EXIST
{
}
// *****************************************************************

#if defined (ACE_HAS_EXPLICIT_TEMPLATE_INSTANTIATION)

template class ACE_Cached_Mem_Pool_Node<TRB_POSIX_MAX_Result>;
template class ACE_Free_List <ACE_Cached_Mem_Pool_Node<TRB_POSIX_MAX_Result> >;
template class ACE_Locked_Free_List <ACE_Cached_Mem_Pool_Node<TRB_POSIX_MAX_Result>, ACE_SYNCH_NULL_MUTEX>;
//template class ACE_Cached_Allocator <TRB_POSIX_MAX_Result, ACE_SYNCH_NULL_MUTEX>;
template class TRB_Cached_List_Allocator <TRB_POSIX_MAX_Result, ACE_SYNCH_NULL_MUTEX>;

#elif defined (ACE_HAS_TEMPLATE_INSTANTIATION_PRAGMA)

#pragma instantiate ACE_Cached_Mem_Pool_Node<TRB_POSIX_MAX_Result>
#pragma instantiate ACE_Free_List <ACE_Cached_Mem_Pool_Node<TRB_POSIX_MAX_Result> >
#pragma instantiate ACE_Locked_Free_List <ACE_Cached_Mem_Pool_Node<TRB_POSIX_MAX_Result>, ACE_SYNCH_NULL_MUTEX>
//#pragma instantiate ACE_Cached_Allocator <TRB_POSIX_MAX_Result, ACE_SYNCH_NULL_MUTEX>
#pragma instantiate TRB_Cached_List_Allocator <TRB_POSIX_MAX_Result, ACE_SYNCH_NULL_MUTEX>

#endif /* ACE_HAS_EXPLICIT_TEMPLATE_INSTANTIATION */

#endif /* ACE_HAS_AIO_CALLS */
