/* -*- C++ -*- */

//=============================================================================
/**
 *  @file    POSIX_Proactor.h
 *
 *  POSIX_Proactor.h,v 4.44 2002/12/06 03:20:24 shuston Exp
 *
 *  @author Irfan Pyarali <irfan@cs.wustl.edu>
 *  @author Tim Harrison <harrison@cs.wustl.edu>
 *  @author Alexander Babu Arulanthu <alex@cs.wustl.edu>
 *  @author Roger Tragin <r.tragin@computer.org>
 *  @author Alexander Libman <alibman@baltimore.com>
 */
//=============================================================================

#ifndef TRB_POSIX_PROACTOR_H 
#define TRB_POSIX_PROACTOR_H 

#include "TProactor/TPROACTOR_Export.h"

#if !defined (ACE_LACKS_PRAGMA_ONCE)
#pragma once
#endif /* ACE_LACKS_PRAGMA_ONCE */

#if defined (ACE_HAS_AIO_CALLS) || defined(ACE_HAS_AIO_EMULATION)
// POSIX implementation of Proactor depends on the <aio_> family of
// system calls.

#include "TProactor/Proactor_Impl.h"
#include "TProactor/POSIX_AIO_Processor.h"
#include "TProactor/POSIX_AIO_Dispatcher.h"

class ACE_Allocator;

// *********************************************************************
/**
 * @class TRB_POSIX_Wakeup_Completion
 *
 * This result object is used by the <end_event_loop> of the
 * TRB_Proactor interface to wake up all the threads blocking
 * for completions.
 */
class TPROACTOR_Export TRB_POSIX_Wakeup_Completion : public TRB_POSIX_Asynch_Result
{
public:
  /// Constructor.
  TRB_POSIX_Wakeup_Completion (TRB_Handler &handler,
                               const void *act = 0,
                               ACE_HANDLE event = ACE_INVALID_HANDLE,
                               int priority = 0,
                               int signal_number = ACE_SIGRTMIN);

  /// Destructor.
  virtual ~TRB_POSIX_Wakeup_Completion (void);


  /// This method calls the <handler>'s <handle_wakeup> method.
  virtual void complete (size_t bytes_transferred = 0,
                         int success = 1,
                         const void *completion_key = 0,
                         u_long error = 0);
};

// *********************************************************************
/**
 * @class TRB_POSIX_Asynch_Timer
 *
 * @brief This class is posted to the completion port when a timer
 * expires. When the <complete method> of this object is
 * called, the <handler>'s <handle_timeout> method will be
 * called.
 */
class TPROACTOR_Export TRB_POSIX_Asynch_Timer : public TRB_POSIX_Asynch_Result
{

  /// The factory method for this class is with the POSIX_Proactor
  /// class.
  friend class TRB_POSIX_Proactor;

protected:
  /// Constructor.
  TRB_POSIX_Asynch_Timer (TRB_Handler &handler,
                          const void *act,
                          const ACE_Time_Value &tv,
                          ACE_HANDLE event = ACE_INVALID_HANDLE,
                          int priority = 0,
                          int signal_number = ACE_SIGRTMIN);

  /// Destructor.
  virtual ~TRB_POSIX_Asynch_Timer (void) {}

  /// This method calls the <handler>'s handle_timeout method.
  virtual void complete (size_t bytes_transferred,
                         int success,
                         const void *completion_key,
                         u_long error = 0);

  /// Time value requested by caller
  ACE_Time_Value time_;
};

// *********************************************************************
union U_POSIX_MAX_Result
{
    void * p1; 
    char c1  [sizeof(TRB_POSIX_Asynch_Read_Stream_Result)];
    char c2  [sizeof(TRB_POSIX_Asynch_Write_Stream_Result)];
    char c3  [sizeof(TRB_POSIX_Asynch_Read_File_Result)];
    char c4  [sizeof(TRB_POSIX_Asynch_Write_File_Result)];
    char c5  [sizeof(TRB_POSIX_Asynch_Read_Dgram_Result)];
    char c6  [sizeof(TRB_POSIX_Asynch_Write_Dgram_Result)];
    char c7  [sizeof(TRB_POSIX_Asynch_Accept_Result)];
    char c8  [sizeof(TRB_POSIX_Asynch_Connect_Result)];
    char c9  [sizeof(TRB_POSIX_Asynch_Transmit_File_Result)];
    char c10 [sizeof(TRB_POSIX_Asynch_Timer)];
    char c11 [sizeof(TRB_POSIX_Wakeup_Completion)];
};

class TPROACTOR_Export TRB_POSIX_MAX_Result
{
public:
    U_POSIX_MAX_Result fill_data_;
};

// *********************************************************************
/**
 * @class TRB_POSIX_Proactor
 *
 * @brief POSIX implementation of the Proactor.
 *
 */
class TPROACTOR_Export TRB_POSIX_Proactor : public TRB_Proactor_Impl
{
public:
  enum Proactor_Type
  {
    /// Base class type
    PROACTOR_POSIX  = 0,

    /// Aio_suspend() based
    PROACTOR_AIOCB  = 1,

    /// Signals notifications
    PROACTOR_SIG    = 2,

    /// SUN specific aiowait()
    PROACTOR_SUN    = 3,

    /// Callback notifications
    PROACTOR_CB     = 4,
    
    /// Linux io_submit/io_getevents
    PROACTOR_LINUX  = 5
  };


  enum SystemType  // open for future extention
  {
    OS_UNDEFINED= 0x0000,
    OS_WIN      = 0x0100,          // for future
    OS_WIN_NT   = OS_WIN | 0x0001,
    OS_WIN_2000 = OS_WIN | 0x0002,
    OS_SUN      = 0x0200,          // Sun Solaris family
    OS_SUN_55   = OS_SUN | 0x0001,
    OS_SUN_56   = OS_SUN | 0x0002,
    OS_SUN_57   = OS_SUN | 0x0004,
    OS_SUN_58   = OS_SUN | 0x0008,
    OS_HPUX     = 0x0400,          // HPUX family
    OS_HPUX_11  = OS_HPUX | 0x0001,
    OS_LINUX    = 0x0800,          // Linux family
    OS_FREEBSD  = 0x1000,          // FreeBSD family
    OS_IRIX     = 0x2000,          // SGI IRIX family
    OS_OPENBSD  = 0x4000           // OpenBSD familty
  };

  //enum Opcode {
  //  READ = 1,
  //  WRITE = 2
  //};

  /// Constructor.
  TRB_POSIX_Proactor (size_t max_op = ACE_AIO_DEFAULT_SIZE,
                     int sig_num = 0,         // interrupt signal , 0-use pipe 
                     int leader_type = 0,     // 0-shared /1-dedicated
                     int start_aio_type = 1); // 1-any thread/ 0-only leader

  /// Build Lego-Proactor for your own style;
  TRB_POSIX_Proactor (const TRB_POSIX_AIO_Config & prime_processor_cfg,
                      const TRB_POSIX_AIO_Config & second_processor_cfg);


  /// Virtual destructor.
  virtual ~TRB_POSIX_Proactor (void);

  /// Open The Proactor
  int open (void);

  /// Close down the Proactor.
  int close (void);

  /// 
  virtual Proactor_Type  get_impl_type (void);

  /**
   * Dispatch a single set of events.  If <wait_time> elapses before
   * any events occur, return 0.  Return 1 on success i.e., when a
   * completion is dispatched, non-zero (-1) on errors and errno is
   * set accordingly.
   */
  virtual int handle_events (ACE_Time_Value &wait_time);

  /**
   * Block indefinitely until at least one event is dispatched.
   * Dispatch a single set of events.  If <wait_time> elapses before
   * any events occur, return 0.  Return 1 on success i.e., when a
   * completion is dispatched, non-zero (-1) on errors and errno is
   * set accordingly.
   */
  virtual int handle_events (void);

  /**
   * Post a result to the completion port of the Proactor.  If errors
   * occur, the result will be deleted by this method.  If successful,
   * the result will be deleted by the Proactor when the result is
   * removed from the completion port. Therefore, the result should
   * have been dynamically allocated and should be orphaned by the
   * user once this method is called.
   */
  virtual int post_completion (TRB_POSIX_Asynch_Result *result);

  
  virtual int cancel_aio (ACE_HANDLE h);
  virtual int cancel_aio (TRB_Handler * handler);
  virtual int cancel_all_aio (void);

  
  /// This function is a no-op function for Unix systems. Returns 0.
  /// Not now
  virtual int register_handle (ACE_HANDLE handle,
                               const void *completion_key);

  virtual int unregister_handle (ACE_HANDLE handle);

  /// @@ This is a no-op on POSIX platforms. Returns 0.
  int wake_up_dispatch_threads (void);

  /// @@ This is a no-op on POSIX platforms. Returns 0.
  int close_dispatch_threads (int wait);

  /// @@ This is a no-op on POSIX platforms. Returns 0.
  size_t number_of_threads (void) const;
  void number_of_threads (size_t threads);

  /// This is a no-op in POSIX. Returns ACE_INVALID_HANDLE.
  virtual ACE_HANDLE get_handle (void) const;


  // Methods used to create Asynch IO factory and result objects. We
  // create the right objects here in these methods.

  virtual TRB_Asynch_Read_Stream_Impl *create_asynch_read_stream (void);
  virtual TRB_Asynch_Read_Stream_Result_Impl *
    create_asynch_read_stream_result (TRB_Handler &handler,
                                      ACE_HANDLE handle,
                                      ACE_Message_Block &message_block,
                                      size_t bytes_to_read,
                                      const void *act,
                                      ACE_HANDLE event = ACE_INVALID_HANDLE,
                                      int priority = 0,
                                      int signal_number = ACE_SIGRTMIN);

  virtual TRB_Asynch_Write_Stream_Impl *create_asynch_write_stream (void);
  virtual TRB_Asynch_Write_Stream_Result_Impl *
    create_asynch_write_stream_result (TRB_Handler &handler,
                                       ACE_HANDLE handle,
                                       ACE_Message_Block &message_block,
                                       size_t bytes_to_write,
                                       const void *act,
                                       ACE_HANDLE event = ACE_INVALID_HANDLE,
                                       int priority = 0,
                                       int signal_number = ACE_SIGRTMIN);

  virtual TRB_Asynch_Read_File_Impl *create_asynch_read_file (void);
  virtual TRB_Asynch_Read_File_Result_Impl *
    create_asynch_read_file_result (TRB_Handler &handler,
                                    ACE_HANDLE handle,
                                    ACE_Message_Block &message_block,
                                    size_t bytes_to_read,
                                    const void *act,
                                    u_long offset,
                                    u_long offset_high,
                                    ACE_HANDLE event = ACE_INVALID_HANDLE,
                                    int priority = 0,
                                    int signal_number = ACE_SIGRTMIN);

  virtual TRB_Asynch_Write_File_Impl *create_asynch_write_file (void);
  virtual TRB_Asynch_Write_File_Result_Impl *
    create_asynch_write_file_result (TRB_Handler &handler,
                                     ACE_HANDLE handle,
                                     ACE_Message_Block &message_block,
                                     size_t bytes_to_write,
                                     const void *act,
                                     u_long offset,
                                     u_long offset_high,
                                     ACE_HANDLE event = ACE_INVALID_HANDLE,
                                     int priority = 0,
                                     int signal_number = ACE_SIGRTMIN);

  virtual TRB_Asynch_Read_Dgram_Impl *create_asynch_read_dgram (void);
  virtual TRB_Asynch_Read_Dgram_Result_Impl *
    create_asynch_read_dgram_result (TRB_Handler &handler,
                                     ACE_HANDLE handle,
                                     ACE_Message_Block *message_block,
                                     size_t bytes_to_read,
                                     int flags,
                                     int protocol_family,
                                     const void* act,
                                     ACE_HANDLE event = ACE_INVALID_HANDLE,
                                     int priority = 0,
                                     int signal_number = ACE_SIGRTMIN);

  virtual TRB_Asynch_Write_Dgram_Impl *create_asynch_write_dgram (void);
  virtual TRB_Asynch_Write_Dgram_Result_Impl *
    create_asynch_write_dgram_result (TRB_Handler &handler,
                                      ACE_HANDLE handle,
                                      const ACE_Addr & remote_sap,
                                      ACE_Message_Block *message_block,
                                      size_t bytes_to_write,
                                      int flags,
                                      const void* act,
                                      ACE_HANDLE event = ACE_INVALID_HANDLE,
                                      int priority = 0,
                                      int signal_number = ACE_SIGRTMIN);

  virtual TRB_Asynch_Accept_Impl *create_asynch_accept (void);
  virtual TRB_Asynch_Accept_Result_Impl *
    create_asynch_accept_result (TRB_Handler &handler,
                                 ACE_HANDLE listen_handle,
                                 ACE_HANDLE accept_handle,
                                 ACE_Message_Block &message_block,
                                 size_t bytes_to_read,
                                 const void *act,
                                 ACE_HANDLE event = ACE_INVALID_HANDLE,
                                 int priority = 0,
                                 int signal_number = ACE_SIGRTMIN);

  virtual TRB_Asynch_Connect_Impl *create_asynch_connect (void);
  virtual TRB_Asynch_Connect_Result_Impl *
    create_asynch_connect_result (TRB_Handler & handler,
                                  ACE_HANDLE connect_handle,
                                  const ACE_Addr & remote_sap,
                                  const void *act,
                                  ACE_HANDLE event = ACE_INVALID_HANDLE,
                                  int priority = 0,
                                  int signal_number = ACE_SIGRTMIN);

  virtual TRB_Asynch_Transmit_File_Impl *create_asynch_transmit_file (void);
  virtual TRB_Asynch_Transmit_File_Result_Impl *
    create_asynch_transmit_file_result (TRB_Handler &handler,
                                        ACE_HANDLE socket,
                                        ACE_HANDLE file,
                                        TRB_Asynch_Transmit_File::Header_And_Trailer *header_and_trailer,
                                        size_t bytes_to_write,
                                        u_long offset,
                                        u_long offset_high,
                                        size_t bytes_per_send,
                                        u_long flags,
                                        const void *act,
                                        ACE_HANDLE event = ACE_INVALID_HANDLE,
                                        int priority = 0,
                                        int signal_number = ACE_SIGRTMIN);

  /// Create a timer result object which can be used with the Timer
  /// mechanism of the Proactor.
  virtual TRB_Asynch_Result_Impl *
    create_asynch_timer (TRB_Handler &handler,
                         const void *act,
                         const ACE_Time_Value &tv,
                         ACE_HANDLE event = ACE_INVALID_HANDLE,
                         int priority = 0,
                         int signal_number = ACE_SIGRTMIN);




  /// New !!!
  /// Atomic :  allocation result and start aio
  int start_read_stream (TRB_Handler &handler,
                         ACE_HANDLE handle,
                         ACE_Message_Block &message_block,
                         size_t bytes_to_read,
                         const void *act,
                         ACE_HANDLE event = ACE_INVALID_HANDLE,
                         int priority = 0,
                         int signal_number = ACE_SIGRTMIN);

  
  int start_write_stream (TRB_Handler &handler,
                          ACE_HANDLE handle,
                          ACE_Message_Block &message_block,
                          size_t bytes_to_write,
                          const void *act,
                          ACE_HANDLE event = ACE_INVALID_HANDLE,
                          int priority = 0,
                          int signal_number = ACE_SIGRTMIN);

  int start_read_file (TRB_Handler &handler,
                       ACE_HANDLE handle,
                       ACE_Message_Block &message_block,
                       size_t bytes_to_read,
                       const void *act,
                       u_long offset,
                       u_long offset_high,
                       ACE_HANDLE event = ACE_INVALID_HANDLE,
                       int priority = 0,
                       int signal_number = ACE_SIGRTMIN);

  int start_write_file (TRB_Handler &handler,
                        ACE_HANDLE handle,
                        ACE_Message_Block &message_block,
                        size_t bytes_to_write,
                        const void *act,
                        u_long offset,
                        u_long offset_high,
                        ACE_HANDLE event = ACE_INVALID_HANDLE,
                        int priority = 0,
                        int signal_number = ACE_SIGRTMIN);

  int start_read_dgram (TRB_Handler &handler,
                        ACE_HANDLE handle,
                        ACE_Message_Block *message_block,
                        size_t bytes_to_read,
                        int flags,
                        int protocol_family,
                        const void* act,
                        ACE_HANDLE event = ACE_INVALID_HANDLE,
                        int priority = 0,
                        int signal_number = ACE_SIGRTMIN);

  int start_write_dgram (TRB_Handler &handler,
                         ACE_HANDLE handle,
                         const ACE_Addr & remote_sap,
                         ACE_Message_Block *message_block,
                         size_t bytes_to_write,
                         int flags,
                         const void* act,
                         ACE_HANDLE event = ACE_INVALID_HANDLE,
                         int priority = 0,
                         int signal_number = ACE_SIGRTMIN);

  int start_accept (TRB_Handler &handler,
                    ACE_HANDLE listen_handle,
                    ACE_HANDLE accept_handle,
                    ACE_Message_Block &message_block,
                    size_t bytes_to_read,
                    const void *act,
                    ACE_HANDLE event = ACE_INVALID_HANDLE,
                    int priority = 0,
                    int signal_number = ACE_SIGRTMIN);


  int start_connect (TRB_Handler & handler,
                     ACE_HANDLE connect_handle,
                     const ACE_Addr & remote_sap,
                     const void *act,
                     ACE_HANDLE event = ACE_INVALID_HANDLE,
                     int priority = 0,
                     int signal_number = ACE_SIGRTMIN);

protected:

  
  /**
   * Post <how_many> completions to the completion port so that all
   * threads can wake up. This is used in conjunction with the
   * <run_event_loop>.
   */
  virtual int post_wakeup_completions (int how_many);

  static int get_os_id(void);

protected:

  int handle_events_i (ACE_Time_Value * wait_time);

  //virtual int start_aio (TRB_POSIX_Asynch_Result *result);
  int start_aio (TRB_POSIX_Asynch_Result *result);



private:

  int  create_result_allocator (void);
  void delete_result_allocator (void);
  void * allocate_any_result (void);


  /// Handler to handle the wakeups. This works in conjunction with the
  /// <TRB_Proactor::run_event_loop>.
  TRB_Handler wakeup_handler_;

  int os_id_;

  int flg_open_ ;
  int flg_disable_start_;

  ACE_SYNCH_MUTEX  mutex_;

  TRB_POSIX_AIO_Dispatcher  dispatcher_;

  TRB_POSIX_AIO_Config      prime_config_;
  TRB_POSIX_AIO_Config      second_config_;

  TRB_POSIX_AIO_Processor * prime_aio_processor_;
  TRB_POSIX_AIO_Processor * second_aio_processor_;

  ACE_Allocator  * aio_allocator_;
};



/*
 * For old compatibility
 *
 */
class TPROACTOR_Export TRB_POSIX_AIOCB_Proactor: public TRB_POSIX_Proactor 
{
public:
  TRB_POSIX_AIOCB_Proactor (size_t max_op = ACE_AIO_DEFAULT_SIZE,
                          int sig_num = 0,        // use pipe, no interrupt signal
                          int leader_type = 0,    // 0-shared /1-dedicated
                          int start_aio_type = 1); // 1-any thread/ 0-only leader
};


class TPROACTOR_Export TRB_POSIX_SIG_Proactor: public TRB_POSIX_Proactor 
{
public:
  TRB_POSIX_SIG_Proactor (size_t max_op = ACE_AIO_DEFAULT_SIZE,
                          int sig_num = ACE_SIGRTMIN, // interrupt signal
                          int leader_type = 0,    // 0-shared /1-dedicated
                          int start_aio_type = 1); // 1-any thread/ 0-only leader
};


// **********************************************************************

class TPROACTOR_Export TRB_Select_Proactor: public TRB_POSIX_Proactor 
{
public:
  TRB_Select_Proactor (size_t max_op = ACE_AIO_DEFAULT_SIZE,
                          int sig_num = 0,        // use pipe, no interrupt signal
                          int leader_type = 0,    // 0-shared /1-dedicated
                          int start_aio_type = 1); // 1-any thread/ 0-only leader
};

class TPROACTOR_Export TRB_Poll_Proactor: public TRB_POSIX_Proactor 
{
public:
  TRB_Poll_Proactor (size_t max_op = ACE_AIO_DEFAULT_SIZE,
                     int sig_num = 0,        // use pipe, no interrupt signal
                     int leader_type = 0,    // 0-shared /1-dedicated
                     int start_aio_type = 1); // 1-any thread/ 0-only leader
};

class TPROACTOR_Export TRB_Dev_Poll_Proactor: public TRB_POSIX_Proactor 
{
public:
  TRB_Dev_Poll_Proactor (size_t max_op = ACE_AIO_DEFAULT_SIZE,
                          int sig_num = 0,        // use pipe, no interrupt signal
                          int leader_type = 0,    // 0-shared /1-dedicated
                          int start_aio_type = 1); // 1-any thread/ 0-only leader
};

class TPROACTOR_Export TRB_Event_Poll_Proactor: public TRB_POSIX_Proactor 
{
public:
  TRB_Event_Poll_Proactor (size_t max_op = ACE_AIO_DEFAULT_SIZE,
                          int sig_num = 0,        // use pipe, no interrupt signal
                          int leader_type = 0,    // 0-shared /1-dedicated
                          int start_aio_type = 1); // 1-any thread/ 0-only leader
};

class TPROACTOR_Export TRB_LINUX_Proactor: public TRB_POSIX_Proactor 
{
public:
  TRB_LINUX_Proactor (size_t max_op = ACE_AIO_DEFAULT_SIZE,
                          int sig_num = 0,        // use pipe, no interrupt signal
                          int leader_type = 0,    // 0-shared /1-dedicated
                          int start_aio_type = 1); // 1-any thread/ 0-only leader
};

class TPROACTOR_Export TRB_Linux_RT_Proactor: public TRB_POSIX_Proactor 
{
public:
  TRB_Linux_RT_Proactor (size_t max_op = ACE_AIO_DEFAULT_SIZE,
                          int sig_num = 0,        // use pipe, no interrupt signal
                          int leader_type = 1,    // 0-shared /1-dedicated
                          int start_aio_type = 1); // 1-any thread/ 0-only leader
};

class TPROACTOR_Export TRB_SUN_Port_Proactor: public TRB_POSIX_Proactor 
{
public:
  TRB_SUN_Port_Proactor (size_t max_op = ACE_AIO_DEFAULT_SIZE,
                          int sig_num = 0,        // use pipe, no interrupt signal
                          int leader_type = 0,    // 0-shared /1-dedicated
                          int start_aio_type = 1); // 1-any thread/ 0-only leader
};

class TPROACTOR_Export TRB_Dummy_Proactor: public TRB_POSIX_Proactor 
{
public:
  TRB_Dummy_Proactor(void);
};

#endif /* ACE_HAS_AIO_CALLS */
#endif //TRB_POSIX_PROACTOR_H 
